@@ -46,6 +46,7 @@
 #include <wicked/bridge.h>
 #include <wicked/vlan.h>
 #include <wicked/fsm.h>
+#include <wicked/wireless.h>
 
 #include "wicked-client.h"
 #include "appconfig.h"
@@ -92,6 +93,7 @@ ni_ifstatus_code_name(unsigned int status)
 		{ "no-device",			NI_WICKED_ST_NO_DEVICE		},
 		{ "device-not-running",		NI_WICKED_ST_NOT_RUNNING	},
 		{ "no-config",			NI_WICKED_ST_NO_CONFIG		},
+		{ "no-carrier",			NI_WICKED_ST_NO_CARRIER		},
 		{ "setup-in-progress",		NI_WICKED_ST_IN_PROGRESS	},
 		{ "config-changed",		NI_WICKED_ST_CHANGED_CONFIG	},
 		{ "enslaved",			NI_WICKED_ST_ENSLAVED		},
@@ -211,10 +213,44 @@ __ifstatus_of_device_lease(ni_netdev_t *dev, ni_addrconf_lease_t *lease, unsigne
 	}
 }
 
+static ni_addrconf_lease_t*
+ifstatus_primary_lease(ni_netdev_t *dev,
+		ni_addrconf_lease_t *lease)
+{
+	ni_addrconf_lease_t *l;
+
+	if (!ni_addrconf_flag_bit_is_set(lease->flags, NI_ADDRCONF_FLAGS_FALLBACK))
+		return NULL;
+
+	for (l = dev->leases; l; l = l->next) {
+		if (l->family != lease->family)
+			continue;
+
+		if (!ni_addrconf_flag_bit_is_set(l->flags, NI_ADDRCONF_FLAGS_PRIMARY))
+			continue;
+
+		if (ni_log_level_at(NI_LOG_DEBUG1)) {
+			ni_stringbuf_t buf = NI_STRINGBUF_INIT_DYNAMIC;
+
+			ni_addrconf_flags_format(&buf, l->flags, "|");
+			ni_debug_application("%s: primary lease %s:%s, state=%s, flags=%s",
+					dev->name,
+					ni_addrfamily_type_to_name(l->family),
+					ni_addrconf_type_to_name(l->type),
+					ni_addrconf_state_to_name(l->state),
+					buf.string);
+			ni_stringbuf_destroy(&buf);
+		}
+
+		return l;
+	}
+	return NULL;
+}
+
 static void
 __ifstatus_of_device_leases(ni_netdev_t *dev, unsigned int *st)
 {
-	ni_addrconf_lease_t *lease;
+	ni_addrconf_lease_t *lease, *primary;
 
 	if (!dev || !st)
 		return;
@@ -239,6 +275,16 @@ __ifstatus_of_device_leases(ni_netdev_t *dev, unsigned int *st)
 					NI_ADDRCONF_FLAGS_OPTIONAL))
 			continue;
 
+		/* Do not consider fallback lease status, if there is a
+		 * corresponding primary+granted lease */
+		if ((primary = ifstatus_primary_lease(dev, lease))) {
+			unsigned int tmp = NI_WICKED_ST_OK;
+
+			__ifstatus_of_device_lease(dev, primary, &tmp);
+			if (tmp == NI_WICKED_ST_OK)
+				continue;
+		}
+
 		__ifstatus_of_device_lease(dev, lease, st);
 	}
 }
@@ -291,7 +337,7 @@ __ifstatus_of_device(ni_netdev_t *dev)
 
 	if (!ni_ifcheck_device_link_is_up(dev) &&
 	    ni_ifcheck_device_link_required(dev))
-		return NI_WICKED_ST_IN_PROGRESS;
+		return NI_WICKED_ST_NO_CARRIER;
 
 	__ifstatus_of_device_leases(dev, &st);
 #if 0
@@ -372,6 +418,23 @@ ni_ifstatus_show_iflink(const ni_netdev_t *dev, ni_bool_t verbose)
 	printf("\n");
 }
 
+static inline void
+ni_ifstatus_show_iftype_wireless(const ni_wireless_t *wlan, ni_bool_t verbose)
+{
+	ni_stringbuf_t ssid = NI_STRINGBUF_INIT_DYNAMIC;
+
+	printf(", state %s", ni_wireless_assoc_state_to_name(wlan->assoc.state));
+
+	if (wlan->assoc.bssid.len > 0){
+		printf(", ssid %s", ni_wireless_ssid_print(&wlan->assoc.ssid, &ssid));
+
+		if (wlan->assoc.auth_mode)
+			printf(", %s", wlan->assoc.auth_mode);
+	}
+
+	ni_stringbuf_destroy(&ssid);
+}
+
 static inline void
 ni_ifstatus_show_iftype(const ni_netdev_t *dev, ni_bool_t verbose)
 {
@@ -396,6 +459,11 @@ ni_ifstatus_show_iftype(const ni_netdev_t *dev, ni_bool_t verbose)
 		}
 		break;
 
+	case NI_IFTYPE_WIRELESS:
+		if (dev->wireless)
+			ni_ifstatus_show_iftype_wireless(dev->wireless, verbose);
+		break;
+
 	default:
 		/* ... */
 		break;
@@ -428,6 +496,31 @@ ni_ifstatus_show_iftype(const ni_netdev_t *dev, ni_bool_t verbose)
 	printf("\n");
 }
 
+
+static inline void
+ni_ifstatus_show_wireless(const ni_netdev_t *dev, ni_bool_t verbose)
+{
+	struct timeval now;
+	int duration;
+	ni_wireless_t *wlan = dev->wireless;
+
+	if (!wlan || !verbose)
+		return;
+
+	if (wlan->assoc.bssid.len > 0){
+		if (ni_timer_get_time(&now) == 0)
+			duration = now.tv_sec - wlan->assoc.established_time.tv_sec;
+		else
+			duration = 0;
+
+		if_printf("", "wireless:", "bssid %s, signal %d, channel %u (%uMHz), duration %us\n",
+			ni_link_address_print(&wlan->assoc.bssid),
+			wlan->assoc.signal, ni_wireless_frequency_to_channel(wlan->assoc.frequency),
+			wlan->assoc.frequency, duration);
+
+	}
+}
+
 static inline void
 ni_ifstatus_show_addrs(ni_netdev_t *dev, ni_bool_t verbose)
 {
@@ -569,10 +662,14 @@ __show_leases_by_family(const ni_netdev_t *dev, ni_bool_t verbose, sa_family_t f
 			ni_addrconf_type_to_name(lease->type),
 			ni_addrconf_state_to_name(lease->state));
 
-		if (verbose && lease->flags) {
+		if ((verbose && lease->flags) ||
+		    ni_addrconf_flag_bit_is_set(lease->flags, NI_ADDRCONF_FLAGS_FALLBACK)) {
+			ni_stringbuf_t tmp = NI_STRINGBUF_INIT_DYNAMIC;
 			ni_stringbuf_puts(&buf, " [");
-			ni_addrconf_flags_format(&buf, lease->flags, ",");
+			ni_addrconf_flags_format(&tmp, lease->flags, ",");
+			ni_stringbuf_puts(&buf, tmp.string);
 			ni_stringbuf_puts(&buf, "]");
+			ni_stringbuf_destroy(&tmp);
 		}
 	}
 	if (buf.string)
@@ -608,6 +705,7 @@ ni_ifstatus_show_control(const ni_netdev_t *dev, ni_bool_t verbose)
 		ni_stringbuf_printf(&buf, "none");
 
 	if_printf("", "control:", "%s\n", buf.string);
+	ni_stringbuf_destroy(&buf);
 }
 
 static void
@@ -747,7 +845,7 @@ ni_do_ifstatus(int argc, char **argv)
 			ni_string_array_copy(&opt_ifconfig, sources);
 	}
 
-	if (!ni_ifconfig_load(fsm, opt_global_rootdir, &opt_ifconfig, TRUE, TRUE)) {
+	if (!ni_ifconfig_load(fsm, opt_global_rootdir, &opt_ifconfig, TRUE)) {
 		status = NI_WICKED_ST_ERROR;
 		goto cleanup;
 	}
@@ -801,6 +899,7 @@ ni_do_ifstatus(int argc, char **argv)
 		if (dev) {
 			ni_ifstatus_show_iflink (dev, opt_verbose > OPT_NORMAL);
 			ni_ifstatus_show_iftype (dev, opt_verbose > OPT_NORMAL);
+			ni_ifstatus_show_wireless(dev, opt_verbose > OPT_NORMAL);
 
 			/* TODO: Hmm... this is the running config only;
 			 *              show current config info too?
@@ -834,6 +933,7 @@ ni_do_ifstatus(int argc, char **argv)
 			case NI_WICKED_ST_UNCONFIGURED:
 			case NI_WICKED_ST_NOT_RUNNING:
 			case NI_WICKED_ST_IN_PROGRESS:
+			case NI_WICKED_ST_NO_CARRIER:
 				break;
 			default:
 				status = NI_WICKED_ST_OK;
@@ -950,6 +1050,7 @@ ni_ifstatus_display_result(ni_fsm_t *fsm, ni_string_array_t *names, ni_ifworker_
 			case NI_WICKED_ST_UNCONFIGURED:
 			case NI_WICKED_ST_NOT_RUNNING:
 			case NI_WICKED_ST_IN_PROGRESS:
+			case NI_WICKED_ST_NO_CARRIER:
 				break;
 			default:
 				status = NI_WICKED_ST_OK;
@@ -965,5 +1066,7 @@ ni_ifstatus_display_result(ni_fsm_t *fsm, ni_string_array_t *names, ni_ifworker_
 			status = rc;
 	}
 
+	ni_uint_array_destroy(&stcodes);
+	ni_uint_array_destroy(&stflags);
 	return status;
 }
