@@ -37,6 +37,7 @@
 
 #include <wicked/netinfo.h>
 #include <wicked/logging.h>
+#include <wicked/socket.h>
 #include <wicked/fsm.h>
 
 #include "client/ifconfig.h"
@@ -48,114 +49,124 @@
 
 struct ni_nanny_fsm_monitor {
 	const ni_timer_t *      timer;
-	unsigned long		timeout;
+	ni_timeout_t		timeout;
 	ni_ifworker_array_t *	marked;
 };
 
-static xml_node_t *
-__ni_ifup_generate_match_dev(xml_node_t *node, ni_ifworker_t *w)
+static ni_bool_t
+ni_ifup_generate_match_device(xml_node_t *node, ni_ifworker_t *w)
 {
+	/*
+	 * Generate <device>${w->name}</device> match (self reference) node.
+	 */
 	if (!node || !w || ni_string_empty(w->name))
-		return NULL;
-	return xml_node_new_element(NI_NANNY_IFPOLICY_MATCH_DEV, node, w->name);
+		return FALSE;
+
+	return !!xml_node_new_element(NI_NANNY_IFPOLICY_MATCH_DEV, node, w->name);
 }
 
 static ni_bool_t
-__ni_ifup_generate_match_link_port_ref(xml_node_t *match, xml_node_t *port)
-{
-	const char *type = xml_node_get_attr(port, NI_CLIENT_IFCONFIG_PORT_TYPE);
-	ni_iftype_t ptype = ni_linktype_name_to_type(type);
-	xml_node_t *ref, *ovsbr;
-
-	switch (ptype) {
-	case NI_IFTYPE_OVS_BRIDGE:
-		ovsbr = xml_node_get_child(port, NI_CLIENT_IFCONFIG_BRIDGE);
-		if (!ovsbr || ni_string_empty(ovsbr->cdata))
-			return FALSE;
-
-		if (!(ref = xml_node_new(NI_NANNY_IFPOLICY_MATCH_REF, match)))
-			return FALSE;
-
-		if (!xml_node_new_element(NI_NANNY_IFPOLICY_MATCH_DEV, ref, ovsbr->cdata)) {
-			xml_node_free(ref);
-			return FALSE;
-		}
-		break;
-
-	default:
-		/* other port types need master only */
-		break;
-	}
-	return TRUE;
-}
-
-static ni_bool_t
-__ni_ifup_generate_match_link_ref(xml_node_t *match, xml_node_t *link)
-{
-	xml_node_t *ref, *master, *port;
-
-	if (!(master = xml_node_get_child(link, NI_CLIENT_IFCONFIG_MASTER)))
-		return TRUE; /* <link> does not contain a <master> node */
-
-	if (ni_string_empty(master->cdata))
-		return FALSE;
-
-	if (ni_string_eq(master->cdata, ni_linktype_type_to_name(NI_IFTYPE_OVS_SYSTEM))) {
-		if ((port = xml_node_get_child(link, NI_CLIENT_IFCONFIG_LINK_PORT)))
-			return __ni_ifup_generate_match_link_port_ref(match, port);
-
-		return FALSE;
-	}
-
-	if (!(ref = xml_node_new(NI_NANNY_IFPOLICY_MATCH_REF, match)))
-		return FALSE;
-
-	if (!xml_node_new_element(NI_NANNY_IFPOLICY_MATCH_DEV, ref, master->cdata)) {
-		xml_node_free(ref);
-		return FALSE;
-	}
-
-	if ((port = xml_node_get_child(link, NI_CLIENT_IFCONFIG_LINK_PORT)))
-		return __ni_ifup_generate_match_link_port_ref(match, port);
-
-	return TRUE; /* master ref at least */
-}
-
-static ni_bool_t
-__ni_ifup_generate_match_master_ref(xml_node_t *match, ni_ifworker_t *master)
+ni_ifup_generate_match_device_ref(xml_node_t *match, const char *name)
 {
 	xml_node_t *ref;
 
-	if (!master || ni_string_empty(master->name))
+	/*
+	 * Generate <reference><device>${name}</device></reference>
+	 * match node to a related (lower, master, …) interface name.
+	 */
+	if (!match || ni_string_empty(name))
 		return FALSE;
 
 	if (!(ref = xml_node_new(NI_NANNY_IFPOLICY_MATCH_REF, match)))
 		return FALSE;
 
-	if (!xml_node_new_element(NI_NANNY_IFPOLICY_MATCH_DEV, ref, master->name)) {
-		xml_node_free(ref);
-		return FALSE;
-	}
+	if (xml_node_new_element(NI_NANNY_IFPOLICY_MATCH_DEV, ref, name))
+		return TRUE;
 
-	return TRUE;
+	xml_node_detach(ref);
+	xml_node_free(ref);
+	return FALSE;
 }
 
 static ni_bool_t
-__ni_ifup_generate_match_refs(xml_node_t *match, ni_ifworker_t *w)
+ni_ifup_generate_match_lower_ref(xml_node_t *match, ni_ifworker_t *w)
 {
-	xml_node_t *link;
+	/*
+	 * If we have a lower/underlying interface (e.g. lower=bond0)
+	 * reference and our (e.g. vlan bond0.42) interface can't
+	 * be setup or even can't be created or exist without it,
+	 * e.g.:
+	 *     ip link add dev bond0.42 link bond0 type vlan id 42
+	 *
+	 * In the config, the lower reference does not have a fix
+	 * node name, but is provided by type specific nodes, e.g.
+	 * for vlan|macvlan:
+	 *     <vlan|macvlan|…>
+	 *       <device>${w->lowerdev}</device>
+	 *     </vlan|macvlan|…>
+	 * for ovs vlan (fake) bridge's parent bridge:
+	 *     <ovs-bridge>
+	 *       <vlan>
+	 *         <parent>${w->lowerdev}</device>
+	 *       </vlan>
+	 *     </ovs-bridge>
+	 *
+	 * The w->lowerdev points to the worker resolved from the
+	 * config by fsm while (config) hierarchy build, so we
+	 * don't need to "find it manually".
+	 */
+	if (!w->lowerdev)
+		return TRUE;
 
-	if (w->masterdev)
-		return __ni_ifup_generate_match_master_ref(match, w->masterdev);
+	return ni_ifup_generate_match_device_ref(match, w->lowerdev->name);
+}
 
-	if ((link = xml_node_get_child(w->config.node, NI_CLIENT_IFCONFIG_LINK)))
-		return __ni_ifup_generate_match_link_ref(match, link);
+static ni_bool_t
+ni_ifup_generate_match_master_ref(xml_node_t *match, ni_ifworker_t *w)
+{
+	/*
+	 * If we are a port (e.g. bond0.42 vlan) bound to a master
+	 * (e.g. br42 bridge) required for the link setup, e.g.:
+	 *     ip link set dev bond0.42 master br42
+	 *
+	 * The <link> node provides a typed port union:
+	 *   <link>
+	 *     <master>${w->masterdev}</master>
+	 *     <port type="${type-name}">${type-data}</port>
+	 *   </link>
+	 *
+	 * The port type is the type of the effective master and
+	 * its content is type specific port configuration node.
+	 *
+	 * The w->masterdev points to the worker resolved from the
+	 * config by fsm while (config) hierarchy build, so we
+	 * don't need to "find it manually".
+	 */
+	if (!w->masterdev)
+		return TRUE;
 
-	return TRUE; /* no refs is not an error */
+	return ni_ifup_generate_match_device_ref(match, w->masterdev->name);
+}
+
+static ni_bool_t
+ni_ifup_generate_match_refs(xml_node_t *match, ni_ifworker_t *w)
+{
+	/*
+	 * Some interface types may have a lower (underlying)
+	 * interface and (most) can be also bound as port to
+	 * an another master reference.
+	 */
+	if (!ni_ifup_generate_match_lower_ref(match, w))
+		return FALSE;
+
+	if (!ni_ifup_generate_match_master_ref(match, w))
+		return FALSE;
+
+	return TRUE; /* no refa is not an error */
 }
 
 static xml_node_t *
-__ni_ifup_generate_match(const char *name, ni_ifworker_t *w)
+ni_ifup_generate_match(const char *name, ni_ifworker_t *w)
 {
 	xml_node_t *match;
 
@@ -166,45 +177,25 @@ __ni_ifup_generate_match(const char *name, ni_ifworker_t *w)
 		"generate policy match for %s (type %s)", w->name,
 		ni_linktype_type_to_name(w->iftype));
 
-	if (!__ni_ifup_generate_match_dev(match, w))
-		goto error;
-
-	/* Ignore child dependency for following device types:
-	 *  - ovs-system: otherwise ovs-system would require all ports
-	 *    in all ovs-bridges and want to get at least one up ...
-	 *    this is not what we want :-)
-	 */
-	switch (w->iftype) {
-	case NI_IFTYPE_OVS_SYSTEM:
-		goto done;
-		break;
-	default:
-		if (ni_string_eq(w->name, ni_linktype_type_to_name(NI_IFTYPE_OVS_SYSTEM)))
-			goto done;
-		break;
-	}
-
-	if (!__ni_ifup_generate_match_refs(match, w)) {
-		ni_debug_application("%s: unable to generate policy match device references",
+	if (!ni_ifup_generate_match_device(match, w)) {
+		ni_error("%s: unable to generate policy device match",
 				w->name);
 		goto error;
 	}
 
-	if (w->children.count) {
-		xml_node_t *or;
-		unsigned int i;
+	/*
+	 * The ovs-system datapath does not have any
+	 * (should not have any) references to other
+	 * interfaces...
+	 */
+	if (w->iftype == NI_IFTYPE_OVS_SYSTEM ||
+	    ni_string_eq(w->name, ni_linktype_type_to_name(NI_IFTYPE_OVS_SYSTEM)))
+		goto done;
 
-		if (!(or = xml_node_new(NI_NANNY_IFPOLICY_MATCH_COND_OR, match)))
-			goto error;
-
-		for (i = 0; i < w->children.count; i++) {
-			ni_ifworker_t *child = w->children.data[i];
-			xml_node_t *cnode;
-
-			cnode = xml_node_new(NI_NANNY_IFPOLICY_MATCH_COND_CHILD, or);
-			if (!cnode || !__ni_ifup_generate_match_dev(cnode, child))
-				goto error;
-		}
+	if (!ni_ifup_generate_match_refs(match, w)) {
+		ni_error("%s: unable to generate policy device reference match",
+				w->name);
+		goto error;
 	}
 
 done:
@@ -226,7 +217,7 @@ ni_ifup_start_policy(ni_ifworker_t *w)
 
 	ni_debug_application("%s: hiring nanny", w->name);
 
-	match = __ni_ifup_generate_match(NI_NANNY_IFPOLICY_MATCH, w);
+	match = ni_ifup_generate_match(NI_NANNY_IFPOLICY_MATCH, w);
 	if (!match)
 		goto error;
 
@@ -299,6 +290,79 @@ ni_ifup_hire_nanny(ni_ifworker_array_t *array, ni_bool_t set_persistent)
  * We want to wait for this signal and when it is >= device-up return TRUE.
  * After timeout we fail...
  */
+static ni_bool_t
+ni_nanny_fsm_monitor_wait_by_startmode(ni_ifworker_t *worker)
+{
+	/*
+	 * We don't need to wait for hotplug interfaces,
+	 * but only for "boot" and "manual" (started by
+	 * dedicated "ifup <name>" but not in "ifup all").
+	 */
+	if (ni_string_eq(worker->control.mode, "hotplug"))
+		return FALSE;
+	if (ni_string_eq(worker->control.mode, "ifplugd"))
+		return FALSE;
+
+	return TRUE;
+}
+
+static ni_bool_t
+ni_nanny_fsm_monitor_remove_port(ni_ifworker_array_t *workers, ni_ifworker_t *port)
+{
+	/*
+	 * A L2 master (bond,team,bridge,ovs-bridge) calculates it's
+	 * link-up (carrier/up & running) state from it's ports using
+	 * own logic (e.g. monitoring or stp) and usually requires one
+	 * port only to inherit carrier itself.
+	 *
+	 * While a bond/team are HA constructs ("cable redundancy" to
+	 * the same "peer" switch), bridge acts as switch and connects
+	 * multiple ports (the broadcast domains / switches behind the
+	 * "cable") with each other into one (broadcast domain[s] set).
+	 * Note: A bridge with disabled STP is often mocking carrier,
+	 * that is, it may report carrier even there is no port added.
+	 *
+	 * When master finished/reached it's target state, we can stop
+	 * monitoring setup of it's hotplug ports, which are allowed to
+	 * be missed completely, but in case a port is a non-hotplug one
+	 * (what makes sense for bridges), we should NOT remove it from
+	 * further monitoring.
+	 *
+	 * Both, hotplug and non-hotplug ports as well as the master
+	 * itself, may be also configured to not require link aka to
+	 * ignore carrier. In this case, the setup of the interface
+	 * will finish earlier, but we still need to wait until the
+	 * "up" run ignoring carrier finished for a non-hotplug port.
+	 */
+	if (ni_nanny_fsm_monitor_wait_by_startmode(port))
+		return FALSE;
+
+	return ni_ifworker_array_remove(workers, port);
+}
+
+static void
+ni_nanny_fsm_monitor_remove_ports(ni_ifworker_array_t *workers, ni_ifworker_t *worker)
+{
+	ni_ifworker_t *port;
+	unsigned int i;
+
+	for (i = 0; i < workers->count; ) {
+		port = workers->data[i];
+
+		if (port->masterdev == worker &&
+		    ni_nanny_fsm_monitor_remove_port(workers, port))
+			continue;
+		i++;
+	}
+}
+
+static void
+ni_nanny_fsm_monitor_remove_finished(ni_ifworker_array_t *workers, ni_ifworker_t *worker)
+{
+	ni_nanny_fsm_monitor_remove_ports(workers, worker);
+	ni_ifworker_array_remove(workers, worker);
+}
+
 static void
 ni_nanny_fsm_monitor_handler(ni_dbus_connection_t *conn, ni_dbus_message_t *msg, void *user_data)
 {
@@ -342,7 +406,7 @@ ni_nanny_fsm_monitor_handler(ni_dbus_connection_t *conn, ni_dbus_message_t *msg,
 		if (!ni_string_eq(w->name, ifname))
 			continue;
 
-		ni_ifworker_array_remove_with_children(monitor->marked, w);
+		ni_nanny_fsm_monitor_remove_finished(monitor->marked, w);
 		break;
 	}
 
@@ -355,9 +419,9 @@ ni_nanny_fsm_monitor_timeout(void *user_data, const ni_timer_t *timer)
 	ni_nanny_fsm_monitor_t *monitor = user_data;
 
 	if (monitor && timer == monitor->timer) {
+		ni_warn("Interface wait time (%us) reached", NI_TIMEOUT_SEC(monitor->timeout));
 		monitor->timer = NULL;
 		monitor->timeout = 0;
-		ni_info("Interface wait time reached");
 	}
 }
 
@@ -386,7 +450,7 @@ ni_nanny_fsm_monitor_new(ni_fsm_t *fsm)
 }
 
 ni_bool_t
-ni_nanny_fsm_monitor_arm(ni_nanny_fsm_monitor_t *monitor, unsigned long timeout)
+ni_nanny_fsm_monitor_arm(ni_nanny_fsm_monitor_t *monitor, ni_timeout_t timeout)
 {
 	if (monitor) {
 		monitor->timeout = timeout;
@@ -408,14 +472,14 @@ ni_nanny_fsm_monitor_run(ni_nanny_fsm_monitor_t *monitor, ni_ifworker_array_t *m
 
 	monitor->marked = ni_ifworker_array_clone(marked);
 	while (!ni_caught_terminal_signal()) {
-		long timeout;
+		ni_timeout_t timeout;
 
 		if (!monitor->marked || !monitor->marked->count)
 			break;
 
 		timeout = ni_timer_next_timeout();
 		if (monitor->timeout == 0 ||
-		    (monitor->timeout > 0 && timeout < 0))
+		    (monitor->timeout > 0 && timeout == NI_TIMEOUT_INFINITE))
 			break;
 
 		if (ni_socket_wait(timeout) != 0)
@@ -449,11 +513,12 @@ ni_nanny_fsm_monitor_free(ni_nanny_fsm_monitor_t *monitor)
 	free(monitor);
 }
 
-static int
-ni_do_ifup_nanny(int argc, char **argv)
+int
+ni_do_ifup(int argc, char **argv)
 {
 	enum  { OPT_HELP, OPT_IFCONFIG, OPT_CONTROL_MODE, OPT_STAGE, OPT_TIMEOUT,
 		OPT_SKIP_ACTIVE, OPT_SKIP_ORIGIN, OPT_PERSISTENT, OPT_TRANSIENT,
+		OPT_DRY_RUN,
 #ifdef NI_TEST_HACKS
 		OPT_IGNORE_PRIO, OPT_IGNORE_STARTMODE,
 #endif
@@ -473,28 +538,28 @@ ni_do_ifup_nanny(int argc, char **argv)
 		{ "ignore-startmode",no_argument, NULL,	OPT_IGNORE_STARTMODE },
 #endif
 		{ "persistent",	no_argument, NULL,	OPT_PERSISTENT },
+		{ "dry-run",	no_argument, NULL,	OPT_DRY_RUN },
 		{ NULL }
 	};
 
 	ni_ifmatcher_t ifmatch;
-	ni_ifworker_array_t ifmarked;
+	ni_ifworker_array_t ifmarked = NI_IFWORKER_ARRAY_INIT;
 	ni_nanny_fsm_monitor_t *monitor = NULL;
 	ni_string_array_t opt_ifconfig = NI_STRING_ARRAY_INIT;
 	ni_string_array_t ifnames = NI_STRING_ARRAY_INIT;
 	ni_bool_t check_prio = TRUE, set_persistent = FALSE;
 	ni_bool_t opt_transient = FALSE;
+	ni_log_fn_t *dry_run = NULL;
 	int c, status = NI_WICKED_RC_USAGE;
-	unsigned int timeout = 0;
+	unsigned int seconds = 0;
 	ni_fsm_t *fsm;
 
 	fsm = ni_fsm_new();
 	ni_assert(fsm);
 	ni_fsm_require_register_type("reachable", ni_ifworker_reachability_check_new);
 
-	memset(&ifmatch, 0, sizeof(ifmatch));
-	memset(&ifmarked, 0, sizeof(ifmarked));
-
 	/* Allow ifup on all interfaces we have config for */
+	memset(&ifmatch, 0, sizeof(ifmatch));
 	ifmatch.require_configured = FALSE;
 	ifmatch.allow_persistent = TRUE;
 	ifmatch.require_config = TRUE;
@@ -515,16 +580,9 @@ ni_do_ifup_nanny(int argc, char **argv)
 			break;
 
 		case OPT_TIMEOUT:
-			if (!strcmp(optarg, "infinite")) {
-				timeout = NI_IFWORKER_INFINITE_TIMEOUT;
-			} else {
-				unsigned int sec;
-
-				if (ni_parse_uint(optarg, &sec, 10) < 0) {
-					ni_error("ifup: cannot parse timeout option \"%s\"", optarg);
-					goto usage;
-				}
-				timeout = sec * 1000; /* sec -> msec */
+			if (ni_parse_seconds_timeout(optarg, &seconds)) {
+				ni_error("ifup: cannot parse timeout option \"%s\"", optarg);
+				goto usage;
 			}
 			break;
 
@@ -546,12 +604,16 @@ ni_do_ifup_nanny(int argc, char **argv)
 			break;
 #endif
 
+		case OPT_TRANSIENT:
+			opt_transient = TRUE;
+			break;
+
 		case OPT_PERSISTENT:
 			set_persistent = TRUE;
 			break;
 
-		case OPT_TRANSIENT:
-			opt_transient = TRUE;
+		case OPT_DRY_RUN:
+			dry_run = ni_note;
 			break;
 
 		default:
@@ -586,6 +648,8 @@ usage:
 #endif
 				"  --persistent\n"
 				"      Set interface into persistent mode (no regular ifdown allowed)\n"
+				"  --dry-run\n"
+				"      Show interface hierarchies as notice with (+) markers and exit.\n"
 				);
 			goto cleanup;
 		}
@@ -615,18 +679,18 @@ usage:
 		}
 	}
 
-	if (!ni_ifconfig_load(fsm, opt_global_rootdir, &opt_ifconfig, check_prio, TRUE)) {
+	if (!ni_ifconfig_load(fsm, opt_global_rootdir, &opt_ifconfig, check_prio)) {
 		status = NI_WICKED_RC_NOT_CONFIGURED;
 		goto cleanup;
 	}
 
 	/* Set timeout how long the action is allowed to wait */
-	if (timeout) {
-		fsm->worker_timeout = timeout; /* One set by user */
+	if (seconds) {
+		fsm->worker_timeout = NI_TIMEOUT_FROM_SEC(seconds);
 	} else
 	if (ni_wait_for_interfaces) {
 		fsm->worker_timeout = ni_fsm_find_max_timeout(fsm,
-				ni_wait_for_interfaces*1000);
+				NI_TIMEOUT_FROM_SEC(ni_wait_for_interfaces));
 	} else {
 		fsm->worker_timeout = ni_fsm_find_max_timeout(fsm,
 				NI_IFWORKER_DEFAULT_TIMEOUT);
@@ -636,20 +700,19 @@ usage:
 		ni_debug_application("wait for interfaces infinitely");
 	else
 		ni_debug_application("wait %u seconds for interfaces",
-					fsm->worker_timeout/1000);
+					NI_TIMEOUT_SEC(fsm->worker_timeout));
 
 	ni_nanny_fsm_monitor_arm(monitor, fsm->worker_timeout);
 
 	if (ni_fsm_build_hierarchy(fsm, TRUE) < 0) {
-		ni_error("ifup: unable to build device hierarchy");
+		ni_error("ifup: unable to build interface hierarchy");
 		/* Severe error we always explicitly return */
 		status = NI_WICKED_RC_ERROR;
 		goto cleanup;
 	}
 
-	status = NI_WICKED_RC_SUCCESS;
-
 	/* Get workers that match given criteria */
+	status = NI_WICKED_RC_SUCCESS;
 	while (optind < argc) {
 		ifmatch.name = argv[optind++];
 
@@ -670,7 +733,10 @@ usage:
 			ni_string_array_append(&ifnames, ifmatch.name);
 	}
 
-	ni_fsm_pull_in_children(&ifmarked, fsm);
+	ni_fsm_print_system_hierarchy(fsm, NULL, dry_run);
+	ni_fsm_print_config_hierarchy(fsm, &ifmarked, dry_run);
+	if (dry_run)
+		goto cleanup;
 
 	if (!ni_ifup_hire_nanny(&ifmarked, set_persistent))
 		status = NI_WICKED_RC_NOT_CONFIGURED;
@@ -685,7 +751,7 @@ usage:
 
 	/*
 	 * Do not report any errors to systemd -- returning an error
-	 * here, will cause sytemd to stop the network completely.
+	 * here, will cause systemd to stop the network completely.
 	 */
 	if (opt_systemd)
 		status = NI_LSB_RC_SUCCESS;
@@ -695,290 +761,7 @@ cleanup:
 	ni_nanny_fsm_monitor_free(monitor);
 	ni_ifworker_array_destroy(&ifmarked);
 	ni_string_array_destroy(&opt_ifconfig);
+	ni_fsm_free(fsm);
 	return status;
 }
 
-static int
-ni_do_ifup_direct(int argc, char **argv)
-{
-	enum  { OPT_HELP, OPT_IFCONFIG, OPT_CONTROL_MODE, OPT_STAGE, OPT_TIMEOUT,
-		OPT_SKIP_ACTIVE, OPT_SKIP_ORIGIN, OPT_PERSISTENT, OPT_TRANSIENT,
-#ifdef NI_TEST_HACKS
-		OPT_IGNORE_PRIO, OPT_IGNORE_STARTMODE,
-#endif
-	};
-
-	static struct option ifup_options[] = {
-		{ "help",	no_argument,       NULL,	OPT_HELP },
-		{ "ifconfig",	required_argument, NULL,	OPT_IFCONFIG },
-		{ "mode",	required_argument, NULL,	OPT_CONTROL_MODE },
-		{ "boot-stage",	required_argument, NULL,	OPT_STAGE },
-		{ "skip-active",required_argument, NULL,	OPT_SKIP_ACTIVE },
-		{ "skip-origin",required_argument, NULL,	OPT_SKIP_ORIGIN },
-		{ "timeout",	required_argument, NULL,	OPT_TIMEOUT },
-		{ "transient", 	no_argument,		NULL,	OPT_TRANSIENT },
-#ifdef NI_TEST_HACKS
-		{ "ignore-prio",no_argument, NULL,	OPT_IGNORE_PRIO },
-		{ "ignore-startmode",no_argument, NULL,	OPT_IGNORE_STARTMODE },
-#endif
-		{ "persistent",	no_argument, NULL,	OPT_PERSISTENT },
-		{ NULL }
-	};
-
-	ni_ifmatcher_t ifmatch;
-	ni_ifmarker_t ifmarker;
-	ni_ifworker_array_t ifmarked;
-	ni_string_array_t opt_ifconfig = NI_STRING_ARRAY_INIT;
-	ni_string_array_t ifnames = NI_STRING_ARRAY_INIT;
-	ni_bool_t check_prio = TRUE;
-	ni_bool_t opt_transient = FALSE;
-	unsigned int nmarked;
-	ni_fsm_t *fsm;
-	int c, status = NI_WICKED_RC_USAGE;
-	unsigned int timeout = 0;
-	const char *ptr;
-
-	fsm = ni_fsm_new();
-	ni_assert(fsm);
-	ni_fsm_require_register_type("reachable", ni_ifworker_reachability_check_new);
-
-	memset(&ifmatch, 0, sizeof(ifmatch));
-	memset(&ifmarker, 0, sizeof(ifmarker));
-	memset(&ifmarked, 0, sizeof(ifmarked));
-
-	/* Allow ifup on all interfaces we have config for */
-	ifmatch.require_configured = FALSE;
-	ifmatch.allow_persistent = TRUE;
-	ifmatch.require_config = TRUE;
-
-	ifmarker.target_range.min = __NI_FSM_STATE_MAX - 1;
-	ifmarker.target_range.max = __NI_FSM_STATE_MAX;
-
-	/*
-	 * Workaround to consider WAIT_FOR_INTERFACES variable
-	 * in network/config (bnc#863371, bnc#862530 timeouts).
-	 * Correct would be to get it from compat layer, but
-	 * the network/config is sourced in systemd service...
-	 */
-	if ((ptr = getenv("WAIT_FOR_INTERFACES"))) {
-		unsigned int sec;
-
-		if (ni_parse_uint(ptr, &sec, 10) == 0 &&
-		    (sec * 1000 > fsm->worker_timeout)) {
-			ni_debug_application("wait %u sec for interfaces", sec);
-			timeout = sec * 1000;
-		}
-	}
-
-	optind = 1;
-	while ((c = getopt_long(argc, argv, "", ifup_options, NULL)) != EOF) {
-		switch (c) {
-		case OPT_IFCONFIG:
-			ni_string_array_append(&opt_ifconfig, optarg);
-			break;
-
-		case OPT_CONTROL_MODE:
-			ifmatch.mode = optarg;
-			break;
-
-		case OPT_STAGE:
-			ifmatch.boot_stage= optarg;
-			break;
-
-		case OPT_TIMEOUT:
-			if (!strcmp(optarg, "infinite")) {
-				timeout = NI_IFWORKER_INFINITE_TIMEOUT;
-			} else if (ni_parse_uint(optarg, &timeout, 10) >= 0) {
-				timeout *= 1000; /* sec -> msec */
-			} else {
-				ni_error("ifup: cannot parse timeout option \"%s\"", optarg);
-				goto usage;
-			}
-			break;
-
-		case OPT_SKIP_ORIGIN:
-			ifmatch.skip_origin = optarg;
-			break;
-
-		case OPT_SKIP_ACTIVE:
-			ifmatch.skip_active = TRUE;
-			break;
-
-#ifdef NI_TEST_HACKS
-		case OPT_IGNORE_PRIO:
-			check_prio = FALSE;
-			break;
-
-		case OPT_IGNORE_STARTMODE:
-			ifmatch.ignore_startmode = TRUE;
-			break;
-#endif
-
-		case OPT_PERSISTENT:
-			ifmarker.persistent = TRUE;
-			break;
-
-		case OPT_TRANSIENT:
-			opt_transient = TRUE;
-			break;
-
-		default:
-		case OPT_HELP:
-usage:
-			fprintf(stderr,
-				"wicked [options] ifup [ifup-options] <ifname ...>|all\n"
-				"\nSupported ifup-options:\n"
-				"  --help\n"
-				"      Show this help text.\n"
-				"  --transient\n"
-				"      Enable transient interface return codes\n"
-				"  --ifconfig <pathname>\n"
-				"      Read interface configuration(s) from file/directory rather than using system config\n"
-				"  --mode <label>\n"
-				"      Only touch interfaces with matching control <mode>\n"
-				"  --boot-stage <label>\n"
-				"      Only touch interfaces with matching <boot-stage>\n"
-				"  --skip-active\n"
-				"      Do not touch running interfaces\n"
-				"  --skip-origin <name>\n"
-				"      Skip interfaces that have a configuration origin of <name>\n"
-				"      Usually, you would use this with the name \"firmware\" to avoid\n"
-				"      touching interfaces that have been set up via firmware (like iBFT) previously\n"
-				"  --timeout <sec>\n"
-				"      Timeout after <sec> seconds\n"
-#ifdef NI_TEST_HACKS
-				"  --ignore-prio\n"
-				"      Ignore checking the config origin priorities\n"
-				"  --ignore-startmode\n"
-				"      Ignore checking the STARTMODE=off and STARTMODE=manual configs\n"
-#endif
-				"  --persistent\n"
-				"      Set interface into persistent mode (no regular ifdown allowed)\n"
-				);
-			goto cleanup;
-		}
-	}
-
-	if (optind >= argc) {
-		fprintf(stderr, "Missing interface argument\n");
-		goto usage;
-	}
-
-	if (!ni_fsm_create_client(fsm)) {
-		/* Severe error we always explicitly return */
-		status = NI_WICKED_RC_ERROR;
-		goto cleanup;
-	}
-
-	if (!ni_fsm_refresh_state(fsm)) {
-		/* Severe error we always explicitly return */
-		status = NI_WICKED_RC_ERROR;
-		goto cleanup;
-	}
-
-	if (opt_ifconfig.count == 0) {
-		const ni_string_array_t *sources = ni_config_sources("ifconfig");
-
-		if (sources && sources->count)
-			ni_string_array_copy(&opt_ifconfig, sources);
-
-		if (opt_ifconfig.count == 0) {
-			ni_error("ifup: unable to load interface config source list");
-			status = NI_WICKED_RC_NOT_CONFIGURED;
-			goto cleanup;
-		}
-	}
-
-	if (!ni_ifconfig_load(fsm, opt_global_rootdir, &opt_ifconfig, check_prio, TRUE)) {
-		status = NI_WICKED_RC_NOT_CONFIGURED;
-		goto cleanup;
-	}
-
-	/* Set timeout how long the action is allowed to wait */
-	if (timeout) {
-		fsm->worker_timeout = timeout; /* One set by user */
-	} else
-	if (ni_wait_for_interfaces) {
-		fsm->worker_timeout = ni_fsm_find_max_timeout(fsm,
-				ni_wait_for_interfaces*1000);
-	} else {
-		fsm->worker_timeout = ni_fsm_find_max_timeout(fsm,
-				NI_IFWORKER_DEFAULT_TIMEOUT);
-	}
-
-	if (fsm->worker_timeout == NI_IFWORKER_INFINITE_TIMEOUT)
-		ni_debug_application("wait for interfaces infinitely");
-	else
-		ni_debug_application("wait %u seconds for interfaces",
-					fsm->worker_timeout/1000);
-
-	if (ni_fsm_build_hierarchy(fsm, TRUE) < 0) {
-		ni_error("ifup: unable to build device hierarchy");
-		/* Severe error we always explicitly return */
-		status = NI_WICKED_RC_ERROR;
-		goto cleanup;
-	}
-
-	/* Get workers that match given criteria */
-	nmarked = 0;
-	while (optind < argc) {
-		ifmatch.name = argv[optind++];
-
-		if (!strcmp(ifmatch.name, "boot")) {
-			ifmatch.name = "all";
-			ifmatch.mode = "boot";
-		}
-
-		ni_fsm_get_matching_workers(fsm, &ifmatch, &ifmarked);
-
-		if (ni_string_eq(ifmatch.name, "all") ||
-		    ni_string_empty(ifmatch.name)) {
-			ni_string_array_destroy(&ifnames);
-			break;
-		}
-
-		if (ni_string_array_index(&ifnames, ifmatch.name) < 0)
-			ni_string_array_append(&ifnames, ifmatch.name);
-	}
-
-	ni_fsm_pull_in_children(&ifmarked, fsm);
-
-	/* Mark and start selected workers */
-	if (ifmarked.count)
-		nmarked = ni_fsm_mark_matching_workers(fsm, &ifmarked, &ifmarker);
-
-	if (nmarked == 0) {
-		ni_note("ifup: no matching interfaces");
-		status = NI_WICKED_RC_SUCCESS;
-	} else {
-		if (ni_fsm_schedule(fsm) != 0)
-			ni_fsm_mainloop(fsm);
-
-		ni_fsm_wait_tentative_addrs(fsm);
-
-		status = ni_ifstatus_display_result(fsm, &ifnames, &ifmarked,
-			opt_transient);
-
-		/*
-		 * Do not report any errors to systemd -- returning an error
-		 * here, will cause sytemd to stop the network completely.
-		 */
-		if (opt_systemd)
-			status = NI_LSB_RC_SUCCESS;
-	}
-
-cleanup:
-	ni_ifworker_array_destroy(&ifmarked);
-	ni_string_array_destroy(&ifnames);
-	ni_string_array_destroy(&opt_ifconfig);
-	return status;
-}
-
-int
-ni_do_ifup(int argc, char **argv)
-{
-	if (ni_config_use_nanny())
-		return ni_do_ifup_nanny(argc, argv);
-	else
-		return ni_do_ifup_direct(argc, argv);
-}
