@@ -4,7 +4,8 @@
  *		wicked:
  *		compat:
  *		firmware:
- *	along their assosciated sub-types.
+ *		dracut:
+ *	along their associated sub-types.
  *
  *	Copyright (C) 2010-2014 SUSE LINUX Products GmbH, Nuernberg, Germany.
  *
@@ -41,8 +42,12 @@
 #include <wicked/logging.h>
 #include <wicked/netinfo.h>
 
+#include "appconfig.h"
 #include "wicked-client.h"
 #include "client/ifconfig.h"
+#include "client/read-config.h"
+#include "dracut/dracut.h"
+#include "firmware.h"
 
 #if defined(COMPAT_AUTO) || defined(COMPAT_SUSE)
 extern ni_bool_t	__ni_suse_get_ifconfig(const char *, const char *,
@@ -53,39 +58,25 @@ extern ni_bool_t	__ni_redhat_get_ifconfig(const char *, const char *,
 						ni_compat_ifconfig_t *);
 #endif
 
-
-typedef struct ni_ifconfig_type	ni_ifconfig_type_t;
-struct ni_ifconfig_type {
-	const char *			name;
-	struct {
-		ni_bool_t		(*read)(xml_document_array_t *,
-						const char *,
-						const char *,
-						const char *,
-						ni_bool_t,
-						ni_bool_t);
-	    const ni_ifconfig_type_t *	(*guess)(const ni_ifconfig_type_t *,
-						const char *root,
-						const char *path);
-	} ops;
-};
-
 static ni_bool_t	ni_ifconfig_read_wicked(xml_document_array_t *,
 						const char *,
 						const char *,
 						const char *,
+						ni_ifconfig_kind_t,
 						ni_bool_t,
 						ni_bool_t);
 static ni_bool_t	ni_ifconfig_read_wicked_xml(xml_document_array_t *,
 						const char *,
 						const char *,
 						const char *,
+						ni_ifconfig_kind_t,
 						ni_bool_t,
 						ni_bool_t);
 static ni_bool_t	ni_ifconfig_read_compat(xml_document_array_t *,
 						const char *,
 						const char *,
 						const char *,
+						ni_ifconfig_kind_t,
 						ni_bool_t,
 						ni_bool_t);
 #if defined(COMPAT_AUTO) || defined(COMPAT_SUSE)
@@ -93,6 +84,7 @@ static ni_bool_t	ni_ifconfig_read_compat_suse(xml_document_array_t *,
 						const char *,
 						const char *,
 						const char *,
+						ni_ifconfig_kind_t,
 						ni_bool_t,
 						ni_bool_t);
 #endif
@@ -101,6 +93,7 @@ static ni_bool_t	ni_ifconfig_read_compat_redhat(xml_document_array_t *,
 						const char *,
 						const char *,
 						const char *,
+						ni_ifconfig_kind_t,
 						ni_bool_t,
 						ni_bool_t);
 #endif
@@ -108,13 +101,15 @@ static ni_bool_t	ni_ifconfig_read_firmware(xml_document_array_t *,
 						const char *,
 						const char *,
 						const char *,
+						ni_ifconfig_kind_t,
 						ni_bool_t,
 						ni_bool_t);
 
-static const ni_ifconfig_type_t *
-__ni_ifconfig_find_map(const ni_ifconfig_type_t *map, const char *name, size_t len)
+const ni_ifconfig_type_t *
+ni_ifconfig_find_map(const ni_ifconfig_type_t *map, const char *name, size_t len)
 {
 	const ni_ifconfig_type_t *pos = map;
+
 	if (pos) {
 		while (pos->name) {
 			if (name && strlen(pos->name) == len &&
@@ -126,13 +121,13 @@ __ni_ifconfig_find_map(const ni_ifconfig_type_t *map, const char *name, size_t l
 	return pos;
 }
 
-static const ni_ifconfig_type_t *
+const ni_ifconfig_type_t *
 ni_ifconfig_find_type(const ni_ifconfig_type_t *map, const char *root,
 			const char *path, const char *name, size_t len)
 {
 	const ni_ifconfig_type_t *ret;
 
-	ret = __ni_ifconfig_find_map(map, name, len);
+	ret = ni_ifconfig_find_map(map, name, len);
 
 	if (!name && ret && ret->ops.guess) {
 		ret = ret->ops.guess(map, root, path);
@@ -140,6 +135,38 @@ ni_ifconfig_find_type(const ni_ifconfig_type_t *map, const char *root,
 	return ret;
 }
 
+ni_bool_t
+ni_ifconfig_read_subtype(xml_document_array_t *array, const ni_ifconfig_type_t *type,
+			const char *root, const char *path, ni_ifconfig_kind_t kind,
+			ni_bool_t prio, ni_bool_t raw, const char *supertype)
+{
+	const ni_ifconfig_type_t *map;
+	const char *sub_path = path;
+	const char *sub_name = NULL;
+	char *sub_type = NULL;
+	ni_bool_t ret = FALSE;
+	size_t len;
+
+	if (!array || !type)
+		return ret;
+
+	len = strcspn(path, ":");
+	if (path[len] == ':') {
+		sub_name = len ? path : NULL;
+		sub_path = path + len + 1;
+	}
+
+	map = ni_ifconfig_find_type(type, root, path, sub_name, len);
+	if (map && map->name && map->ops.read) {
+		ni_string_printf(&sub_type, "%s:%s", supertype, map->name);
+		ret = map->ops.read(array, sub_type, root, sub_path, kind, prio, raw);
+		ni_string_free(&sub_type);
+	} else {
+		ni_error("Unsupported ifconfig type %s:%.*s", supertype, (int)len, sub_name);
+	}
+	return ret;
+}
+
 const ni_ifconfig_type_t *
 ni_ifconfig_guess_compat_type(const ni_ifconfig_type_t *map,
 				const char *root, const char *path)
@@ -147,42 +174,81 @@ ni_ifconfig_guess_compat_type(const ni_ifconfig_type_t *map,
 	(void)path;
 
 #ifdef COMPAT_SUSE
-	return __ni_ifconfig_find_map(map, "suse", sizeof("suse")-1);
+	return ni_ifconfig_find_map(map, "suse", sizeof("suse")-1);
 #endif
 #ifdef COMPAT_REDHAT
-	return __ni_ifconfig_find_map(map, "redhat", sizeof("redhat")-1);
+	return ni_ifconfig_find_map(map, "redhat", sizeof("redhat")-1);
 #endif
 #ifdef COMPAT_AUTO
-	if (ni_file_exists_fmt("%s%s", (root ? root : ""), "/etc/SuSE-release"))
-		return __ni_ifconfig_find_map(map, "suse", sizeof("suse")-1);
+	if (ni_file_exists_fmt("%s%s", (root ? root : ""), "/etc/SuSE-release") ||
+	    ni_file_exists_fmt("%s%s", (root ? root : ""), "/etc/SUSE-brand"))
+		return ni_ifconfig_find_map(map, "suse", sizeof("suse")-1);
 
 	if (ni_file_exists_fmt("%s%s", (root ? root : ""), "/etc/redhat-release"))
-		return __ni_ifconfig_find_map(map, "redhat", sizeof("redhat")-1);
+		return ni_ifconfig_find_map(map, "redhat", sizeof("redhat")-1);
 #endif
 
 	return NULL;
 }
 
+static inline ni_ifconfig_kind_t
+ni_ifconfig_kind_guess(ni_ifconfig_kind_t kind)
+{
+	switch (kind) {
+	case NI_IFCONFIG_KIND_POLICY:
+	case NI_IFCONFIG_KIND_CONFIG:
+		return kind;
+
+	case NI_IFCONFIG_KIND_DEFAULT:
+	default:
+		return NI_IFCONFIG_KIND_POLICY;
+	}
+}
+
+static const ni_intmap_t	ni_ifconfig_kind_map[] = {
+	{ "policy",		NI_IFCONFIG_KIND_POLICY		},
+	{ "config",		NI_IFCONFIG_KIND_CONFIG		},
+	{ NULL,			NI_IFCONFIG_KIND_DEFAULT	}
+};
+
+const char *
+ni_ifconfig_kind_to_name(ni_ifconfig_kind_t kind)
+{
+	return ni_format_uint_mapped(kind, ni_ifconfig_kind_map);
+}
+
+ni_bool_t
+ni_ifconfig_kind_by_name(const char *name, ni_ifconfig_kind_t *kind)
+{
+	unsigned int type;
+
+	if (!kind || ni_parse_uint_mapped(name, ni_ifconfig_kind_map, &type) < 0)
+		return FALSE;
+
+	*kind = type;
+	return TRUE;
+}
+
 const ni_ifconfig_type_t *
 ni_ifconfig_guess_wicked_type(const ni_ifconfig_type_t *map,
 			const char *root, const char *path)
 {
-	return __ni_ifconfig_find_map(map, "xml", sizeof("xml")-1);
+	return ni_ifconfig_find_map(map, "xml", sizeof("xml")-1);
 }
 
 const ni_ifconfig_type_t *
 ni_ifconfig_guess_type(const ni_ifconfig_type_t *map,
 			const char *root, const char *path)
 {
-	return __ni_ifconfig_find_map(map, "wicked", sizeof("wicked")-1);
+	return ni_ifconfig_find_map(map, "wicked", sizeof("wicked")-1);
 }
 
-static const ni_ifconfig_type_t	__ni_ifconfig_types_wicked[] = {
+static const ni_ifconfig_type_t		ni_ifconfig_types_wicked[] = {
 	{ "xml",	{ .read = ni_ifconfig_read_wicked_xml	} },
 	{ NULL,		{ .guess= ni_ifconfig_guess_wicked_type	} },
 };
 
-static const ni_ifconfig_type_t	__ni_ifconfig_types_compat[] = {
+static const ni_ifconfig_type_t		ni_ifconfig_types_compat[] = {
 #if defined(COMPAT_AUTO) || defined(COMPAT_SUSE)
 	{ "suse",	{ .read = ni_ifconfig_read_compat_suse	} },
 #endif
@@ -192,36 +258,42 @@ static const ni_ifconfig_type_t	__ni_ifconfig_types_compat[] = {
 	{ NULL,		{ .guess= ni_ifconfig_guess_compat_type } },
 };
 
-static const ni_ifconfig_type_t	__ni_ifconfig_types[] = {
+static const ni_ifconfig_type_t		ni_ifconfig_types[] = {
 	{ "wicked",	{ .read = ni_ifconfig_read_wicked	} },
 	{ "compat",	{ .read = ni_ifconfig_read_compat	} },
 	{ "firmware",	{ .read = ni_ifconfig_read_firmware	} },
+	{ "dracut",	{ .read = ni_ifconfig_read_dracut	} },
 	{ NULL,		{ .guess= ni_ifconfig_guess_type	} },
 };
 
 ni_bool_t
-ni_ifconfig_load(ni_fsm_t *fsm, const char *root, ni_string_array_t *opt_ifconfig, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_load(ni_fsm_t *fsm, const char *root, ni_string_array_t *sources, ni_bool_t check_prio)
 {
 	xml_document_array_t docs = XML_DOCUMENT_ARRAY_INIT;
+	ni_ifconfig_kind_t kind = NI_IFCONFIG_KIND_CONFIG; /* load into fsm */
 	unsigned int i;
 
-	for (i = 0; i < opt_ifconfig->count; ++i) {
-		if (!ni_ifconfig_read(&docs, root, opt_ifconfig->data[i], check_prio, raw)) {
-			xml_document_array_destroy(&docs);
-			return FALSE;
-		}
+	for (i = 0; i < sources->count; ++i) {
+		const char *source = sources->data[i];
+
+		if (ni_ifconfig_read(&docs, root, source, kind, check_prio, FALSE))
+			continue;
+
+		xml_document_array_destroy(&docs);
+		return FALSE;
 	}
 
+	if (ni_ifxml_migrate_docs(&docs))
+		ni_debug_readwrite("Migrated configuration to current schema");
+
 	for (i = 0; i < docs.count; i++) {
-		xml_node_t *root, *ifnode;
+		xml_node_t *root;
 		const char *origin;
 
 		root = xml_document_root(docs.data[i]);
-		origin = xml_node_location_filename(root);
-		for (ifnode = root->children; ifnode; ifnode = ifnode->next) {
-			/* We do not fail when unable to generate ifworker */
-			ni_fsm_workers_from_xml(fsm, ifnode, origin);
-		}
+		origin = ni_ifconfig_get_origin(root);
+		/* We do not fail when unable to generate ifworker */
+		ni_fsm_workers_from_xml(fsm, root, origin);
 	}
 
 	xml_document_array_destroy(&docs);
@@ -229,7 +301,7 @@ ni_ifconfig_load(ni_fsm_t *fsm, const char *root, ni_string_array_t *opt_ifconfi
 }
 
 ni_bool_t
-ni_ifconfig_read(xml_document_array_t *array, const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read(xml_document_array_t *array, const char *root, const char *path, ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
 	const ni_ifconfig_type_t *map;
 	const char *_path = path;
@@ -242,9 +314,9 @@ ni_ifconfig_read(xml_document_array_t *array, const char *root, const char *path
 		_path = path + len + 1;
 	}
 
-	map = ni_ifconfig_find_type(__ni_ifconfig_types, root, path, _name, len);
+	map = ni_ifconfig_find_type(ni_ifconfig_types, root, path, _name, len);
 	if (map && map->name && map->ops.read) {
-		return map->ops.read(array, map->name, root, _path, check_prio, raw);
+		return map->ops.read(array, map->name, root, _path, kind, check_prio, raw);
 	}
 
 	ni_error("Unsupported ifconfig type %.*s", (int)len, path);
@@ -252,7 +324,7 @@ ni_ifconfig_read(xml_document_array_t *array, const char *root, const char *path
 }
 
 static ni_config_origin_prio_t
-__ni_ifconfig_origin_get_prio(const char *origin)
+ni_ifconfig_origin_get_prio(const char *origin)
 {
 	ni_config_origin_prio_t prio;
 
@@ -273,66 +345,61 @@ __ni_ifconfig_origin_get_prio(const char *origin)
 
 /*
  * Parse name node and its namespace from xml config.
- * Set ifname if available - it should be at least IF_NAMESIZE bytes long.
  *
- * Return ifindex value or 0 if not available.
+ * Return true when interface name is available/can be resolved.
  */
-static char *
-__ifconfig_read_get_ifname(xml_node_t *ifnode, unsigned int *ifindex)
+static ni_bool_t
+ni_ifconfig_read_get_ifname(xml_node_t *ifnode, char **ifname)
 {
 	xml_node_t *nnode = NULL;
 	const char *namespace;
-	char *ifname = NULL;
+
+	if (!ifnode || !ifname)
+		return FALSE;
 
 	/* Check for   <name> node */
 	nnode = xml_node_get_child(ifnode, "name");
 	if (!nnode || ni_string_empty(nnode->cdata)) {
-		ni_debug_ifconfig("cannot get interface name - "
+		ni_error("cannot get interface name - "
 			"config has no valid <name> node");
-		goto error;
+		return FALSE;
 	}
 
-	ifname = nnode->cdata;
-
 	/* Resolve a namespace if specified */
 	namespace = xml_node_get_attr(nnode, "namespace");
 	if (ni_string_empty(namespace)) {
-		if (ifindex)
-			*ifindex = if_nametoindex(ifname);
+		if (!ni_netdev_name_is_valid(nnode->cdata)) {
+			ni_error("invalid interface name node data: '%s'",
+					nnode->cdata);
+			return FALSE;
+		}
+		if (!ni_string_dup(ifname, nnode->cdata)) {
+			ni_error("unable to allocate interface name: %m");
+			return FALSE;
+		}
 	}
 	else if (ni_string_eq(namespace, "ifindex")) {
 		unsigned int value;
-		char name_buf[IF_NAMESIZE+1];
 
-		if (ni_parse_uint(ifname, &value, 10) < 0) {
-			ni_debug_ifconfig("unable to parse ifindex value "
+		if (ni_parse_uint(nnode->cdata, &value, 10) < 0) {
+			ni_error("unable to parse ifindex value"
 				" specified via <name namespace=\"ifindex\">");
-			goto error;
+			return FALSE;
 		}
-
-		/* Get ifname based on ifindex */
-		if (ni_string_empty(if_indextoname(value, name_buf))) {
-			ni_debug_ifconfig("unable to obtain interface name "
-				"using ifindex value");
-			goto error;
+		if (!ni_netdev_index_to_name(ifname, value)) {
+			ni_error("unable to obtain interface name"
+				" using ifindex value %u", value);
+			return FALSE;
 		}
-
-		ifname = NULL;
-		ni_string_dup(&ifname, name_buf);
-
-		if (ifindex)
-			*ifindex = value;
 	}
 	else {
 		/* TODO: Implement other namespaces */;
+		ni_error("unable to resolve interface name from namespace '%s'",
+			namespace);
+		return FALSE;
 	}
 
-	return ifname;
-
-error:
-	if (ifindex)
-		*ifindex = 0;
-	return NULL;
+	return TRUE;
 }
 
 ni_bool_t
@@ -341,7 +408,7 @@ ni_ifconfig_validate_adding_doc(xml_document_t *config_doc, ni_bool_t check_prio
 	static ni_var_array_t validated_cfgs; /* Array of already processed configs */
 	ni_config_origin_prio_t src_prio, dst_prio;
 	xml_node_t *src_root, *src_child;
-	char *ifname;
+	char *ifname = NULL;
 
 	if (!config_doc)
 		return FALSE;
@@ -350,10 +417,13 @@ ni_ifconfig_validate_adding_doc(xml_document_t *config_doc, ni_bool_t check_prio
 		return TRUE;
 
 	src_root = xml_document_root(config_doc);
-	src_prio = __ni_ifconfig_origin_get_prio(xml_node_location_filename(src_root));
+	src_prio = ni_ifconfig_origin_get_prio(xml_node_location_filename(src_root));
 
-	/* Go through all config_doc's <interfaces> */
-	for (src_child = src_root->children; src_child; src_child = src_child->next) {
+	if (ni_string_empty(src_root->name))
+		src_child = src_root->children;
+	else
+		src_child = src_root;
+	for ( ; src_child; src_child = src_child->next) {
 		int rv;
 
 		if (ni_ifconfig_is_policy(src_child)) {
@@ -361,32 +431,36 @@ ni_ifconfig_validate_adding_doc(xml_document_t *config_doc, ni_bool_t check_prio
 			continue;
 		}
 
-		ifname = __ifconfig_read_get_ifname(src_child, NULL);
-		if (ni_string_empty(ifname))
-			return FALSE;
+		if (!ni_ifconfig_read_get_ifname(src_child, &ifname))
+			goto cleanup;
 
 		rv = ni_var_array_get_uint(&validated_cfgs, ifname, &dst_prio);
 		if (rv < 0)
-			return FALSE;
+			goto cleanup;
 
 		if (rv && dst_prio < src_prio) {
 			ni_warn("Ignoring %s config %s because of higher prio config",
 				ifname, xml_node_location_filename(src_root));
-			return FALSE;
+			goto cleanup;
 		}
 
 		ni_var_array_set_uint(&validated_cfgs, ifname, src_prio);
 	}
 
+	ni_string_free(&ifname);
 	return TRUE;
+cleanup:
+	ni_string_free(&ifname);
+	return FALSE;
 }
 
 /*
  * Read ifconfig
  */
 static ni_bool_t
-ni_ifconfig_read_wicked_xml_file(xml_document_array_t *docs, const char *type,
-			const char *root, const char *pathname, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_wicked_xml_file(xml_document_array_t *docs,
+			const char *type, const char *root, const char *pathname,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
 	ni_client_state_config_t conf = NI_CLIENT_STATE_CONFIG_INIT;
 	char pathbuf[PATH_MAX] = {'\0'};
@@ -422,8 +496,6 @@ ni_ifconfig_read_wicked_xml_file(xml_document_array_t *docs, const char *type,
 
 		xml_node_location_set(node, xml_location_clone(cnode->location));
 		xml_node_reparent(node, cnode);
-		if (ni_ifconfig_migrate(node))
-			ni_debug_readwrite("Migrated \"%s\" config to current schema", conf.origin);
 
 		ni_ifconfig_metadata_clear(node);
 		if (!raw)
@@ -431,10 +503,11 @@ ni_ifconfig_read_wicked_xml_file(xml_document_array_t *docs, const char *type,
 
 		if (ni_ifconfig_validate_adding_doc(doc, check_prio)) {
 			ni_debug_ifconfig("%s: %s", __func__, xml_node_location(node));
-			xml_document_array_append(docs, doc);
-		} else {
-			xml_document_free(doc);
+
+			if (xml_document_expand(docs, doc))
+				continue;
 		}
+		xml_document_free(doc);
 	}
 
 	ni_client_state_config_reset(&conf);
@@ -444,8 +517,9 @@ ni_ifconfig_read_wicked_xml_file(xml_document_array_t *docs, const char *type,
 }
 
 static ni_bool_t
-ni_ifconfig_read_wicked_xml_dir(xml_document_array_t *docs, const char *type,
-			const char *root, const char *pathname, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_wicked_xml_dir(xml_document_array_t *docs,
+			const char *type, const char *root, const char *pathname,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
 	char pathbuf[PATH_MAX] = {'\0'};
 	ni_string_array_t files = NI_STRING_ARRAY_INIT;
@@ -462,7 +536,9 @@ ni_ifconfig_read_wicked_xml_dir(xml_document_array_t *docs, const char *type,
 	if (ni_scandir(pathname, "*.xml", &files) != 0) {
 		for (i = 0; i < files.count; ++i) {
 			/* Ignore wrong xml config files - warning only */
-			if (ni_ifconfig_read_wicked_xml_file(docs, type, pathname, files.data[i], check_prio, raw))
+			if (ni_ifconfig_read_wicked_xml_file(docs, type,
+						pathname, files.data[i],
+						kind, check_prio, raw))
 				empty = FALSE;
 		}
 	}
@@ -475,8 +551,9 @@ ni_ifconfig_read_wicked_xml_dir(xml_document_array_t *docs, const char *type,
 }
 
 ni_bool_t
-ni_ifconfig_read_wicked_xml(xml_document_array_t *array, const char *type,
-			const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_wicked_xml(xml_document_array_t *array,
+			const char *type, const char *root, const char *path,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
 	char *ifconfig_dir = NULL;
 	char pathbuf[PATH_MAX];
@@ -496,44 +573,20 @@ ni_ifconfig_read_wicked_xml(xml_document_array_t *array, const char *type,
 
 	/* At the moment only XML is supported */
 	if (ni_isreg(pathbuf))
-		rv = ni_ifconfig_read_wicked_xml_file(array, type, root, path, check_prio, raw);
+		rv = ni_ifconfig_read_wicked_xml_file(array, type, root, path, kind, check_prio, raw);
 	else if (ni_isdir(pathbuf))
-		rv = ni_ifconfig_read_wicked_xml_dir(array, type, root, path, check_prio, raw);
+		rv = ni_ifconfig_read_wicked_xml_dir(array, type, root, path, kind, check_prio, raw);
 
 	ni_string_free(&ifconfig_dir);
 	return rv;
 }
 
 ni_bool_t
-ni_ifconfig_read_wicked(xml_document_array_t *array, const char *type,
-			const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_wicked(xml_document_array_t *array,
+			const char *type, const char *root, const char *path,
+			ni_ifconfig_kind_t kind, ni_bool_t prio, ni_bool_t raw)
 {
-	const ni_ifconfig_type_t *map;
-	const char *_path = path;
-	const char *_name = NULL;
-	char *_type = NULL;
-	size_t len;
-
-	len = strcspn(path, ":");
-	if (path[len] == ':') {
-		_name = len ? path : NULL;
-		_path = path + len + 1;
-	}
-
-	map = ni_ifconfig_find_type(__ni_ifconfig_types_wicked, root, path, _name, len);
-	if (map && map->name && map->ops.read) {
-		ni_string_printf(&_type, "%s:%s", type, map->name);
-		if (map->ops.read(array, _type, root, _path, check_prio, raw)) {
-			ni_string_free(&_type);
-			return TRUE;
-		}
-		ni_string_free(&_type);
-	}
-	else {
-		ni_error("Unsupported ifconfig type %s:%.*s", type, (int)len, _name);
-	}
-
-	return FALSE;
+	return ni_ifconfig_read_subtype(array, ni_ifconfig_types_wicked, root, path, kind, prio, raw, type);
 }
 
 /*
@@ -541,8 +594,9 @@ ni_ifconfig_read_wicked(xml_document_array_t *array, const char *type,
  */
 #if defined(COMPAT_AUTO) || defined(COMPAT_SUSE)
 ni_bool_t
-ni_ifconfig_read_compat_suse(xml_document_array_t *array, const char *type,
-			const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_compat_suse(xml_document_array_t *array,
+			const char *type, const char *root, const char *path,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
 	ni_compat_ifconfig_t conf;
 	ni_bool_t rv;
@@ -551,7 +605,16 @@ ni_ifconfig_read_compat_suse(xml_document_array_t *array, const char *type,
 
 	/* TODO: apply timeout */
 	if ((rv = __ni_suse_get_ifconfig(root, path, &conf))) {
-		ni_compat_generate_interfaces(array, &conf, check_prio, raw);
+#if 0
+		/* TODO:
+		 * we currently convert config to policy later
+		 */
+		kind = ni_ifconfig_kind_guess(kind);
+#endif
+		if (kind == NI_IFCONFIG_KIND_POLICY)
+			ni_compat_generate_policies(array, &conf, check_prio, raw);
+		else
+			ni_compat_generate_interfaces(array, &conf, check_prio, raw);
 	}
 	ni_compat_ifconfig_destroy(&conf);
 	return rv;
@@ -560,15 +623,25 @@ ni_ifconfig_read_compat_suse(xml_document_array_t *array, const char *type,
 
 #if defined(COMPAT_AUTO) || defined(COMPAT_REDHAT)
 ni_bool_t
-ni_ifconfig_read_compat_redhat(xml_document_array_t *array, const char *type,
-			const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_compat_redhat(xml_document_array_t *array,
+			const char *type, const char *root, const char *path,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
 	ni_compat_ifconfig_t conf;
 	ni_bool_t rv;
 
 	ni_compat_ifconfig_init(&conf, type);
 	if ((rv = __ni_redhat_get_ifconfig(root, path, &conf))) {
-		ni_compat_generate_interfaces(array, &conf, check_prio, raw);
+#if 0
+		/* TODO:
+		 * we currently convert config to policy later
+		 */
+		kind = ni_ifconfig_kind_guess(kind);
+#endif
+		if (kind == NI_IFCONFIG_KIND_POLICY)
+			ni_compat_generate_policies(array, &conf, check_prio, raw);
+		else
+			ni_compat_generate_interfaces(array, &conf, check_prio, raw);
 	}
 
 	ni_compat_ifconfig_destroy(&conf);
@@ -577,49 +650,26 @@ ni_ifconfig_read_compat_redhat(xml_document_array_t *array, const char *type,
 #endif
 
 ni_bool_t
-ni_ifconfig_read_compat(xml_document_array_t *array, const char *type,
-			const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_compat(xml_document_array_t *array,
+			const char *type, const char *root, const char *path,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
-	const ni_ifconfig_type_t *map;
-	const char *_path = path;
-	const char *_name = NULL;
-	char *_type = NULL;
-	size_t len;
-
-	len = strcspn(path, ":");
-	if (path[len] == ':') {
-		_name = len ? path : NULL;
-		_path = path + len + 1;
-	}
-
-	map = ni_ifconfig_find_type(__ni_ifconfig_types_compat, root, path, _name, len);
-	if (map && map->name && map->ops.read) {
-		ni_string_printf(&_type, "%s:%s", type, map->name);
-		if (map->ops.read(array, _type, root, _path, check_prio, raw)) {
-			ni_string_free(&_type);
-			return TRUE;
-		}
-		ni_string_free(&_type);
-	}
-	else {
-		ni_error("Unsupported ifconfig type %s:%.*s", type, (int)len, _name);
-	}
-
-	return FALSE;
+	return ni_ifconfig_read_subtype(array, ni_ifconfig_types_compat, root, path, kind, check_prio, raw, type);
 }
 
 ni_bool_t
-ni_ifconfig_read_firmware(xml_document_array_t *array, const char *type,
-			const char *root, const char *path, ni_bool_t check_prio, ni_bool_t raw)
+ni_ifconfig_read_firmware(xml_document_array_t *array,
+			const char *type, const char *root, const char *path,
+			ni_ifconfig_kind_t kind, ni_bool_t check_prio, ni_bool_t raw)
 {
-	xml_document_t *config_doc;
 	ni_client_state_config_t conf = NI_CLIENT_STATE_CONFIG_INIT;
+	xml_document_array_t docs = XML_DOCUMENT_ARRAY_INIT;
 	xml_node_t *rnode, *cnode, *next;
+	unsigned int i;
 
-	config_doc = ni_netconfig_firmware_discovery(root, path);
-	if (!config_doc) {
+	if (!ni_netif_firmware_discover_ifconfig(&docs, type, root, path)) {
 		ni_error("unable to get firmware interface definitions from %s:%s",
-			type, path);
+				type, path);
 		return FALSE;
 	}
 
@@ -627,35 +677,43 @@ ni_ifconfig_read_firmware(xml_document_array_t *array, const char *type,
 	 * Firmware is expected to provide a more exact origin
 	 * than we can set here, just read it from the nodes.
 	 */
-	rnode = xml_document_root(config_doc);
-	for (cnode = rnode->children; cnode; cnode = next) {
-		xml_document_t *doc;
-		xml_node_t *node;
+	for (i = 0; i < docs.count; ++i) {
+		xml_document_t *doc = docs.data[i];
 
-		next = cnode->next;
-		doc = xml_document_new();
-		node = xml_document_root(doc);
+		if (xml_document_is_empty(doc))
+			continue;
 
-		if (!ni_ifconfig_metadata_get_from_node(&conf, rnode))
-			ni_ifconfig_format_origin(&conf.origin, type, path);
+		rnode = xml_document_root(doc);
+		for (cnode = rnode->children; cnode; cnode = next) {
+			xml_document_t *config;
+			xml_node_t *node;
 
-		xml_node_reparent(node, cnode);
-		xml_node_location_relocate(node, conf.origin);
+			next = cnode->next;
+			if (!(config = xml_document_new()))
+				continue;
+			node = xml_document_root(config);
 
-		ni_ifconfig_metadata_clear(node);
-		if (!raw)
-			ni_ifconfig_metadata_add_to_node(node, &conf);
+			if (!ni_ifconfig_metadata_get_from_node(&conf, rnode))
+				ni_ifconfig_format_origin(&conf.origin, type, path);
 
-		if (ni_ifconfig_validate_adding_doc(doc, check_prio)) {
-			ni_debug_ifconfig("%s: %s", __func__, xml_node_location(node));
-			xml_document_array_append(array, doc);
-		} else {
-			xml_document_free(doc);
+			xml_node_reparent(node, cnode);
+			xml_node_location_relocate(node, conf.origin);
+
+			ni_ifconfig_metadata_clear(node);
+			if (!raw)
+				ni_ifconfig_metadata_add_to_node(node, &conf);
+
+			if (ni_ifconfig_validate_adding_doc(config, check_prio)) {
+				ni_debug_ifconfig("%s: %s", __func__, xml_node_location(node));
+				if (xml_document_expand(array, config))
+					continue;
+			}
+			xml_document_free(config);
 		}
 	}
 
 	ni_client_state_config_reset(&conf);
-	xml_document_free(config_doc);
+	xml_document_array_destroy(&docs);
 
 	return TRUE;
 }
@@ -680,7 +738,11 @@ ni_ifconfig_metadata_add_to_node(xml_node_t *root, ni_client_state_config_t *con
 	if (!root || !root->children || !conf || ni_string_empty(conf->origin))
 		return;
 
-	while ((ifnode = xml_node_get_next_child(root, root->children->name, ifnode))) {
+	if (ni_string_empty(root->name))
+		ifnode = root->children;
+	else
+		ifnode = root;
+	for ( ; ifnode; ifnode = ifnode->next) {
 		xml_node_add_attr(ifnode,
 				NI_CLIENT_STATE_XML_CONFIG_ORIGIN_NODE,
 					conf->origin);
@@ -707,7 +769,11 @@ ni_ifconfig_metadata_get_from_node(ni_client_state_config_t *conf, xml_node_t *r
 		return FALSE;
 
 	ni_client_state_config_reset(conf);
-	while ((ifnode = xml_node_get_next_child(root, root->children->name, ifnode))) {
+	if (ni_string_empty(root->name))
+		ifnode = root->children;
+	else
+		ifnode = root;
+	for ( ; ifnode; ifnode = ifnode->next) {
 		/* only first   node with proper meta data attributes is processed */
 		const char *origin;
 		const char *str;
@@ -738,10 +804,13 @@ ni_ifconfig_metadata_clear(xml_node_t *root)
 	if (!root || !root->children)
 		return;
 
-	while ((ifnode = xml_node_get_next_child(root, root->children->name, ifnode))) {
+	if (ni_string_empty(root->name))
+		ifnode = root->children;
+	else
+		ifnode = root;
+	for ( ; ifnode; ifnode = ifnode->next) {
 		xml_node_del_attr(ifnode, NI_CLIENT_STATE_XML_CONFIG_ORIGIN_NODE);
 		xml_node_del_attr(ifnode, NI_CLIENT_STATE_XML_CONFIG_UUID_NODE);
 		xml_node_del_attr(ifnode, NI_CLIENT_STATE_XML_CONFIG_OWNER_NODE);
 	}
 }
-
