@@ -16,6 +16,9 @@
 #include <limits.h>
 #include <errno.h>
 #include <net/if_arp.h>
+#ifdef HAVE_SYSTEMD_SD_DAEMON_H
+#include <systemd/sd-daemon.h>
+#endif
 
 #include <wicked/types.h>
 #include <wicked/logging.h>
@@ -482,8 +485,14 @@ dhcp4_supplicant(void)
 	if (opt_recover_state)
 		dhcp4_recover_state(opt_state_file);
 
+#ifdef HAVE_SYSTEMD_SD_DAEMON_H
+	if (opt_systemd) {
+		sd_notify(0, "READY=1");
+	}
+#endif
+
 	while (!ni_caught_terminal_signal()) {
-		long timeout;
+		ni_timeout_t timeout;
 
 		do {
 			timeout = ni_timer_next_timeout();
@@ -566,7 +575,8 @@ dhcp4_protocol_event(enum ni_dhcp4_event ev, const ni_dhcp4_device_t *dev, ni_ad
 	ni_dbus_variant_t *var;
 	int argc = 0;
 
-	ni_debug_dhcp("%s(ev=%u, dev=%d, uuid=%s)", __func__, ev, dev->link.ifindex,
+	ni_debug_dhcp("%s(ev=%s, dev=%s[%u], uuid=%s)", __func__,
+			ni_dhcp4_event_name(ev), dev->ifname, dev->link.ifindex,
 			dev->config? ni_uuid_print(&dev->config->uuid) : "<none>");
 
 	dev_object = ni_dbus_server_find_object_by_handle(dhcp4_dbus_server, dev);
