@@ -130,10 +130,11 @@ struct ni_capture {
 	} retrans;
 
 	void *			user_data;
+	char *			desc;
 };
 
 static int		ni_capture_set_filter(ni_capture_t *, const ni_capture_protinfo_t *);
-static ssize_t		__ni_capture_send(const ni_capture_t *, const ni_buffer_t *);
+static ssize_t		ni_capture_send_buf(const ni_capture_t *, const ni_buffer_t *);
 
 static uint32_t
 checksum_partial(uint32_t sum, const void *data, uint16_t len)
@@ -327,7 +328,7 @@ ni_capture_inspect_udp_header(void *data, size_t bytes, size_t *payload_len,
 void
 ni_capture_arm_retransmit(ni_capture_t *capture)
 {
-	ni_timeout_arm(&capture->retrans.deadline, &capture->retrans.timeout);
+	ni_timeout_arm_sec(&capture->retrans.deadline, &capture->retrans.timeout);
 }
 
 void
@@ -343,7 +344,7 @@ ni_capture_force_retransmit(ni_capture_t *capture, unsigned int delay)
 	if (timerisset(&capture->retrans.deadline)) {
 		struct timeval *deadline = &capture->retrans.deadline;
 
-		gettimeofday(deadline, NULL);
+		ni_timer_get_time(deadline);
 		deadline->tv_sec += delay;
 	}
 }
@@ -354,12 +355,12 @@ ni_capture_force_retransmit(ni_capture_t *capture, unsigned int delay)
 static void
 ni_capture_retransmit(ni_capture_t *capture)
 {
-	int rv;
-
-	ni_debug_socket("%s: retransmit request", capture->ifname);
+	ni_debug_socket("%s: %s%sretransmit request", capture->ifname,
+			capture->desc ?: "", capture->desc ? " " : "");
 
 	if (capture->retrans.buffer == NULL) {
-		ni_error("ni_capture_retransmit: no message!?");
+		ni_error("%s: %s%sni_capture_retransmit: no message!?",
+			capture->ifname, capture->desc ?: "", capture->desc ? " " : "");
 		ni_capture_disarm_retransmit(capture);
 		return;
 	}
@@ -370,12 +371,10 @@ ni_capture_retransmit(ni_capture_t *capture)
 	if (capture->retrans.timeout.timeout_callback)
 		capture->retrans.timeout.timeout_callback(capture->retrans.timeout.timeout_data);
 
-	rv = __ni_capture_send(capture, capture->retrans.buffer);
+	ni_capture_send_buf(capture, capture->retrans.buffer);
 
 	/* We don't care whether sending failed or not. Quite possibly
 	 * it's a temporary condition, so continue */
-	if (rv < 0)
-		ni_warn("%s: sending message failed", capture->ifname);
 	ni_capture_arm_retransmit(capture);
 }
 
@@ -386,7 +385,7 @@ ni_capture_retransmit(ni_capture_t *capture)
  * callbacks nested in callbacks...
  */
 static int
-__ni_capture_socket_get_timeout(const ni_socket_t *sock, struct timeval *tv)
+ni_capture_socket_get_timeout(const ni_socket_t *sock, struct timeval *tv)
 {
 	ni_capture_t *capture;
 
@@ -402,7 +401,7 @@ __ni_capture_socket_get_timeout(const ni_socket_t *sock, struct timeval *tv)
 }
 
 static void
-__ni_capture_socket_check_timeout(ni_socket_t *sock, const struct timeval *now)
+ni_capture_socket_check_timeout(ni_socket_t *sock, const struct timeval *now)
 {
 	ni_capture_t *capture;
 
@@ -418,8 +417,8 @@ __ni_capture_socket_check_timeout(ni_socket_t *sock, const struct timeval *now)
 /*
  * Capture receive handling
  */
-int
-__ni_capture_recv(int fd, void *buf, size_t len, ni_bool_t *partial_csum, ni_sockaddr_t *from)
+static int
+ni_capture_recv_raw(int fd, void *buf, size_t len, ni_bool_t *partial_csum, ni_sockaddr_t *from)
 {
 #if defined(PACKET_AUXDATA)
 	/* use 2 times bigger buffer to catch possible additions... */
@@ -495,20 +494,21 @@ ni_capture_from_hwaddr_print(const ni_sockaddr_t *from)
 }
 
 int
-ni_capture_recv(ni_capture_t *capture, ni_buffer_t *bp, ni_sockaddr_t *from, const char *hint)
+ni_capture_recv(ni_capture_t *capture, ni_buffer_t *bp, ni_sockaddr_t *from)
 {
 	void *payload;
 	size_t payload_len;
 	ssize_t bytes;
 	ni_bool_t partial_checksum = FALSE;
 	const char *lladdr;
+	const char *hint = capture->desc;
 
-	bytes = __ni_capture_recv(capture->sock->__fd, capture->buffer,
+	bytes = ni_capture_recv_raw(capture->sock->__fd, capture->buffer,
 				  capture->mtu, &partial_checksum, from);
 
 	if (bytes < 0) {
 		ni_error("%s: %s cannot read %s%spacket from socket: %m",
-				capture->ifname, __FUNCTION__,
+				capture->ifname, __func__,
 				hint ? hint : "", hint ? " " : "");
 		return -1;
 	}
@@ -540,7 +540,7 @@ ni_capture_recv(ni_capture_t *capture, ni_buffer_t *bp, ni_sockaddr_t *from, con
 
 	default:
 		ni_error("%s: %s cannot handle ethertype %u", capture->ifname,
-				__FUNCTION__, capture->protocol);
+				__func__, capture->protocol);
 		return -1;
 	}
 
@@ -587,25 +587,36 @@ ni_capture_devinfo_init(ni_capture_devinfo_t *devinfo, const char *ifname, const
 	devinfo->mtu = link->mtu ? link->mtu : MTU_MAX;
 	devinfo->hwaddr = link->hwaddr;
 
-	if (devinfo->hwaddr.len == 0) {
-		ni_error("%s: empty MAC address, cannot do packet level networking yet",
-			ifname);
-		return -1;
-	}
-	if (devinfo->hwaddr.type == ARPHRD_VOID) {
-		ni_error("%s: void arp type, cannot do packet level networking yet",
-			ifname);
-		return -1;
-	}
-
-	if (devinfo->hwaddr.type == ARPHRD_NONE) {
-		ni_warn("%s: no arp type, trying to use ether for capturing", ifname);
-		devinfo->hwaddr.type = ARPHRD_ETHER;
-	}
-
 	return 0;
 }
 
+ni_bool_t
+ni_capture_devinfo_copy(ni_capture_devinfo_t *dst, const ni_capture_devinfo_t *src)
+{
+	if (!dst || !src)
+		return FALSE;
+
+	if (!ni_string_dup(&dst->ifname, src->ifname))
+		return FALSE;
+
+	dst->iftype = src->iftype;
+	dst->ifindex = src->ifindex;
+	dst->mtu = src->mtu;
+	dst->hwaddr = src->hwaddr;
+
+	return TRUE;
+}
+
+void
+ni_capture_devinfo_destroy(ni_capture_devinfo_t *devinfo)
+{
+	if (devinfo) {
+		ni_string_free(&devinfo->ifname);
+		memset(devinfo, 0, sizeof(*devinfo));
+		ni_link_address_init(&devinfo->hwaddr);
+	}
+}
+
 int
 ni_capture_devinfo_refresh(ni_capture_devinfo_t *devinfo, const char *ifname, const ni_linkinfo_t *link)
 {
@@ -628,17 +639,6 @@ ni_capture_devinfo_refresh(ni_capture_devinfo_t *devinfo, const char *ifname, co
 		return -1;
 	}
 
-	if (devinfo->hwaddr.len == 0) {
-		ni_error("%s: empty MAC address, cannot do packet level networking yet",
-			ifname);
-		return -1;
-	}
-	if (devinfo->hwaddr.type == ARPHRD_VOID) {
-		ni_error("%s: void arp type, cannot do packet level networking yet",
-			ifname);
-		return -1;
-	}
-
 	return 0;
 }
 
@@ -650,7 +650,7 @@ ni_capture_devinfo_refresh(ni_capture_devinfo_t *devinfo, const char *ifname, co
  * some point.
  */
 static void
-__ni_capture_enable_packet_auxdata(int fd)
+ni_capture_enable_packet_auxdata(int fd)
 {
 #if defined(PACKET_AUXDATA)
 	int on = 1;
@@ -664,7 +664,7 @@ __ni_capture_enable_packet_auxdata(int fd)
 }
 
 static void
-__ni_capture_init_once(void)
+ni_capture_init_once(void)
 {
 	static ni_bool_t done = FALSE;
 
@@ -677,7 +677,8 @@ __ni_capture_init_once(void)
 }
 
 ni_capture_t *
-ni_capture_open(const ni_capture_devinfo_t *devinfo, const ni_capture_protinfo_t *protinfo, void (*receive)(ni_socket_t *))
+ni_capture_open(const ni_capture_devinfo_t *devinfo, const ni_capture_protinfo_t *protinfo,
+		void (*receive)(ni_socket_t *), const char* desc)
 {
 	ni_packetaddr_t	addr;
 	ni_capture_t *capture = NULL;
@@ -696,13 +697,13 @@ ni_capture_open(const ni_capture_devinfo_t *devinfo, const ni_capture_protinfo_t
 	/* Destination address defaults to broadcast */
 	destaddr = protinfo->eth_destaddr;
 
-	if (destaddr.len == 0
+	if (destaddr.len == 0 && ni_link_address_length(devinfo->hwaddr.type) > 0
 	 && ni_link_address_get_broadcast(devinfo->hwaddr.type, &destaddr) < 0) {
 		ni_error("cannot get broadcast address for %s (bad iftype)", devinfo->ifname);
 		return NULL;
 	}
 
-	__ni_capture_init_once();
+	ni_capture_init_once();
 
 	if ((fd = socket (PF_PACKET, SOCK_DGRAM, htons(protinfo->eth_protocol))) < 0) {
 		ni_error("socket: %m");
@@ -737,7 +738,7 @@ ni_capture_open(const ni_capture_devinfo_t *devinfo, const ni_capture_protinfo_t
 		goto failed;
 	}
 
-	__ni_capture_enable_packet_auxdata(fd);
+	ni_capture_enable_packet_auxdata(fd);
 
 	capture->mtu = devinfo->mtu;
 	if (capture->mtu == 0)
@@ -745,9 +746,10 @@ ni_capture_open(const ni_capture_devinfo_t *devinfo, const ni_capture_protinfo_t
 	capture->buffer = xmalloc(capture->mtu);
 
 	capture->sock->receive = receive;
-	capture->sock->get_timeout = __ni_capture_socket_get_timeout;
-	capture->sock->check_timeout = __ni_capture_socket_check_timeout;
+	capture->sock->get_timeout = ni_capture_socket_get_timeout;
+	capture->sock->check_timeout = ni_capture_socket_check_timeout;
 	capture->sock->user_data = capture;
+	ni_string_dup(&capture->desc, desc);
 	ni_socket_activate(capture->sock);
 	return capture;
 
@@ -776,7 +778,8 @@ ni_capture_set_filter(ni_capture_t *cap, const ni_capture_protinfo_t *protinfo)
 
 	case ETHERTYPE_IP:
 		if (protinfo->ip_protocol != IPPROTO_UDP && protinfo->ip_protocol != IPPROTO_TCP) {
-			ni_error("cannot build capture filter for IP proto %d, port %d: not supported",
+			ni_error("%s: %s%scannot build capture filter for IP proto %d, port %d: not supported",
+					cap->ifname, cap->desc ?: "", cap->desc ? " " : "",
 					protinfo->ip_protocol, protinfo->ip_port);
 			return -1;
 		}
@@ -789,32 +792,35 @@ ni_capture_set_filter(ni_capture_t *cap, const ni_capture_protinfo_t *protinfo)
 		break;
 
 	default:
-		ni_error("cannot build capture filter for ether type 0x%04x: not supported", protinfo->eth_protocol);
+		ni_error("%s: %s%scannot build capture filter for ether type 0x%04x: not supported",
+				cap->ifname, cap->desc ?: "", cap->desc ? " " : "",
+				protinfo->eth_protocol);
 		return -1;
 	}
 
 	if (setsockopt(cap->sock->__fd, SOL_SOCKET, SO_ATTACH_FILTER, &pf, sizeof(pf)) < 0) {
-		ni_error("SO_ATTACH_FILTER: %m");
+		ni_error("%s: %s%sSO_ATTACH_FILTER: %m", cap->ifname, cap->desc ?: "", cap->desc ? " " : "");
 		return -1;
 	}
 
 	return 0;
 }
 
-ssize_t
-__ni_capture_send(const ni_capture_t *capture, const ni_buffer_t *buf)
+static ssize_t
+ni_capture_send_buf(const ni_capture_t *capture, const ni_buffer_t *buf)
 {
 	ssize_t rv;
 
 	if (capture == NULL) {
-		ni_error("%s: no capture handle", __FUNCTION__);
+		ni_error("%s: no capture handle", __func__);
 		return -1;
 	}
 
 	rv = sendto(capture->sock->__fd, ni_buffer_head(buf), ni_buffer_count(buf), 0,
 			&capture->addr.sa, sizeof(capture->addr));
 	if (rv < 0)
-		ni_error("unable to send dhcp packet: %m");
+		ni_debug_socket("%s: unable to send %s%spacket: %m", capture->ifname,
+				capture->desc ?: "", capture->desc ? " " : "");
 
 	return rv;
 }
@@ -823,8 +829,10 @@ ssize_t
 ni_capture_send(ni_capture_t *capture, const ni_buffer_t *buf, const ni_timeout_param_t *tmo)
 {
 	ssize_t rv;
+	int err;
 
-	rv = __ni_capture_send(capture, buf);
+	rv = ni_capture_send_buf(capture, buf);
+	err = errno;
 	if (tmo) {
 		capture->retrans.buffer = buf;
 		capture->retrans.timeout = *tmo;
@@ -832,6 +840,7 @@ ni_capture_send(ni_capture_t *capture, const ni_buffer_t *buf, const ni_timeout_
 	} else {
 		ni_capture_disarm_retransmit(capture);
 	}
+	errno = err;
 	return rv;
 }
 
@@ -845,6 +854,7 @@ ni_capture_free(ni_capture_t *capture)
 	if (capture->buffer)
 		free(capture->buffer);
 	ni_string_free(&capture->ifname);
+	ni_string_free(&capture->desc);
 	free(capture);
 }
 
