@@ -32,10 +32,11 @@
 #include "model.h"
 #include "debug.h"
 #include "dhcp.h"
+#include "dhcp6/options.h"
 
 static dbus_bool_t		__ni_objectmodel_callback_info_to_dict(const ni_objectmodel_callback_info_t *, ni_dbus_variant_t *);
-static dbus_bool_t		__ni_objectmodel_route_to_dict(const ni_route_t *, ni_dbus_variant_t *);
-static ni_route_t *		__ni_objectmodel_route_from_dict(ni_route_table_t **, const ni_dbus_variant_t *);
+static dbus_bool_t		ni_objectmodel_route_to_dict(const ni_route_t *, ni_dbus_variant_t *);
+static dbus_bool_t		ni_objectmodel_route_from_dict(ni_route_t *, const ni_dbus_variant_t *);
 static dbus_bool_t		ni_objectmodel_rule_to_dict(const ni_rule_t *, ni_dbus_variant_t *);
 static dbus_bool_t		ni_objectmodel_rule_from_dict(ni_rule_t *, const ni_dbus_variant_t *);
 
@@ -615,7 +616,7 @@ __ni_objectmodel_address_from_dict(ni_address_t **list, const ni_dbus_variant_t
 		const ni_dbus_variant_t *var;
 		const char *label;
 
-		ap = ni_address_new(local_addr.ss_family, prefixlen, &local_addr, NULL);
+		ap = ni_address_create(local_addr.ss_family, prefixlen, &local_addr, NULL);
 		if (!ap)
 			return NULL;
 
@@ -678,7 +679,7 @@ __ni_objectmodel_address_from_dict(ni_address_t **list, const ni_dbus_variant_t
  *  </array>
  */
 dbus_bool_t
-__ni_objectmodel_get_route_list(ni_route_table_t *list,
+__ni_objectmodel_get_route_list(ni_route_table_t *list, unsigned int family,
 				ni_dbus_variant_t *result,
 				DBusError *error)
 {
@@ -694,6 +695,8 @@ __ni_objectmodel_get_route_list(ni_route_table_t *list,
 			if ((rp = tab->routes.data[i]) == NULL)
 				continue;
 
+			if (family != AF_UNSPEC && family != rp->family)
+				continue;
 			if (rp->family != rp->destination.ss_family)
 				continue;
 
@@ -702,7 +705,7 @@ __ni_objectmodel_get_route_list(ni_route_table_t *list,
 				return FALSE;
 			ni_dbus_variant_init_dict(dict);
 
-			rv = __ni_objectmodel_route_to_dict(rp, dict);
+			rv = ni_objectmodel_route_to_dict(rp, dict);
 		}
 	}
 
@@ -746,10 +749,11 @@ __ni_objectmodel_get_rule_list(ni_rule_array_t *rules, unsigned int family,
  * Build a route list from a dbus dict
  */
 dbus_bool_t
-__ni_objectmodel_set_route_list(ni_route_table_t **list,
+__ni_objectmodel_set_route_list(ni_route_table_t **list, unsigned int family,
 				const ni_dbus_variant_t *argument,
 				DBusError *error)
 {
+	ni_route_t *route;
 	unsigned int i;
 
 	if (!ni_dbus_variant_is_dict_array(argument)) {
@@ -763,7 +767,16 @@ __ni_objectmodel_set_route_list(ni_route_table_t **list,
 	for (i = 0; i < argument->array.len; ++i) {
 		ni_dbus_variant_t *dict = &argument->variant_array_value[i];
 
-		(void) __ni_objectmodel_route_from_dict(list, dict);
+		if (!(route = ni_route_new())) {
+			ni_error("%s: unable to allocate route structure", __func__);
+			return FALSE;
+		}
+
+		route->family = family;
+		if (!ni_objectmodel_route_from_dict(route, dict))
+			ni_route_free(route);
+		else if (!ni_route_tables_add_route(list, route))
+			ni_route_free(route);
 	}
 	return TRUE;
 }
@@ -821,7 +834,7 @@ __ni_objectmodel_set_rule_list(ni_rule_array_t **rules, unsigned int family,
  *   </dict>
  */
 dbus_bool_t
-__ni_objectmodel_get_route_dict(ni_route_table_t *list,
+__ni_objectmodel_get_route_dict(ni_route_table_t *list, unsigned int family,
 				ni_dbus_variant_t *result,
 				DBusError *error)
 {
@@ -837,6 +850,8 @@ __ni_objectmodel_get_route_dict(ni_route_table_t *list,
 			if ((rp = tab->routes.data[i]) == NULL)
 				continue;
 
+			if (family != AF_UNSPEC && family != rp->family)
+				continue;
 			if (rp->family != rp->destination.ss_family)
 				continue;
 
@@ -844,7 +859,7 @@ __ni_objectmodel_get_route_dict(ni_route_table_t *list,
 			dict = ni_dbus_dict_add(result, "route");
 			ni_dbus_variant_init_dict(dict);
 
-			rv = __ni_objectmodel_route_to_dict(rp, dict);
+			rv = ni_objectmodel_route_to_dict(rp, dict);
 		}
 	}
 
@@ -852,11 +867,11 @@ __ni_objectmodel_get_route_dict(ni_route_table_t *list,
 }
 
 dbus_bool_t
-__ni_objectmodel_set_route_dict(ni_route_table_t **list,
-				const ni_dbus_variant_t *dict,
-				DBusError *error)
+__ni_objectmodel_set_route_dict(ni_route_table_t **list, unsigned int family,
+				const ni_dbus_variant_t *dict, DBusError *error)
 {
 	ni_dbus_variant_t *var;
+	ni_route_t *route;
 
 	if (!list || !ni_dbus_variant_is_dict(dict)) {
 		if (error) {
@@ -872,7 +887,17 @@ __ni_objectmodel_set_route_dict(ni_route_table_t **list,
 	while ((var = ni_dbus_dict_get_next(dict, "route", var)) != NULL) {
 		if (!ni_dbus_variant_is_dict(var))
 			return FALSE;
-		(void) __ni_objectmodel_route_from_dict(list, var);
+
+		if (!(route = ni_route_new())) {
+			ni_error("%s: unable to allocate route structure", __func__);
+			return FALSE;
+		}
+
+		route->family = family;
+		if (!ni_objectmodel_route_from_dict(route, var))
+			ni_route_free(route);
+		else if (!ni_route_tables_add_route(list, route))
+			ni_route_free(route);
 	}
 	return TRUE;
 }
@@ -951,7 +976,7 @@ __ni_objectmodel_set_rule_dict(ni_rule_array_t **rules, unsigned int family,
  * Common functions to represent an assigned route as a dict
  */
 static dbus_bool_t
-__ni_objectmodel_route_to_dict(const ni_route_t *rp, ni_dbus_variant_t *dict)
+ni_objectmodel_route_to_dict(const ni_route_t *rp, ni_dbus_variant_t *dict)
 {
 	const ni_route_nexthop_t *nh;
 	ni_dbus_variant_t *child;
@@ -1057,7 +1082,7 @@ __ni_objectmodel_route_nexthop_from_dict(ni_route_nexthop_t *nh, const ni_dbus_v
 	 */
 	if (ni_dbus_dict_get(nhdict, "gateway")) {
 		if (!__ni_objectmodel_dict_get_sockaddr(nhdict, "gateway", &nh->gateway)) {
-			ni_debug_dbus("%s: invalid route hop gateway %u", __func__, value);
+			ni_debug_dbus("%s: invalid route hop gateway", __func__);
 			return FALSE;
 		}
 	}
@@ -1122,18 +1147,17 @@ __ni_objectmodel_route_kern_from_dict(ni_route_t *rp, const ni_dbus_variant_t *r
 	return TRUE;
 }
 
-static ni_route_t *
-__ni_objectmodel_route_from_dict(ni_route_table_t **list, const ni_dbus_variant_t *dict)
+static dbus_bool_t
+ni_objectmodel_route_from_dict(ni_route_t *rp, const ni_dbus_variant_t *dict)
 {
 	ni_stringbuf_t buf = NI_STRINGBUF_INIT_DYNAMIC;
 	const ni_dbus_variant_t *nhdict, *child;
-	ni_route_t *rp = NULL;
 	uint32_t value;
 	ni_bool_t scope_ok = FALSE;
 	ni_bool_t table_ok = FALSE;
 
-	if (!(rp = ni_route_new()))
-		goto failure;
+	if (!rp || !dict)
+		return FALSE;
 
 	rp->type = RTN_UNICAST;
 	rp->table = RT_TABLE_MAIN;
@@ -1181,24 +1205,29 @@ __ni_objectmodel_route_from_dict(ni_route_table_t **list, const ni_dbus_variant_
 
 	if (!__ni_objectmodel_dict_get_sockaddr_prefix(dict, "destination",
 				&rp->destination, &rp->prefixlen)) {
-		/* omitted destination just means it is a default route */
-		rp->destination.ss_family = rp->nh.gateway.ss_family;
+		/*
+		 * Omitted/missing destination means it is a default route.
+		 * Try rp->family (pre-)initialized family from selector as
+		 * the lease->family, otherwise we require destination.
+		 */
+		rp->destination.ss_family = rp->family;
 		rp->prefixlen = 0;
-	}
+	} else if (rp->family == AF_UNSPEC)
+		rp->family = rp->destination.ss_family;
 
-	/*
-	 * Hmm... check device default route / default route without a gateway?
-	 */
 	if (rp->destination.ss_family == AF_UNSPEC) {
 		ni_debug_dbus("%s: unknown route destination family", __func__);
 		goto failure;
-	} else
+	}
+	if (rp->family != rp->destination.ss_family) {
+		ni_debug_dbus("%s: unknown route family and destination family mix", __func__);
+		goto failure;
+	}
 	if (rp->nh.gateway.ss_family != AF_UNSPEC &&
 	    rp->nh.gateway.ss_family != rp->destination.ss_family) {
 		ni_debug_dbus("%s: unknown route with destination/gateway family mix", __func__);
 		goto failure;
 	}
-	rp->family = rp->destination.ss_family;
 
 	__ni_objectmodel_dict_get_sockaddr(dict, "pref-source", &rp->pref_src);
 	if (ni_dbus_dict_get_uint32(dict, "priority", &value))
@@ -1248,17 +1277,13 @@ __ni_objectmodel_route_from_dict(ni_route_table_t **list, const ni_dbus_variant_
 			rp->owner = NI_ADDRCONF_NONE;
 	}
 
-	if (ni_route_tables_add_route(list, rp))
-		return rp;
+	return TRUE;
 
 failure:
 	ni_debug_dbus("%s: Cannot get complete route from dbus dict (%s)",
 			__func__, ni_route_print(&buf, rp));
 	ni_stringbuf_destroy(&buf);
-	if (rp) {
-		ni_route_free(rp);
-	}
-	return NULL;
+	return FALSE;
 }
 
 static dbus_bool_t
@@ -1657,6 +1682,126 @@ __ni_objectmodel_get_addrconf_dhcp4_dict(const struct ni_addrconf_lease_dhcp4 *d
 	__ni_objectmodel_get_addrconf_dhcp_opts_dict(dhcp4->options, dict, 1, 65535);
 }
 
+static void
+ni_objectmodel_get_addrconf_dhcp6_ia_entry_dict(ni_dbus_variant_t *dict, const ni_dhcp6_ia_addr_t *iadr, const char *type)
+{
+	ni_sockaddr_t addr;
+
+	if (!dict || !iadr || !type)
+		return;
+
+	switch (iadr->type) {
+	case NI_DHCP6_OPTION_IA_ADDRESS:
+		ni_dbus_dict_add_string(dict, "type", type);
+		ni_sockaddr_set_ipv6(&addr, iadr->addr, 0);
+		__ni_objectmodel_dict_add_sockaddr(dict, "address", &addr);
+		ni_dbus_dict_add_uint32(dict, "preferred-lft", iadr->preferred_lft);
+		ni_dbus_dict_add_uint32(dict, "valid-lft", iadr->valid_lft);
+		break;
+	case NI_DHCP6_OPTION_IA_PREFIX:
+		ni_dbus_dict_add_string(dict, "type", type);
+		ni_sockaddr_set_ipv6(&addr, iadr->addr, 0);
+		__ni_objectmodel_dict_add_sockaddr_prefix(dict, "prefix", &addr, iadr->plen);
+		ni_dbus_dict_add_uint32(dict, "preferred-lft", iadr->preferred_lft);
+		ni_dbus_dict_add_uint32(dict, "valid-lft", iadr->valid_lft);
+		if (iadr->excl && !IN6_IS_ADDR_UNSPECIFIED(&iadr->excl->addr) &&
+				iadr->excl->plen && iadr->plen <= iadr->excl->plen) {
+			ni_sockaddr_set_ipv6(&addr, iadr->excl->addr, 0);
+			__ni_objectmodel_dict_add_sockaddr_prefix(dict, "exclude", &addr, iadr->excl->plen);
+		}
+		break;
+	default:
+		break;
+	}
+}
+
+static void
+ni_objectmodel_get_addrconf_dhcp6_ia_entry_array(ni_dbus_variant_t *dict, const ni_dhcp6_ia_addr_t *entries)
+{
+	ni_dbus_variant_t *array = NULL;
+	ni_dbus_variant_t *edict = NULL;
+	const ni_dhcp6_ia_addr_t *iadr;
+	const char *type;
+
+	if (!dict || !entries)
+		return;
+
+	for (iadr = entries; iadr; iadr = iadr->next) {
+		if (iadr->status.code != NI_DHCP6_STATUS_SUCCESS)
+			continue;
+
+		if (IN6_IS_ADDR_UNSPECIFIED(&iadr->addr))
+			continue;
+
+		if (!(type = ni_dhcp6_option_name(iadr->type)))
+			continue;
+
+		if (!array) {
+			if (!(array = ni_dbus_dict_add(dict, "entries")))
+				continue;
+			ni_dbus_dict_array_init(array);
+		}
+		if (!(edict = ni_dbus_dict_array_add(array)))
+			continue;
+
+		ni_objectmodel_get_addrconf_dhcp6_ia_entry_dict(edict, iadr, type);
+	}
+
+}
+
+static void
+ni_objectmodel_get_addrconf_dhcp6_ia_dict(ni_dbus_variant_t *dict, const ni_dhcp6_ia_t *ia, const char *type)
+{
+	if (!dict || !ia || !type)
+		return;
+
+	ni_dbus_dict_add_string(dict, "type", type);
+	ni_dbus_dict_add_uint32(dict, "iaid", ia->iaid);
+	ni_dbus_dict_add_int64(dict,  "acquired", ia->acquired.tv_sec);
+	switch (ia->type) {
+	case NI_DHCP6_OPTION_IA_NA:
+	case NI_DHCP6_OPTION_IA_PD:
+		ni_dbus_dict_add_uint32(dict, "renewal-time", ia->renewal_time);
+		ni_dbus_dict_add_uint32(dict, "rebind-time", ia->rebind_time);
+		break;
+	case NI_DHCP6_OPTION_IA_TA:
+	default:
+		break;
+	}
+	ni_objectmodel_get_addrconf_dhcp6_ia_entry_array(dict, ia->addrs);
+}
+
+static void
+ni_objectmodel_get_addrconf_dhcp6_ia_array(ni_dbus_variant_t *dict, const ni_dhcp6_ia_t *entries)
+{
+	ni_dbus_variant_t *array = NULL;
+	ni_dbus_variant_t *edict = NULL;
+	const ni_dhcp6_ia_t *ia;
+	const char *type;
+
+	if (!dict || !entries)
+		return;
+
+	for (ia = entries; ia; ia = ia->next) {
+		if (ia->status.code != NI_DHCP6_STATUS_SUCCESS)
+			continue;
+
+		if (!(type = ni_dhcp6_option_name(ia->type)))
+			continue;
+
+		if (!array) {
+			if (!(array = ni_dbus_dict_add(dict, "ias")))
+				continue;
+			ni_dbus_dict_array_init(array);
+		}
+
+		if (!(edict = ni_dbus_dict_array_add(array)))
+			continue;
+
+		ni_objectmodel_get_addrconf_dhcp6_ia_dict(edict, ia, type);
+	}
+}
+
 static void
 __ni_objectmodel_get_addrconf_dhcp6_dict(const struct ni_addrconf_lease_dhcp6 *dhcp6,
 					ni_dbus_variant_t *dict)
@@ -1683,8 +1828,6 @@ __ni_objectmodel_get_addrconf_dhcp6_dict(const struct ni_addrconf_lease_dhcp6 *d
 		ni_dbus_dict_add_bool(dict, "rapid-commit", dhcp6->rapid_commit);
 	}
 
-	/* Hmm... status + ia_list: only if we need it */
-
 	if (dhcp6->boot_url) {
 		ni_dbus_dict_add_string(dict, "bootfile-url", dhcp6->boot_url);
 	}
@@ -1693,9 +1836,44 @@ __ni_objectmodel_get_addrconf_dhcp6_dict(const struct ni_addrconf_lease_dhcp6 *d
 							&dhcp6->boot_params);
 	}
 
+	ni_objectmodel_get_addrconf_dhcp6_ia_array(dict, dhcp6->ia_list);
 	__ni_objectmodel_get_addrconf_dhcp_opts_dict(dhcp6->options, dict, 0, 65535);
 }
 
+static dbus_bool_t
+__ni_objectmodel_get_nis_info(const ni_nis_info_t *nis, ni_dbus_variant_t *dict, DBusError *error)
+{
+	ni_dbus_variant_t *domains;
+	const ni_nis_domain_t *dom;
+	unsigned int i;
+
+	ni_dbus_variant_init_dict(dict);
+
+	if (nis->domainname || nis->default_servers.count) {
+		ni_dbus_dict_add_string(dict, "domainname", nis->domainname);
+		ni_dbus_dict_add_uint32(dict, "binding", nis->default_binding);
+		__ni_objectmodel_set_string_array(dict, "servers", &nis->default_servers);
+	}
+
+	if (nis->domains.count && (domains = ni_dbus_dict_add(dict, "domains"))) {
+
+		ni_dbus_dict_array_init(domains);
+		for (i = 0; i < nis->domains.count; ++i) {
+			if (!(dom = nis->domains.data[i]))
+				continue;
+
+			if (!(dict = ni_dbus_dict_array_add(domains)))
+				continue;
+
+			ni_dbus_variant_init_dict(dict);
+			ni_dbus_dict_add_string(dict, "domainname", dom->domainname);
+			ni_dbus_dict_add_uint32(dict, "binding", dom->binding);
+			__ni_objectmodel_set_string_array(dict, "servers", &dom->servers);
+		}
+	}
+	return TRUE;
+}
+
 dbus_bool_t
 __ni_objectmodel_get_addrconf_lease(const ni_addrconf_lease_t *lease,
 						ni_dbus_variant_t *result,
@@ -1724,7 +1902,7 @@ __ni_objectmodel_get_addrconf_lease(const ni_addrconf_lease_t *lease,
 	if (lease->routes) {
 		child = ni_dbus_dict_add(result, "routes");
 		ni_dbus_dict_array_init(child);
-		if (!__ni_objectmodel_get_route_list(lease->routes, child, error))
+		if (!__ni_objectmodel_get_route_list(lease->routes, lease->family, child, error))
 			return FALSE;
 	}
 	if (lease->rules && lease->rules->count) {
@@ -1746,17 +1924,8 @@ __ni_objectmodel_get_addrconf_lease(const ni_addrconf_lease_t *lease,
 		__ni_objectmodel_set_string_array(child, "search", &resolv->dns_search);
 	}
 
-	if (lease->nis) {
-		ni_nis_info_t *nis = lease->nis;
-
-		child = ni_dbus_dict_add(result, "nis");
-		ni_dbus_variant_init_dict(child);
-
-		if (nis->domainname)
-			ni_dbus_dict_add_string(child, "domainname", nis->domainname);
-		ni_dbus_dict_add_uint32(child, "binding", nis->default_binding);
-		__ni_objectmodel_set_string_array(child, "servers", &nis->default_servers);
-	}
+	if (lease->nis && (child = ni_dbus_dict_add(result, "nis")))
+		__ni_objectmodel_get_nis_info(lease->nis, child, error);
 
 	__ni_objectmodel_set_string_array(result, "log-servers", &lease->log_servers);
 	__ni_objectmodel_set_string_array(result, "ntp-servers", &lease->ntp_servers);
@@ -1870,6 +2039,167 @@ __ni_objectmodel_set_addrconf_dhcp4_data(struct ni_addrconf_lease_dhcp4 *dhcp4,
 	return TRUE;
 }
 
+static dbus_bool_t
+ni_objectmodel_set_addrconf_dhcp6_ia_entry_dict(ni_dhcp6_ia_addr_t *iadr, const ni_dbus_variant_t *dict)
+{
+	const ni_dbus_variant_t *var;
+	const char *type = NULL;
+	ni_sockaddr_t addr;
+	unsigned int plen;
+
+	if (!iadr || !dict || !ni_dbus_variant_is_dict(dict))
+		return FALSE;
+
+	if (!ni_dbus_dict_get_string(dict, "type", &type))
+		return FALSE;
+
+	if (ni_string_eq(type, ni_dhcp6_option_name(NI_DHCP6_OPTION_IA_ADDRESS)))
+		iadr->type = NI_DHCP6_OPTION_IA_ADDRESS;
+	else
+	if (ni_string_eq(type, ni_dhcp6_option_name(NI_DHCP6_OPTION_IA_PREFIX)))
+		iadr->type = NI_DHCP6_OPTION_IA_PREFIX;
+	else
+		return FALSE;
+
+	switch (iadr->type) {
+	case NI_DHCP6_OPTION_IA_ADDRESS:
+		if (!__ni_objectmodel_dict_get_sockaddr(dict, "address", &addr) ||
+		    !ni_sockaddr_is_ipv6_specified(&addr))
+			return FALSE;
+		iadr->addr = addr.six.sin6_addr;
+		if (!ni_dbus_dict_get_uint32(dict, "preferred-lft", &iadr->preferred_lft))
+			return FALSE;
+		if (!ni_dbus_dict_get_uint32(dict, "valid-lft", &iadr->valid_lft))
+			return FALSE;
+		break;
+	case NI_DHCP6_OPTION_IA_PREFIX:
+		if (!__ni_objectmodel_dict_get_sockaddr_prefix(dict, "prefix", &addr, &plen) ||
+				addr.ss_family != AF_INET6 || !plen || plen > 128)
+			return FALSE;
+		iadr->addr = addr.six.sin6_addr;
+		iadr->plen = plen;
+		if (!ni_dbus_dict_get_uint32(dict, "preferred-lft", &iadr->preferred_lft))
+			return FALSE;
+		if (!ni_dbus_dict_get_uint32(dict, "valid-lft", &iadr->valid_lft))
+			return FALSE;
+		if ((var = ni_dbus_dict_get(dict, "exclude"))) {
+			if (!__ni_objectmodel_get_sockaddr_prefix(var, &addr, &plen) ||
+			    !ni_sockaddr_is_ipv6_specified(&addr) || iadr->plen >= plen || plen > 128)
+				return FALSE;
+			iadr->excl = ni_dhcp6_ia_pd_excl_new(addr.six.sin6_addr, plen);
+		}
+		break;
+	default:
+		return FALSE;
+	}
+	return TRUE;
+}
+
+static dbus_bool_t
+ni_objectmodel_set_addrconf_dhcp6_ia_entry_array(ni_dhcp6_ia_addr_t **entries, const ni_dbus_variant_t *array)
+{
+	const ni_dbus_variant_t *edict;
+	ni_dhcp6_ia_addr_t *iadr;
+	unsigned int i;
+
+	if (!entries || !array || !ni_dbus_variant_is_dict_array(array))
+		return FALSE;
+
+	ni_dhcp6_ia_addr_list_destroy(entries);
+	for (i = 0; i < array->array.len; ++i) {
+		edict = &array->variant_array_value[i];
+		if (!ni_dbus_variant_is_dict(edict))
+			continue;
+
+		if (!(iadr = ni_dhcp6_ia_addr_new(0, in6addr_any, 0)))
+			continue;
+
+		if (!ni_objectmodel_set_addrconf_dhcp6_ia_entry_dict(iadr, edict) ||
+		    !ni_dhcp6_ia_addr_list_append(entries, iadr))
+			ni_dhcp6_ia_addr_free(iadr);
+	}
+	return TRUE;
+}
+
+static dbus_bool_t
+ni_objectmodel_set_addrconf_dhcp6_ia_dict(ni_dhcp6_ia_t *ia, const ni_dbus_variant_t *dict)
+{
+	const ni_dbus_variant_t *array;
+	const char *type = NULL;
+	int64_t i64;
+
+	if (!ia || !dict || !ni_dbus_variant_is_dict(dict))
+		return FALSE;
+
+	if (!ni_dbus_dict_get_string(dict, "type", &type))
+		return FALSE;
+
+	if (ni_string_eq(type, ni_dhcp6_option_name(NI_DHCP6_OPTION_IA_PD)))
+		ia->type = NI_DHCP6_OPTION_IA_PD;
+	else
+	if (ni_string_eq(type, ni_dhcp6_option_name(NI_DHCP6_OPTION_IA_NA)))
+		ia->type = NI_DHCP6_OPTION_IA_NA;
+	else
+	if (ni_string_eq(type, ni_dhcp6_option_name(NI_DHCP6_OPTION_IA_TA)))
+		ia->type = NI_DHCP6_OPTION_IA_TA;
+	else
+		return FALSE;
+
+	if (!ni_dbus_dict_get_uint32(dict, "iaid", &ia->iaid))
+		return FALSE;
+	if (ni_dbus_dict_get_int64(dict, "acquired", &i64)) {
+		ia->acquired.tv_sec = i64;
+		ia->acquired.tv_usec = 0;
+	} else {
+		return FALSE;
+	}
+
+	switch (ia->type) {
+	case NI_DHCP6_OPTION_IA_PD:
+	case NI_DHCP6_OPTION_IA_NA:
+		if (!ni_dbus_dict_get_uint32(dict, "renewal-time", &ia->renewal_time))
+			return FALSE;
+		if (!ni_dbus_dict_get_uint32(dict, "rebind-time", &ia->rebind_time))
+			return FALSE;
+		break;
+	case NI_DHCP6_OPTION_IA_TA:
+	default:
+		break;
+	}
+
+	if ((array = ni_dbus_dict_get(dict, "entries")) &&
+		!ni_objectmodel_set_addrconf_dhcp6_ia_entry_array(&ia->addrs, array))
+		return FALSE;
+	else
+		return TRUE;
+}
+
+static dbus_bool_t
+ni_objectmodel_set_addrconf_dhcp6_ia_array(ni_dhcp6_ia_t **entries, const ni_dbus_variant_t *array)
+{
+	const ni_dbus_variant_t *edict;
+	ni_dhcp6_ia_t *ia;
+	unsigned int i;
+
+	if (!entries || !array || !ni_dbus_variant_is_dict_array(array))
+		return FALSE;
+
+	ni_dhcp6_ia_list_destroy(entries);
+	for (i = 0; i < array->array.len; ++i) {
+		edict = &array->variant_array_value[i];
+		if (!ni_dbus_variant_is_dict(edict))
+			continue;
+
+		if (!(ia = ni_dhcp6_ia_new(0, 0)))
+			continue;
+
+		if (!ni_objectmodel_set_addrconf_dhcp6_ia_dict(ia, edict) ||
+		    !ni_dhcp6_ia_list_append(entries, ia))
+			ni_dhcp6_ia_free(ia);
+	}
+	return TRUE;
+}
+
 static dbus_bool_t
 __ni_objectmodel_set_addrconf_dhcp6_data(struct ni_addrconf_lease_dhcp6 *dhcp6,
 					const ni_dbus_variant_t *dict,
@@ -1916,11 +2246,74 @@ __ni_objectmodel_set_addrconf_dhcp6_data(struct ni_addrconf_lease_dhcp6 *dhcp6,
 						error, "bootfile-params"))
 		return FALSE;
 
+	if ((var = ni_dbus_dict_get(dict, "ias")) &&
+		!ni_objectmodel_set_addrconf_dhcp6_ia_array(&dhcp6->ia_list, var))
+		return FALSE;
+
 	__ni_objectmodel_set_addrconf_dhcp_opts_dict(&dhcp6->options, dict, 0, 65535);
 
 	return TRUE;
 }
 
+static dbus_bool_t
+__ni_objectmodel_set_nis_info(ni_nis_info_t **result, const ni_dbus_variant_t *dict, DBusError *error)
+{
+	const ni_dbus_variant_t *servers, *domains;
+	const char *string_value;
+	uint32_t value32;
+	ni_nis_info_t *nis;
+
+	if (!result || !dict || !ni_dbus_variant_is_dict(dict))
+		return FALSE;
+
+	*result = NULL;
+	if (!(nis = ni_nis_info_new()))
+		return FALSE;
+
+	if (__ni_objectmodel_get_domain_string(dict, "domainname", &string_value))
+		ni_string_dup(&nis->domainname, string_value);
+
+	if (ni_dbus_dict_get_uint32(dict, "binding", &value32))
+		nis->default_binding = value32;
+
+	servers = ni_dbus_dict_get(dict, "servers");
+	if (servers && !__ni_objectmodel_get_address_array(&nis->default_servers,
+			servers, error, "nis servers")) {
+		ni_nis_info_free(nis);
+		return FALSE;
+	}
+
+	domains = ni_dbus_dict_get(dict, "domains");
+	if (domains && ni_dbus_variant_is_dict_array(domains)) {
+		unsigned int i;
+
+		for (i = 0; i < domains->array.len; ++i) {
+			ni_nis_domain_t *dom;
+
+			dict = &domains->variant_array_value[i];
+			if (!ni_dbus_variant_is_dict(dict) ||
+			    !__ni_objectmodel_get_domain_string(dict, "domainname", &string_value))
+				continue;
+
+			if (ni_nis_domain_find(nis, string_value))
+				continue;
+
+			if ((dom = ni_nis_domain_new(nis, string_value))) {
+				if (ni_dbus_dict_get_uint32(dict, "binding", &value32))
+					dom->binding = value32;
+
+				if ((servers = ni_dbus_dict_get(dict, "servers"))) {
+					__ni_objectmodel_get_address_array(&dom->servers, servers,
+									error, "nis-domain servers");
+				}
+			}
+		}
+	}
+
+	*result = nis;
+	return TRUE;
+}
+
 dbus_bool_t
 __ni_objectmodel_set_addrconf_lease(ni_addrconf_lease_t *lease,
 						const ni_dbus_variant_t *argument,
@@ -1955,7 +2348,7 @@ __ni_objectmodel_set_addrconf_lease(ni_addrconf_lease_t *lease,
 		return FALSE;
 
 	if ((child = ni_dbus_dict_get(argument, "routes")) != NULL
-	 && !__ni_objectmodel_set_route_list(&lease->routes, child, error))
+	 && !__ni_objectmodel_set_route_list(&lease->routes, lease->family, child, error))
 		return FALSE;
 
 	if ((child = ni_dbus_dict_get(argument, "rules")) != NULL
@@ -1965,23 +2358,9 @@ __ni_objectmodel_set_addrconf_lease(ni_addrconf_lease_t *lease,
 	if (!__ni_objectmodel_set_resolver_dict(&lease->resolver, argument, error))
 		return FALSE;
 
-	if ((child = ni_dbus_dict_get(argument, "nis")) != NULL) {
-		ni_nis_info_t *nis = ni_nis_info_new();
-		ni_dbus_variant_t *list;
-
-		lease->nis = nis;
-		if (__ni_objectmodel_get_domain_string(child, "domainname",
-								&string_value))
-			ni_string_dup(&nis->domainname, string_value);
-
-		if (ni_dbus_dict_get_uint32(child, "binding", &value32))
-			nis->default_binding = value32;
-
-		if ((list = ni_dbus_dict_get(child, "servers")) != NULL
-		 && !__ni_objectmodel_get_address_array(&nis->default_servers, list,
-							error, "servers"))
-			return FALSE;
-	}
+	if ((child = ni_dbus_dict_get(argument, "nis"))
+	 && !__ni_objectmodel_set_nis_info(&lease->nis, child, error))
+		return FALSE;
 
 	if ((child = ni_dbus_dict_get(argument, "log-servers")) != NULL
 	 && !__ni_objectmodel_get_address_array(&lease->log_servers, child, error,
