@@ -18,6 +18,7 @@
 
 #include <wicked/netinfo.h>
 #include <wicked/logging.h>
+#include <wicked/time.h>
 #include <wicked/xml.h>
 #include "netinfo_priv.h"
 #include "appconfig.h"
@@ -29,7 +30,7 @@
 #include "duid.h"
 
 
-static unsigned int	ni_dhcp4_do_bits(const ni_config_dhcp4_t *, unsigned int);
+static unsigned int	ni_dhcp4_do_bits(const ni_dhcp4_device_t *, unsigned int);
 static const char *	ni_dhcp4_print_doflags(unsigned int);
 static void		ni_dhcp4_config_set_request_options(const char *, ni_uint_array_t *, const ni_string_array_t *);
 
@@ -51,6 +52,12 @@ ni_dhcp4_device_new(const char *ifname, const ni_linkinfo_t *link)
 	dev->users = 1;
 	dev->listen_fd = -1;
 	dev->link.ifindex = link->ifindex;
+	/*
+	 * it's either a fresh link and we have to perform dad anyway
+	 * or we just (re-)started and "may have moved to a new link",
+	 * so assume a reconnect to retrigger dad in next lease commit.
+	 */
+	dev->link.reconnect = TRUE;
 
 	if (ni_capture_devinfo_init(&dev->system, dev->ifname, link) < 0) {
 		ni_error("%s: cannot set up %s for DHCP4", __func__, ifname);
@@ -58,7 +65,7 @@ ni_dhcp4_device_new(const char *ifname, const ni_linkinfo_t *link)
 		return NULL;
 	}
 
-	dev->start_time = time(NULL);
+	ni_timer_get_time(&dev->start_time);
 	dev->fsm.state = NI_DHCP4_STATE_INIT;
 
 	/* append to end of list */
@@ -67,6 +74,55 @@ ni_dhcp4_device_new(const char *ifname, const ni_linkinfo_t *link)
 	return dev;
 }
 
+static void
+ni_dhcp4_device_free(ni_dhcp4_device_t *dev)
+{
+	ni_dhcp4_device_t **pos;
+
+	ni_assert(dev->users == 0);
+	ni_debug_dhcp("%s: Deleting dhcp4 device with index %u",
+			dev->ifname, dev->link.ifindex);
+
+	ni_dhcp4_device_drop_buffer(dev);
+	ni_dhcp4_device_drop_lease(dev);
+	ni_dhcp4_device_drop_best_offer(dev);
+	ni_dhcp4_device_stop(dev);
+
+	ni_capture_devinfo_destroy(&dev->system);
+
+	ni_string_free(&dev->ifname);
+
+	for (pos = &ni_dhcp4_active; *pos; pos = &(*pos)->next) {
+		if (*pos == dev) {
+			*pos = dev->next;
+			break;
+		}
+	}
+	free(dev);
+}
+
+/*
+ * Refcount handling
+ */
+ni_dhcp4_device_t *
+ni_dhcp4_device_get(ni_dhcp4_device_t *dev)
+{
+	ni_assert(dev->users);
+	dev->users++;
+	return dev;
+}
+
+void
+ni_dhcp4_device_put(ni_dhcp4_device_t *dev)
+{
+	ni_assert(dev->users);
+	if (--(dev->users) == 0)
+		ni_dhcp4_device_free(dev);
+}
+
+/*
+ * Lookup dhcp4 client device, netdev, linkinfo
+ */
 ni_dhcp4_device_t *
 ni_dhcp4_device_by_index(unsigned int ifindex)
 {
@@ -80,27 +136,70 @@ ni_dhcp4_device_by_index(unsigned int ifindex)
 	return NULL;
 }
 
+static ni_netdev_t *
+ni_dhcp4_device_netdev(const ni_dhcp4_device_t *dev)
+{
+	ni_netconfig_t *nc;
+	ni_netdev_t *ifp;
+
+	if (!dev || !(nc = ni_global_state_handle(0)))
+		return NULL;
+	if (!(ifp = ni_netdev_by_index(nc, dev->link.ifindex)))
+		return NULL;
+	return ifp;
+}
+
+static ni_bool_t
+ni_dhcp4_device_link_is_up(const ni_dhcp4_device_t *dev)
+{
+	const ni_netdev_t *ifp;
+
+	if (!(ifp = ni_dhcp4_device_netdev(dev)))
+		return FALSE;
+	return ni_netdev_link_is_up(ifp);
+}
+
+ni_bool_t
+ni_dhcp4_timer_arm(const ni_timer_t **timer, ni_timeout_t timeout,
+		ni_timeout_callback_t *callback, ni_dhcp4_device_t *dev)
+{
+	ni_assert(timer && callback && dev);
+
+	ni_dhcp4_timer_disarm(timer);
+	if ((*timer = ni_timer_register(timeout, callback, dev)))
+		return TRUE;
+
+	return FALSE;
+}
+
+void
+ni_dhcp4_timer_disarm(const ni_timer_t **timer)
+{
+	ni_assert(timer);
+
+	if (*timer) {
+		ni_timer_cancel(*timer);
+		*timer = NULL;
+	}
+}
+
+void
+ni_dhcp4_device_timer_disarm(ni_dhcp4_device_t *dev)
+{
+	ni_dhcp4_timer_disarm(&dev->fsm.timer);
+	ni_dhcp4_timer_disarm(&dev->timer.delay);
+	ni_dhcp4_timer_disarm(&dev->timer.defer);
+	ni_dhcp4_timer_disarm(&dev->timer.acquire);
+}
+
 static void
 ni_dhcp4_device_close(ni_dhcp4_device_t *dev)
 {
-	ni_capture_free(dev->capture);
-	dev->capture = NULL;
-
-	if (dev->listen_fd >= 0)
-		close(dev->listen_fd);
-	dev->listen_fd = -1;
-
-	if (dev->defer.timer) {
-		ni_timer_cancel(dev->defer.timer);
-		dev->defer.timer = NULL;
-	}
-	if (dev->fsm.timer) {
-		ni_warn("%s: timer active for %s", __func__, dev->ifname);
-		ni_timer_cancel(dev->fsm.timer);
-		dev->fsm.timer = NULL;
-	}
+	ni_dhcp4_device_disarm_retransmit(dev);
+	ni_dhcp4_device_timer_disarm(dev);
 
 	ni_dhcp4_device_arp_close(dev);
+	ni_dhcp4_socket_close(dev);
 }
 
 void
@@ -132,106 +231,52 @@ ni_dhcp4_device_set_request(ni_dhcp4_device_t *dev, ni_dhcp4_request_t *request)
 	dev->request = request;
 }
 
-void
-ni_dhcp4_device_free(ni_dhcp4_device_t *dev)
-{
-	ni_dhcp4_device_t **pos;
-
-	ni_assert(dev->users == 0);
-	ni_debug_dhcp("%s: Deleting dhcp4 device with index %u",
-			dev->ifname, dev->link.ifindex);
-
-	ni_dhcp4_device_drop_buffer(dev);
-	ni_dhcp4_device_drop_lease(dev);
-	ni_dhcp4_device_drop_best_offer(dev);
-	ni_dhcp4_device_close(dev);
-	ni_string_free(&dev->system.ifname);
-	ni_string_free(&dev->ifname);
-
-	/* Drop existing config and request */
-	ni_dhcp4_device_set_config(dev, NULL);
-	ni_dhcp4_device_set_request(dev, NULL);
-
-	for (pos = &ni_dhcp4_active; *pos; pos = &(*pos)->next) {
-		if (*pos == dev) {
-			*pos = dev->next;
-			break;
-		}
-	}
-	free(dev);
-}
-
-/*
- * Refcount handling
- */
-ni_dhcp4_device_t *
-ni_dhcp4_device_get(ni_dhcp4_device_t *dev)
-{
-	ni_assert(dev->users);
-	dev->users++;
-	return dev;
-}
-
-void
-ni_dhcp4_device_put(ni_dhcp4_device_t *dev)
-{
-	ni_assert(dev->users);
-	if (--(dev->users) == 0)
-		ni_dhcp4_device_free(dev);
-}
-
-
 unsigned int
 ni_dhcp4_device_uptime(const ni_dhcp4_device_t *dev, unsigned int clamp)
 {
-	unsigned int uptime;
+	struct timeval now, uptime;
 
-	uptime = time(NULL) - dev->start_time;
-	return (uptime < clamp)? uptime : clamp;
+	ni_timer_get_time(&now);
+	if (timercmp(&now, &dev->start_time, >))
+		timersub(&now, &dev->start_time, &uptime);
+	else
+		timerclear(&uptime);
+
+	if ((unsigned long)uptime.tv_sec < (unsigned long)clamp)
+		return (unsigned int)uptime.tv_sec;
+	else
+		return clamp;
 }
 
 void
 ni_dhcp4_device_set_lease(ni_dhcp4_device_t *dev, ni_addrconf_lease_t *lease)
 {
-	if (dev->lease != lease) {
-		if (dev->lease)
-			ni_addrconf_lease_free(dev->lease);
-		dev->lease = lease;
-		if (dev->config && lease)
-			lease->uuid = dev->config->uuid;
-	}
+	ni_addrconf_lease_hold(&dev->lease, lease);
+	if (dev->config && lease)
+		lease->uuid = dev->config->uuid;
 }
 
 void
 ni_dhcp4_device_drop_lease(ni_dhcp4_device_t *dev)
 {
-	ni_addrconf_lease_t *lease;
-
-	if ((lease = dev->lease) != NULL) {
-		dev->lease = NULL;
-		ni_addrconf_lease_free(lease);
-	}
+	ni_addrconf_lease_drop(&dev->lease);
 }
 
 void
-ni_dhcp4_device_set_best_offer(ni_dhcp4_device_t *dev, ni_addrconf_lease_t *lease,
+ni_dhcp4_device_set_best_offer(ni_dhcp4_device_t *dev, ni_addrconf_lease_t **lease,
 							int weight)
 {
-	if (dev->best_offer.lease && dev->best_offer.lease != lease)
-		ni_addrconf_lease_free(dev->best_offer.lease);
-	dev->best_offer.lease = lease;
+	ni_addrconf_lease_move(&dev->best_offer.lease, lease);
 	dev->best_offer.weight = weight;
-	if (dev->config && lease)
-		lease->uuid = dev->config->uuid;
+	if (dev->config && dev->best_offer.lease)
+		dev->best_offer.lease->uuid = dev->config->uuid;
 }
 
 void
 ni_dhcp4_device_drop_best_offer(ni_dhcp4_device_t *dev)
 {
 	dev->best_offer.weight = -1;
-	if (dev->best_offer.lease)
-		ni_addrconf_lease_free(dev->best_offer.lease);
-	dev->best_offer.lease = NULL;
+	ni_addrconf_lease_drop(&dev->best_offer.lease);
 }
 
 /*
@@ -240,16 +285,17 @@ ni_dhcp4_device_drop_best_offer(ni_dhcp4_device_t *dev)
 int
 ni_dhcp4_device_refresh(ni_dhcp4_device_t *dev)
 {
-	ni_netconfig_t *nih = ni_global_state_handle(0);
-	int rv;
+	ni_netconfig_t *nc = ni_global_state_handle(0);
+	ni_netdev_t *ifp;
+	int rv = -1;
 
-	if ((rv = __ni_device_refresh_link_info(nih, &dev->link)) < 0) {
-		ni_error("%s: cannot refresh interface: %s",
-				__func__, ni_strerror(rv));
+	ifp = nc ? ni_netdev_by_index(nc, dev->link.ifindex) : NULL;
+	if (!ifp || (rv = __ni_device_refresh_link_info(nc, &ifp->link)) < 0) {
+		ni_error("%s: cannot refresh interface link info", dev->ifname);
 		return rv;
 	}
 
-	return ni_capture_devinfo_refresh(&dev->system, dev->ifname, &dev->link);
+	return ni_capture_devinfo_refresh(&dev->system, dev->ifname, &ifp->link);
 }
 
 /*
@@ -269,10 +315,6 @@ ni_dhcp4_acquire(ni_dhcp4_device_t *dev, const ni_dhcp4_request_t *info)
 
 	config = xcalloc(1, sizeof(*config));
 
-	/* RFC 2131 4.1 suggests these values */
-	config->capture_retry_timeout = NI_DHCP4_RESEND_TIMEOUT_INIT;
-	config->capture_max_timeout = NI_DHCP4_RESEND_TIMEOUT_MAX;
-
 	config->dry_run = info->dry_run;
 	config->start_delay = info->start_delay;
 	config->defer_timeout = info->defer_timeout;
@@ -285,18 +327,21 @@ ni_dhcp4_acquire(ni_dhcp4_device_t *dev, const ni_dhcp4_request_t *info)
 		config->update = info->update;
 		config->update &= ni_config_addrconf_update_mask(NI_ADDRCONF_DHCP, AF_INET);
 	}
-	config->doflags = ni_dhcp4_do_bits(ni_config_dhcp4_find_device(dev->ifname), config->update);
+	config->doflags = ni_dhcp4_do_bits(dev, config->update);
 
 	config->route_priority = info->route_priority;
+	config->route_set_src = info->route_set_src;
 	config->recover_lease = info->recover_lease;
 	config->release_lease = info->release_lease;
 	config->broadcast = info->broadcast;
 
-	config->max_lease_time = ni_dhcp4_config_max_lease_time();
-	if (config->max_lease_time == 0)
-		config->max_lease_time = ~0U;
-	if (info->lease_time && info->lease_time < config->max_lease_time)
-		config->max_lease_time = info->lease_time;
+	config->max_lease_time = max_t(unsigned int,
+				ni_dhcp4_config_max_lease_time(dev->ifname),
+				NI_DHCP4_LEASE_TIME_MIN);
+	if (info->lease_time && info->lease_time < config->max_lease_time) {
+		config->max_lease_time = max_t(unsigned int,
+				info->lease_time, NI_DHCP4_LEASE_TIME_MIN);
+	}
 
 	/*
 	 * RFC 4702 section 3.1 defines, that a client sending the fqdn
@@ -319,10 +364,10 @@ ni_dhcp4_acquire(ni_dhcp4_device_t *dev, const ni_dhcp4_request_t *info)
 	if (config->fqdn.enabled == NI_TRISTATE_DEFAULT)
 		ni_tristate_set(&config->fqdn.enabled, FALSE);
 	if (config->fqdn.enabled == NI_TRISTATE_ENABLE && ni_string_empty(config->hostname))
-		config->fqdn.update = NI_DHCP4_FQDN_UPDATE_NONE;
+		config->fqdn.update = NI_DHCP_FQDN_UPDATE_NONE;
 
 	if (!ni_dhcp4_parse_client_id(&config->client_id, dev->system.hwaddr.type, info->clientid))
-		ni_dhcp4_set_config_client_id(&config->client_id, dev);
+		ni_dhcp4_set_config_client_id(&config->client_id, dev, info->create_cid);
 
 	if ((classid = info->vendor_class) == NULL)
 		classid = ni_dhcp4_config_vendor_class();
@@ -336,9 +381,10 @@ ni_dhcp4_acquire(ni_dhcp4_device_t *dev, const ni_dhcp4_request_t *info)
 
 	if (ni_log_facility(NI_TRACE_DHCP)) {
 		ni_trace("Received request:");
-		ni_trace("  acquire-timeout %u", config->acquire_timeout);
-		ni_trace("  lease-time      %u", config->max_lease_time);
-		ni_trace("  start-delay     %u", config->start_delay);
+		ni_trace("  start-delay     %s", ni_sprint_timeout(config->start_delay));
+		ni_trace("  defer-timeout   %s", ni_sprint_timeout(config->defer_timeout));
+		ni_trace("  acquire-timeout %s", ni_sprint_timeout(config->acquire_timeout));
+		ni_trace("  max lease-time  %s", ni_sprint_timeout(config->max_lease_time));
 		ni_trace("  hostname        %s", config->hostname[0]? config->hostname : "<none>");
 		if (config->fqdn.enabled == NI_TRISTATE_ENABLE) {
 			ni_trace("  fqdn            update %s, encode %s, qualify %s",
@@ -364,23 +410,15 @@ ni_dhcp4_acquire(ni_dhcp4_device_t *dev, const ni_dhcp4_request_t *info)
 
 	ni_dhcp4_device_set_config(dev, config);
 
-	if (!dev->lease && config->dry_run != NI_DHCP4_RUN_OFFER && config->recover_lease)
-		ni_dhcp4_recover_lease(dev);
-
-	if (dev->lease) {
-		if (config->client_id.len && !ni_opaque_eq(&config->client_id, &dev->lease->dhcp4.client_id)) {
-			ni_debug_dhcp("%s: lease doesn't match request", dev->ifname);
-			ni_dhcp4_device_drop_lease(dev);
-			dev->notify = 1;
-		} else {
-			/* Lease may be good */
-			dev->fsm.state = NI_DHCP4_STATE_REBOOT;
-		}
-	}
-
 	ni_note("%s: Request to acquire DHCPv4 lease with UUID %s",
 		dev->ifname, ni_uuid_print(&config->uuid));
 
+	if (config->dry_run != NI_DHCP4_RUN_OFFER &&
+	    config->recover_lease && ni_dhcp4_recover_lease(dev) == 0)
+		dev->fsm.state = NI_DHCP4_STATE_REBOOT;
+	else
+		dev->fsm.state = NI_DHCP4_STATE_INIT;
+
 	if (ni_dhcp4_device_start(dev) < 0)
 		return -1;
 	return 1;
@@ -410,7 +448,7 @@ ni_dhcp4_restart_leases(void)
  * DHCP4_DO_* masks
  */
 static unsigned int
-ni_dhcp4_do_bits(const ni_config_dhcp4_t *conf, unsigned int update_flags)
+ni_dhcp4_do_bits(const ni_dhcp4_device_t *dev, unsigned int update_flags)
 {
 	static unsigned int	do_mask[32] = {
 	[NI_ADDRCONF_UPDATE_DEFAULT_ROUTE]	= DHCP4_DO_GATEWAY,
@@ -427,8 +465,15 @@ ni_dhcp4_do_bits(const ni_config_dhcp4_t *conf, unsigned int update_flags)
 	[NI_ADDRCONF_UPDATE_BOOT]		= DHCP4_DO_ROOT,
 	[NI_ADDRCONF_UPDATE_TZ]			= DHCP4_DO_POSIX_TZ,
 	};
-	unsigned int bit, result = DHCP4_DO_ARP | DHCP4_DO_CSR
-				 | DHCP4_DO_STATIC_ROUTES;
+	unsigned int bit, result = DHCP4_DO_CSR | DHCP4_DO_STATIC_ROUTES;
+	const ni_config_dhcp4_t *conf = NULL;
+	const ni_netdev_t *ifp;
+
+	conf = ni_config_dhcp4_find_device(dev ? dev->ifname : NULL);
+	if ((ifp = ni_dhcp4_device_netdev(dev))) {
+		if (ifp->link.ifflags & NI_IFF_ARP_ENABLED)
+			result |= DHCP4_DO_ARP;
+	}
 
 	for (bit = 0; bit < 32; ++bit) {
 		if (update_flags & NI_BIT(bit))
@@ -505,11 +550,13 @@ ni_dhcp4_start_release(void *user_data, const ni_timer_t *timer)
 {
 	ni_dhcp4_device_t *dev = user_data;
 
-	if (dev->defer.timer != timer)
+	if (dev->timer.delay != timer) {
+		ni_warn("%s: bad timer handle", __func__);
 		return;
-	dev->defer.timer = NULL;
+	}
+	dev->timer.delay = NULL;
 
-	/* We just send out a singe RELEASE without waiting for the
+	/* We just send out a single RELEASE without waiting for the
 	 * server's reply. We just keep our fingers crossed that it's
 	 * getting out. If it doesn't, it's rather likely the network
 	 * is hosed anyway, so there's little point in delaying. */
@@ -519,14 +566,23 @@ ni_dhcp4_start_release(void *user_data, const ni_timer_t *timer)
 }
 
 int
-ni_dhcp4_release(ni_dhcp4_device_t *dev, const ni_uuid_t *req_uuid)
+ni_dhcp4_drop(ni_dhcp4_device_t *dev, const ni_dhcp4_drop_request_t *req)
 {
 	char *rel_uuid = NULL;
+	const char *action = "drop";
 
-	ni_string_dup(&rel_uuid, ni_uuid_print(req_uuid));
+	if (ni_tristate_is_set(req->release)) {
+		if (ni_tristate_is_enabled(req->release))
+			action = "release";
+	} else {
+		if (dev->config && dev->config->release_lease)
+			action = "release";
+	}
+
+	ni_string_dup(&rel_uuid, ni_uuid_print(&req->uuid));
 	if (dev->lease == NULL || dev->config == NULL) {
-		ni_info("%s: Request to release DHCPv4 lease%s%s: no lease",
-			dev->ifname,
+		ni_info("%s: Request to %s DHCPv4 lease%s%s: no lease",
+			dev->ifname, action,
 			rel_uuid ? " with UUID " : "", rel_uuid ? rel_uuid : "");
 
 		ni_string_free(&rel_uuid);
@@ -535,25 +591,27 @@ ni_dhcp4_release(ni_dhcp4_device_t *dev, const ni_uuid_t *req_uuid)
 		return -NI_ERROR_ADDRCONF_NO_LEASE;
 	}
 
-	ni_note("%s: Request to release DHCPv4 lease%s%s: releasing...",
-		dev->ifname,
+	ni_note("%s: Request to %s DHCPv4 lease%s%s: starting...",
+		dev->ifname, action,
 		rel_uuid ? " with UUID " : "", rel_uuid ? rel_uuid : "");
 	ni_string_free(&rel_uuid);
 
-	dev->lease->uuid = *req_uuid;
-	dev->config->uuid = *req_uuid;
+	dev->lease->uuid = req->uuid;
+	dev->config->uuid = req->uuid;
+	if (ni_tristate_is_enabled(req->release))
+		dev->config->release_lease = TRUE;
+	else
+	if (ni_tristate_is_disabled(req->release))
+		dev->config->release_lease = FALSE;
+
 	dev->fsm.state = NI_DHCP4_STATE_INIT;
 	ni_dhcp4_device_disarm_retransmit(dev);
-	if (dev->fsm.timer) {
-		ni_timer_cancel(dev->fsm.timer);
-		dev->fsm.timer = NULL;
-	}
+	ni_dhcp4_timer_disarm(&dev->fsm.timer);
+
 	ni_dhcp4_device_drop_best_offer(dev);
 	ni_dhcp4_device_arp_close(dev);
 
-	if (dev->defer.timer)
-		ni_timer_cancel(dev->defer.timer);
-	dev->defer.timer = ni_timer_register(0, ni_dhcp4_start_release, dev);
+	ni_dhcp4_timer_arm(&dev->timer.delay, 0, ni_dhcp4_start_release, dev);
 	return 1;
 }
 
@@ -564,6 +622,8 @@ void
 ni_dhcp4_device_event(ni_dhcp4_device_t *dev, ni_netdev_t *ifp, ni_event_t event)
 {
 	switch (event) {
+	case NI_EVENT_DEVICE_CHANGE:
+	case NI_EVENT_DEVICE_RENAME:
 	case NI_EVENT_DEVICE_UP:
 		if (!ni_string_eq(dev->ifname, ifp->name)) {
 			ni_debug_dhcp("%s: Updating interface name to %s",
@@ -575,12 +635,16 @@ ni_dhcp4_device_event(ni_dhcp4_device_t *dev, ni_netdev_t *ifp, ni_event_t event
 		break;
 
 	case NI_EVENT_LINK_DOWN:
-		ni_debug_dhcp("%s: link went down", dev->ifname);
+		ni_debug_dhcp("%s: link went down in state %s", dev->ifname,
+				ni_dhcp4_fsm_state_name(dev->fsm.state));
 		ni_dhcp4_fsm_link_down(dev);
 		break;
 
 	case NI_EVENT_LINK_UP:
-		ni_debug_dhcp("%s: link came up", dev->ifname);
+		/* retrigger dad on lease commit */
+		dev->link.reconnect = TRUE;
+		ni_debug_dhcp("%s: link came up in state %s", dev->ifname,
+				ni_dhcp4_fsm_state_name(dev->fsm.state));
 		ni_dhcp4_fsm_link_up(dev);
 		break;
 
@@ -592,22 +656,17 @@ static void
 ni_dhcp4_device_start_delayed(void *user_data, const ni_timer_t *timer)
 {
 	ni_dhcp4_device_t *dev = user_data;
-	ni_netconfig_t *nc;
-	ni_netdev_t *ifp;
 
-	if (dev->defer.timer != timer) {
+	if (dev->timer.delay != timer) {
 		ni_warn("%s: bad timer handle", __func__);
 		return;
 	}
-	dev->defer.timer = NULL;
+	dev->timer.delay = NULL;
 
-	nc = ni_global_state_handle(0);
-	ifp = ni_netdev_by_index(nc, dev->link.ifindex);
-	ni_dhcp4_fsm_init_device(dev);
-	if (ni_netdev_link_is_up(ifp)) {
+	if (ni_dhcp4_device_link_is_up(dev)) {
 		ni_dhcp4_fsm_link_up(dev);
 	} else {
-		ni_debug_dhcp("%s: defered start until link is up", dev->ifname);
+		ni_debug_dhcp("%s: deferred start until link is up", dev->ifname);
 	}
 }
 
@@ -616,11 +675,9 @@ ni_dhcp4_device_start(ni_dhcp4_device_t *dev)
 {
 	ni_netconfig_t *nc;
 	ni_netdev_t *ifp;
-	ni_int_range_t jitter = { .max = 500, /* msec */ };
-	unsigned long msec = dev->config->start_delay * 1000;
+	unsigned int sec;
 
 	ni_dhcp4_device_drop_buffer(dev);
-	dev->failed = 0;
 
 	nc = ni_global_state_handle(0);
 	if(!nc || !(ifp = ni_netdev_by_index(nc, dev->link.ifindex))) {
@@ -629,12 +686,12 @@ ni_dhcp4_device_start(ni_dhcp4_device_t *dev)
 		return -1;
 	}
 
-	/* Reuse defer pointer for this one-shot timer */
-	msec = ni_timeout_randomize(msec, &jitter);
-	if (dev->defer.timer)
-		ni_timer_cancel(dev->defer.timer);
-	dev->defer.timer = ni_timer_register(msec, ni_dhcp4_device_start_delayed, dev);
+	sec = ni_dhcp4_fsm_start_delay(dev->config->start_delay);
+	ni_dhcp4_timer_arm(&dev->timer.delay, NI_TIMEOUT_FROM_SEC(sec),
+			ni_dhcp4_device_start_delayed, dev);
 
+	ni_dhcp4_defer_timer_arm(dev);
+	ni_dhcp4_acquire_timer_arm(dev);
 	return 1;
 }
 
@@ -672,30 +729,31 @@ ni_dhcp4_device_prepare_message(void *data)
 	/* Build the DHCP4 message */
 	if (ni_dhcp4_build_message(dev, dev->transmit.msg_code, dev->transmit.lease, &dev->message) < 0) {
 		/* This is really terminal */
-		ni_error("unable to build DHCP4 message");
+		ni_error("%s: unable to build %s message with xid 0x%x in state %s",
+			dev->ifname, ni_dhcp4_message_name(dev->transmit.msg_code),
+			dev->dhcp4.xid, ni_dhcp4_fsm_state_name(dev->fsm.state));
 		return -1;
 	}
+
+	ni_debug_dhcp("%s: sending %s with xid 0x%x in state %s",
+			dev->ifname, ni_dhcp4_message_name(dev->transmit.msg_code),
+			dev->dhcp4.xid, ni_dhcp4_fsm_state_name(dev->fsm.state));
 	return 0;
 }
 
 int
-ni_dhcp4_device_send_message(ni_dhcp4_device_t *dev, unsigned int msg_code, const ni_addrconf_lease_t *lease)
+ni_dhcp4_device_send_message_broadcast(ni_dhcp4_device_t *dev, unsigned int msg_code, ni_addrconf_lease_t *lease)
 {
-	ni_timeout_param_t timeout;
 	int rv;
 
 	dev->transmit.msg_code = msg_code;
-	dev->transmit.lease = lease;
+	ni_addrconf_lease_hold(&dev->transmit.lease, lease);
 
 	if (ni_dhcp4_socket_open(dev) < 0) {
 		ni_error("%s: unable to open capture socket", dev->ifname);
 		goto transient_failure;
 	}
 
-	ni_debug_dhcp("%s: sending %s with xid 0x%x in state %s", dev->ifname,
-			ni_dhcp4_message_name(msg_code), htonl(dev->dhcp4.xid),
-			ni_dhcp4_fsm_state_name(dev->fsm.state));
-
 	if ((rv = ni_dhcp4_device_prepare_message(dev)) < 0)
 		return -1;
 
@@ -708,16 +766,9 @@ ni_dhcp4_device_send_message(ni_dhcp4_device_t *dev, unsigned int msg_code, cons
 	case DHCP4_DISCOVER:
 	case DHCP4_REQUEST:
 	case DHCP4_INFORM:
-		memset(&timeout, 0, sizeof(timeout));
-		timeout.timeout = dev->config->capture_retry_timeout;
-		timeout.increment = -1;
-		timeout.max_timeout = dev->config->capture_timeout;
-		timeout.nretries = -1;
-		timeout.jitter.min = -1;/* add a random jitter of +/-1 sec */
-		timeout.jitter.max = 1;
-		timeout.timeout_callback = ni_dhcp4_device_prepare_message;
-		timeout.timeout_data = dev;
-		rv = ni_capture_send(dev->capture, &dev->message, &timeout);
+		dev->transmit.params.timeout_callback = ni_dhcp4_device_prepare_message;
+		dev->transmit.params.timeout_data = dev;
+		rv = ni_capture_send(dev->capture, &dev->message, &dev->transmit.params);
 		break;
 
 	default:
@@ -738,24 +789,24 @@ transient_failure:
 }
 
 int
-ni_dhcp4_device_send_message_unicast(ni_dhcp4_device_t *dev, unsigned int msg_code, const ni_addrconf_lease_t *lease)
+ni_dhcp4_device_send_message_unicast(ni_dhcp4_device_t *dev, unsigned int msg_code, ni_addrconf_lease_t *lease)
 {
 	ni_sockaddr_t addr;
 
 	ni_sockaddr_set_ipv4(&addr, lease->dhcp4.server_id, DHCP4_SERVER_PORT);
 	dev->transmit.msg_code = msg_code;
-	dev->transmit.lease = lease;
+	ni_addrconf_lease_hold(&dev->transmit.lease, lease);
 
 	if (ni_dhcp4_socket_open(dev) < 0) {
 		ni_error("%s: unable to open capture socket", dev->ifname);
 		return -1;
 	}
 
-	ni_debug_dhcp("sending %s with xid 0x%x", ni_dhcp4_message_name(msg_code), htonl(dev->dhcp4.xid));
-
 	if (ni_dhcp4_device_prepare_message(dev) < 0)
 		return -1;
-	if (sendto(dev->listen_fd, ni_buffer_head(&dev->message), ni_buffer_count(&dev->message), 0,
+
+	if (sendto(dev->listen_fd, ni_buffer_head(&dev->message),
+				ni_buffer_count(&dev->message), 0,
 				&addr.sa, sizeof(addr.sin)) < 0)
 		ni_error("%s: sendto failed: %m", dev->ifname);
 	return 0;
@@ -764,11 +815,16 @@ ni_dhcp4_device_send_message_unicast(ni_dhcp4_device_t *dev, unsigned int msg_co
 void
 ni_dhcp4_device_disarm_retransmit(ni_dhcp4_device_t *dev)
 {
-	/* Clear retransmit timer */
+	/* Clear transmit struct except of transmit.start */
+	dev->transmit.msg_code = 0;
+	memset(&dev->transmit.params, 0, sizeof(dev->transmit.params));
+	ni_addrconf_lease_drop(&dev->transmit.lease);
+
+	/* Clear capture retransmit timer params */
 	if (dev->capture)
 		ni_capture_disarm_retransmit(dev->capture);
 
-	/* Drop the message buffer */
+	/* Drop the (raw) message buffer content */
 	ni_dhcp4_device_drop_buffer(dev);
 }
 
@@ -831,14 +887,14 @@ ni_dhcp4_set_dhcpv6_client_id(ni_opaque_t *raw, unsigned int iaid, const ni_opaq
  * Set the client ID as defined in the wicked-config(5)
  */
 ni_bool_t
-ni_dhcp4_set_config_client_id(ni_opaque_t *raw, const ni_dhcp4_device_t *dev)
+ni_dhcp4_set_config_client_id(ni_opaque_t *raw, const ni_dhcp4_device_t *dev,
+				unsigned int create_cid)
 {
 	const ni_config_dhcp4_t *dhcp4;
 	ni_netconfig_t *nc;
 	ni_netdev_t *ndev;
 	unsigned int iaid;
 	ni_opaque_t  duid;
-	unsigned int type;
 
 	if (!raw || !dev || !(nc = ni_global_state_handle(0)))
 		return FALSE;
@@ -846,21 +902,26 @@ ni_dhcp4_set_config_client_id(ni_opaque_t *raw, const ni_dhcp4_device_t *dev)
 	if (!(ndev = ni_netdev_by_index(nc, dev->link.ifindex)))
 		return FALSE;
 
-	dhcp4 = ni_config_dhcp4_find_device(dev->ifname);
-	if (!(type = dhcp4 ? dhcp4->create_cid : 0)) {
+	if (create_cid == NI_CONFIG_DHCP4_CID_TYPE_AUTO) {
+		/* use wicked-config(5) setting if any */
+		if ((dhcp4 = ni_config_dhcp4_find_device(dev->ifname)))
+			create_cid = dhcp4->create_cid;
+	}
+
+	if (create_cid == NI_CONFIG_DHCP4_CID_TYPE_AUTO) {
 		/*
-		 * We should allways use dhcp6 based client-id as
+		 * We should always use dhcp6 based client-id as
 		 * specified in RFC 4361, also on ethernet...
 		 *
 		 * This is also required to update DDNS records for
 		 * DHCPv6 and DHCPv4 IP addresses in same zone (the
 		 * server maintains a dhcid DNS record using it).
 		 *
-		 * Unfortunatelly, it would be a default behavior
+		 * Unfortunately, it would be a default behavior
 		 * change and may cause non-matching lease as well
 		 * as (ipv4 only) ddns update issues:
 		 *
-		 * There are many exising dhcp servers in the wild
+		 * There are many existing dhcp servers in the wild
 		 * relying on or using (static) leases with the
 		 * the "old" dhcp4 link address (mac) client-id,
 		 * incl. updated ddns records using it already...
@@ -872,17 +933,17 @@ ni_dhcp4_set_config_client_id(ni_opaque_t *raw, const ni_dhcp4_device_t *dev)
 		switch (dev->system.hwaddr.type) {
 		case ARPHRD_ETHER:
 #ifndef NI_DHCP4_RFC4361_CID
-			type = NI_CONFIG_DHCP4_CID_TYPE_HWADDR;
+			create_cid = NI_CONFIG_DHCP4_CID_TYPE_HWADDR;
 			break;
 #endif
 		case ARPHRD_INFINIBAND:
 		default:
-			type = NI_CONFIG_DHCP4_CID_TYPE_DHCPv6;
+			create_cid = NI_CONFIG_DHCP4_CID_TYPE_DHCPv6;
 			break;
 		}
 	}
 
-	switch (type) {
+	switch (create_cid) {
 	case NI_CONFIG_DHCP4_CID_TYPE_DHCPv6:
 		if (!ni_iaid_acquire(&iaid, ndev, 0))
 			return FALSE;
@@ -898,6 +959,7 @@ ni_dhcp4_set_config_client_id(ni_opaque_t *raw, const ni_dhcp4_device_t *dev)
 	case NI_CONFIG_DHCP4_CID_TYPE_DISABLE:
 		return TRUE;
 
+	case NI_CONFIG_DHCP4_CID_TYPE_AUTO:
 	default:
 		return FALSE;
 	}
@@ -1000,9 +1062,19 @@ ni_dhcp4_config_server_preference_hwaddr(const ni_hwaddr_t *hwaddr)
 }
 
 unsigned int
-ni_dhcp4_config_max_lease_time(void)
+ni_dhcp4_config_max_lease_time(const char *ifname)
 {
-	return ni_global.config->addrconf.dhcp4.lease_time;
+	const ni_config_dhcp4_t *dhconf = ni_config_dhcp4_find_device(ifname);
+
+	return dhconf && dhconf->lease_time ? dhconf->lease_time : NI_SECONDS_INFINITE;
+}
+
+ni_bool_t
+ni_dhcp4_config_ignore_rfc3927_1_6(const char *ifname)
+{
+	const ni_config_dhcp4_t *dhconf = ni_config_dhcp4_find_device(ifname);
+
+	return dhconf && dhconf->ignore_rfc3927_1_6;
 }
 
 static void
@@ -1066,6 +1138,13 @@ ni_dhcp4_request_free(ni_dhcp4_request_t *req)
 	free(req);
 }
 
+void
+ni_dhcp4_drop_request_init(ni_dhcp4_drop_request_t *req)
+{
+	ni_uuid_init(&req->uuid);
+	req->release = NI_TRISTATE_DEFAULT;
+}
+
 ni_bool_t
 ni_dhcp4_supported(const ni_netdev_t *ifp)
 {
@@ -1074,6 +1153,8 @@ ni_dhcp4_supported(const ni_netdev_t *ifp)
 	 * we've simply did not tested it on other links ...
 	 */
 	switch (ifp->link.hwaddr.type) {
+	case ARPHRD_NONE:
+		break;
 	case ARPHRD_ETHER:
 	case ARPHRD_INFINIBAND:
 		if (ifp->link.masterdev.index) {
