@@ -9,29 +9,6 @@
 #include "config.h"
 #endif
 
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/time.h>
-#include <time.h>
-#include <assert.h>
-
-#include <netinet/in.h>
-#include <netinet/ip.h>
-#include <netinet/udp.h>
-#include <net/if_arp.h>
-#include <net/ethernet.h>
-
-#include <arpa/inet.h>
-
-#include <errno.h>
-#include <limits.h>
-#include <fcntl.h>
-#include <unistd.h>
-#include <stdio.h>
-#include <stdint.h>
-#include <stdlib.h>
-#include <string.h>
-
 #include <wicked/netinfo.h>
 #include <wicked/route.h>
 #include <wicked/logging.h>
@@ -39,12 +16,28 @@
 #include <wicked/resolver.h>
 #include <wicked/nis.h>
 #include <wicked/xml.h>
+
 #include "dhcp4/dhcp4.h"
 #include "dhcp4/protocol.h"
 #include "dhcp.h"
 #include "buffer.h"
 #include "socket_priv.h"
 
+#include <limits.h>
+#include <errno.h>
+
+#include <unistd.h>
+#include <fcntl.h>
+
+#include <netinet/in.h>
+#include <netinet/ip.h>
+#include <netinet/udp.h>
+
+#include <net/if.h>
+#include <net/if_arp.h>
+#include <net/ethernet.h>
+
+
 static void	ni_dhcp4_socket_recv(ni_socket_t *);
 
 /*
@@ -112,7 +105,7 @@ ni_dhcp4_socket_open(ni_dhcp4_device_t *dev)
 		dev->capture = NULL;
 	}
 
-	dev->capture = ni_capture_open(&dev->system, &prot_info, ni_dhcp4_socket_recv);
+	dev->capture = ni_capture_open(&dev->system, &prot_info, ni_dhcp4_socket_recv, "dhcp4");
 	if (!dev->capture)
 		return -1;
 
@@ -120,6 +113,22 @@ ni_dhcp4_socket_open(ni_dhcp4_device_t *dev)
 	return 0;
 }
 
+/*
+ * Close DHCP4 listen/unicast and capture sockets
+ */
+void
+ni_dhcp4_socket_close(ni_dhcp4_device_t *dev)
+{
+	if (dev->capture) {
+		ni_capture_free(dev->capture);
+		dev->capture = NULL;
+	}
+	if (dev->listen_fd >= 0) {
+		close(dev->listen_fd);
+		dev->listen_fd = -1;
+	}
+}
+
 /*
  * This callback is invoked from the socket code when we
  * detect an incoming DHCP4 packet on the raw socket.
@@ -131,7 +140,7 @@ ni_dhcp4_socket_recv(ni_socket_t *sock)
 	ni_sockaddr_t from;
 	ni_buffer_t buf;
 
-	if (ni_capture_recv(capture, &buf, &from, "dhcp4") >= 0) {
+	if (ni_capture_recv(capture, &buf, &from) >= 0) {
 		ni_dhcp4_device_t *dev = ni_capture_get_user_data(capture);
 
 		ni_dhcp4_fsm_process_dhcp4_packet(dev, &buf, &from);
@@ -237,7 +246,7 @@ underflow:
 static int
 ni_dhcp4_option_get_opaque(ni_buffer_t *bp, ni_opaque_t *opaque)
 {
-	unsigned int len = ni_buffer_count(bp);
+	size_t len = ni_buffer_count(bp);
 
 	if (!len || len > sizeof(opaque->data))
 		return -1;
@@ -286,9 +295,9 @@ ni_dhcp4_option_get32(ni_buffer_t *bp, uint32_t *var)
 static int
 ni_dhcp4_option_get_string(ni_buffer_t *bp, char **var, unsigned int *lenp)
 {
-	unsigned int len = ni_buffer_count(bp);
+	size_t len = ni_buffer_count(bp);
 
-	if (len == 0)
+	if (len == 0 || len >= UINT_MAX)
 		return -1;
 
 	if (lenp)
@@ -383,7 +392,7 @@ ni_dhcp4_build_msg_put_fqdn_option(const ni_dhcp4_device_t *dev, ni_buffer_t *ms
 	ni_buffer_init(&databuf, optdata, sizeof(optdata));
 	if (ni_buffer_putc(&databuf, flags) < 0 || /* 0000NEOS   flags  */
 	    ni_buffer_putc(&databuf, 0)     < 0 || /* deprecated rdata1 */
-	    ni_buffer_putc(&databuf, 0)      < 0)   /* deprecared rdata2 */
+	    ni_buffer_putc(&databuf, 0)     < 0)   /* deprecared rdata2 */
 		return 1;
 
 	if (!ni_string_empty(hostname) &&
@@ -605,13 +614,6 @@ __ni_dhcp4_build_msg_put_client_id(const ni_dhcp4_device_t *dev, unsigned int ms
 				"%s: using client-id: %s", dev->ifname,
 				ni_print_hex(options->client_id.data,
 						options->client_id.len));
-	} else
-	if (!message->hwlen) {
-		ni_error("%s: cannot construct %s without usable hw-addr and client-id",
-				dev->ifname, ni_dhcp4_message_name(msg_code));
-		return -1;
-	} else {
-		return 1; /* skipped client-id */
 	}
 	return 0;
 }
@@ -637,34 +639,128 @@ __ni_dhcp4_build_msg_put_server_id(const ni_dhcp4_device_t *dev,
 	return 0;
 }
 
-
 static int
 __ni_dhcp4_build_msg_put_hwspec(const ni_dhcp4_device_t *dev, ni_dhcp4_message_t *message)
 {
 	switch (dev->system.hwaddr.type) {
 	case ARPHRD_ETHER:
 	case ARPHRD_IEEE802:
-		if (dev->system.hwaddr.len && dev->system.hwaddr.len <= sizeof(message->chaddr)) {
+		/*
+		 * See https://www.rfc-editor.org/rfc/rfc2131#section-2
+		 *
+		 *  FIELD  OCTETS DESCRIPTION
+		 *  -----  ------ -----------
+		 *  htype       1   Hardware address type, see ARP section in
+		 *                  "Assigned Numbers" RFC; e.g., '1' = 10mb ethernet.
+		 *  hlen        1   Hardware address length (e.g. '6' for 10mb ethernet).
+		 *  chaddr     16   Client hardware address.
+		 *  [...]
+		 *  flags       2   Flags (see figure 2).
+		 *  ciaddr      4   Client IP address; only filled in if client is in
+		 *                  BOUND, RENEW or REBINDING state and can respond
+		 *                  to ARP requests.
+		 *
+		 *  To work around some clients that cannot accept IP unicast datagrams
+		 *  before the TCP/IP software is configured [...] DHCP uses the 'flags'
+		 *  field. The leftmost bit is defined as the BROADCAST (B) flag.
+		 *
+		 * https://www.rfc-editor.org/rfc/rfc2131#section-3.5
+		 *   The client fills in the 'ciaddr' field only when correctly configured
+		 *   with an IP address in BOUND, RENEWING or REBINDING state.
+		 *
+		 * https://www.rfc-editor.org/rfc/rfc2131#section-4.1
+		 *   A client that cannot receive unicast IP datagrams until its protocol
+		 *   software has been configured with an IP address SHOULD set the
+		 *   BROADCAST bit in the 'flags' field to 1 in any DHCPDISCOVER or
+		 *   DHCPREQUEST messages that client sends.
+		 *   A client that can receive unicast IP datagrams before its protocol
+		 *   software has been configured SHOULD clear the BROADCAST bit to 0.
+		 */
+		message->hwtype = (uint8_t)dev->system.hwaddr.type;
+		if (dev->system.hwaddr.len &&
+		    dev->system.hwaddr.len <= sizeof(message->chaddr)) {
 			message->hwlen = dev->system.hwaddr.len;
-			memcpy(&message->chaddr, dev->system.hwaddr.data, dev->system.hwaddr.len);
+			memcpy(&message->chaddr, dev->system.hwaddr.data,
+						 dev->system.hwaddr.len);
 		}
-		if (ni_tristate_is_enabled(dev->config->broadcast))
+
+		if (ni_tristate_is_enabled(dev->config->broadcast) &&
+		    message->ciaddr == 0)
 			message->flags = htons(BROADCAST_FLAG);
 		break;
 
 	case ARPHRD_IEEE1394:
-	case ARPHRD_INFINIBAND:
-		if (ni_tristate_is_disabled(dev->config->broadcast))
-			ni_warn_once("%s: broadcast is mandatory on infiniband", dev->ifname);
-		/* See http://tools.ietf.org/html/rfc4390
+		/*
+		 * See http://tools.ietf.org/html/rfc2855
 		 *
-		 * Note: set the ciaddr before if needed.
+		 *  'htype' (hardware address type) MUST be 24 [ARPPARAM].
+		 *
+		 *  'hlen' (hardware address length) MUST be 0.
+		 *
+		 *  The 'chaddr' (client hardware address) field is reserved.
+		 *  The sender MUST set this field to zero, and the recipient
+		 *  and the relay agent MUST ignore its value on receipt.
+		 *
+		 *  A DHCP client on 1394 SHOULD set a BROADCAST flag in
+		 *  DHCPDISCOVER and DHCPREQUEST messages (and set 'ciaddr'
+		 *  to zero) to ensure that the server (or the relay agent)
+		 *  broadcasts its reply to the client.
+		 *
+		 *  'client identifier' option MUST be used in DHCP messages
+		 *  from the client [...] due to the lack of the 'chaddr'.
+		 */
+	case ARPHRD_INFINIBAND:
+		/*
+		 * See http://tools.ietf.org/html/rfc4390
+		 *
+		 * "htype" (hardware address type) MUST be 32 [ARPPARAM].
+		 *
+		 * "hlen" (hardware address length) MUST be 0.
+		 *
+		 * "chaddr" (client hardware address) field MUST be zeroed.
+		 *
+		 * "client-identifier" option MUST be used in DHCP messages.
+		 *
+		 * A DHCP client on IPoIB MUST set the BROADCAST flag in
+		 * DHCPDISCOVER and DHCPREQUEST messages (and set "ciaddr"
+		 * to zero) to ensure that the server (or the relay agent)
+		 * broadcasts its reply to the client.
+		 */
+		message->hwtype = dev->system.hwaddr.type;
+
+		if (ni_tristate_is_disabled(dev->config->broadcast)) {
+			ni_warn_once("%s: broadcast is mandatory on infiniband",
+					dev->ifname);
+		}
+
+		/*
+		 * Note: set the ciaddr before if/as needed.
 		 */
-		message->hwlen = 0;
 		if (message->ciaddr == 0)
 			message->flags = htons(BROADCAST_FLAG);
 		break;
 
+	case ARPHRD_NONE:
+		/*
+		 * No specific RFC known.
+		 * A non-arp,non-broadcast point-to-point interface type
+		 * transporting IP packets between two peers and not having
+		 * any hardware packet layer or address.
+		 *
+		 * Used on wwan-qmi interfaces in `raw-ip` mode which seem
+		 * to expect fake htype,hlen ethernet settings with zeroed
+		 * client hardware address in chaddr.
+		 *
+		 * BROADCAST response flag shouldn't be needed (POINTOPOINT);
+		 * set it on config request only.
+		 */
+		message->hwtype = (uint8_t)ARPHRD_ETHER;
+		message->hwlen = ni_link_address_length(ARPHRD_ETHER);
+		if (ni_tristate_is_enabled(dev->config->broadcast) &&
+		    message->ciaddr == 0)
+			message->flags = htons(BROADCAST_FLAG);
+		break;
+
 	default:
 		ni_error("%s: dhcp4 unsupported hardware type %s (0x%x)", dev->ifname,
 				ni_arphrd_type_to_name(dev->system.hwaddr.type),
@@ -697,14 +793,13 @@ __ni_dhcp4_build_msg_init_head(const ni_dhcp4_device_t *dev,
 
 	memset(message, 0, sizeof(*message));
 	message->op = DHCP4_BOOTREQUEST;
-	message->xid = dev->dhcp4.xid;
+	message->xid = htonl(dev->dhcp4.xid);
 	message->secs = htons(ni_dhcp4_device_uptime(dev, 0xFFFF));
 	message->cookie = htonl(MAGIC_COOKIE);
-	message->hwtype = dev->system.hwaddr.type;
 
 	ni_dhcp4_option_put8(msgbuf, DHCP4_MESSAGETYPE, msg_code);
 	ni_debug_verbose(NI_LOG_DEBUG1, NI_TRACE_DHCP,
-			"%s: xid: %x, secs: %u", dev->ifname,
+			"%s: xid: 0x%x, secs: %u", dev->ifname,
 			ntohl(message->xid), ntohs(message->secs));
 	ni_debug_verbose(NI_LOG_DEBUG1, NI_TRACE_DHCP,
 			"%s: using message type: %s", dev->ifname,
@@ -759,7 +854,7 @@ __ni_dhcp4_build_msg_discover(const ni_dhcp4_device_t *dev,
 			return -1;
 	}
 
-	if (options->classid && options->classid[0]) {
+	if (options->classid[0]) {
 		ni_dhcp4_option_puts(msgbuf, DHCP4_CLASSID, options->classid);
 	}
 
@@ -883,7 +978,7 @@ __ni_dhcp4_build_msg_inform(const ni_dhcp4_device_t *dev,
 			return -1;
 	}
 
-	if (options->classid && options->classid[0]) {
+	if (options->classid[0]) {
 		ni_dhcp4_option_puts(msgbuf, DHCP4_CLASSID, options->classid);
 	}
 
@@ -949,7 +1044,7 @@ __ni_dhcp4_build_msg_request_offer(const ni_dhcp4_device_t *dev,
 			return -1;
 	}
 
-	if (options->classid && options->classid[0]) {
+	if (options->classid[0]) {
 		ni_dhcp4_option_puts(msgbuf, DHCP4_CLASSID, options->classid);
 	}
 
@@ -1014,7 +1109,7 @@ __ni_dhcp4_build_msg_request_renew(const ni_dhcp4_device_t *dev,
 			return -1;
 	}
 
-	if (options->classid && options->classid[0]) {
+	if (options->classid[0]) {
 		ni_dhcp4_option_puts(msgbuf, DHCP4_CLASSID, options->classid);
 	}
 
@@ -1079,7 +1174,7 @@ __ni_dhcp4_build_msg_request_rebind(const ni_dhcp4_device_t *dev,
 			return -1;
 	}
 
-	if (options->classid && options->classid[0]) {
+	if (options->classid[0]) {
 		ni_dhcp4_option_puts(msgbuf, DHCP4_CLASSID, options->classid);
 	}
 
@@ -1139,7 +1234,7 @@ __ni_dhcp4_build_msg_request_reboot(const ni_dhcp4_device_t *dev,
 			return -1;
 	}
 
-	if (options->classid && options->classid[0]) {
+	if (options->classid[0]) {
 		ni_dhcp4_option_puts(msgbuf, DHCP4_CLASSID, options->classid);
 	}
 
@@ -1161,8 +1256,12 @@ ni_dhcp4_build_message(const ni_dhcp4_device_t *dev, unsigned int msg_code,
 		return -1;
 	}
 
-	if (IN_LINKLOCAL(ntohl(lease->dhcp4.address.s_addr))) {
-		ni_error("%s: cannot request a link local address", dev->ifname);
+	if (IN_LINKLOCAL(ntohl(lease->dhcp4.address.s_addr)) &&
+			!ni_dhcp4_config_ignore_rfc3927_1_6(dev->ifname)) {
+		ni_error("%s: Rejecting to request link local address offer prohibited by RFC3927 Section 1.6",
+				dev->ifname);
+		ni_error("%s: Fix the dhcp-server config or set `ignore-rfc3927-1-6` see `man 5 wicked-config`",
+				dev->ifname);
 		goto failed;
 	}
 
@@ -1831,7 +1930,6 @@ ni_dhcp4_parse_response(const ni_dhcp4_config_t *config, const ni_dhcp4_message_
 	lease->state = NI_ADDRCONF_STATE_GRANTED;
 	lease->type = NI_ADDRCONF_DHCP;
 	lease->family = AF_INET;
-	ni_timer_get_time(&lease->acquired);
 	lease->fqdn.enabled = NI_TRISTATE_DEFAULT;
 	lease->fqdn.qualify = config->fqdn.qualify;
 
@@ -1903,12 +2001,18 @@ parse_more:
 			ni_debug_dhcp("unable to parse DHCP4 option %s: too short",
 					ni_dhcp4_option_name(option));
 		} else if (ni_buffer_count(&buf)) {
-			ni_debug_dhcp("excess data in DHCP4 option %s - %u bytes left",
+			ni_debug_dhcp("excess data in DHCP4 option %s - %zu bytes left",
 					ni_dhcp4_option_name(option),
 					ni_buffer_count(&buf));
 		}
 	}
 
+	// We should have a msg_type by now
+	if (msg_type < 0) {
+		ni_debug_dhcp("unable to parse DHCP4 response: missing msg type");
+		goto error;
+	}
+
 	if (options->underflow) {
 		ni_debug_dhcp("unable to parse DHCP4 response: truncated packet");
 		goto error;
@@ -2071,7 +2175,7 @@ parse_more:
 			ni_debug_dhcp("unable to parse DHCP4 option %s (%u): too short",
 					ni_dhcp4_option_name(option), option);
 		} else if (ni_buffer_count(&buf)) {
-			ni_debug_dhcp("excess data in DHCP4 option %s (%u): %u data bytes left",
+			ni_debug_dhcp("excess data in DHCP4 option %s (%u): %zu data bytes left",
 					ni_dhcp4_option_name(option), option,
 					ni_buffer_count(&buf));
 		}
@@ -2154,7 +2258,7 @@ parse_more:
 		memset(&local_addr, 0, sizeof(local_addr));
 		local_addr.sin.sin_family = AF_INET;
 		local_addr.sin.sin_addr = lease->dhcp4.address;
-		ap = ni_address_new(AF_INET, pfxlen, &local_addr, &lease->addrs);
+		ap = ni_address_create(AF_INET, pfxlen, &local_addr, &lease->addrs);
 		if (ap && lease->dhcp4.broadcast.s_addr)
 			ni_sockaddr_set_ipv4(&ap->bcast_addr, lease->dhcp4.broadcast, 0);
 	}
