@@ -83,6 +83,10 @@ static void			ni_dhcp6_device_close(ni_dhcp6_device_t *);
 static void			ni_dhcp6_device_free(ni_dhcp6_device_t *);
 
 static void			ni_dhcp6_device_set_config(ni_dhcp6_device_t *, ni_dhcp6_config_t *);
+static void			ni_dhcp6_device_start_timer_init(ni_timeout_param_t *);
+static void			ni_dhcp6_device_start_timer_cancel(ni_dhcp6_device_t *);
+
+static ni_bool_t		ni_dhcp6_device_refresh_mode(ni_dhcp6_device_t *, ni_netdev_t *);
 
 static int			ni_dhcp6_device_transmit_arm_delay(ni_dhcp6_device_t *);
 static void			ni_dhcp6_device_retransmit_arm(ni_dhcp6_device_t *);
@@ -108,6 +112,17 @@ ni_dhcp6_device_new(const char *ifname, const ni_linkinfo_t *link)
 	ni_string_dup(&dev->ifname, ifname);
 	dev->link.ifindex = link->ifindex;
 
+	/*
+	 * https://tools.ietf.org/html/rfc8415#section-18.2 (rfc3315#section-18.1)
+	 * https://tools.ietf.org/html/rfc8415#section-18.2.12 (rfc3315#section-18.1.2)
+	 * https://tools.ietf.org/html/rfc8415#section-18.2.10.1 (rfc3315#section-18.1.8)
+	 *
+	 * it's either a fresh link and we have to perform dad anyway
+	 * or we just (re-)started and "may have moved to a new link",
+	 * so assume a reconnect to retrigger dad in next lease commit.
+	 */
+	dev->link.reconnect = TRUE;
+
 	dev->fsm.state = NI_DHCP6_STATE_INIT;
 
 	/* append to end of list */
@@ -147,6 +162,20 @@ ni_dhcp6_device_put(ni_dhcp6_device_t *dev)
 		ni_dhcp6_device_free(dev);
 }
 
+static ni_netdev_t *
+ni_dhcp6_device_netdev(const ni_dhcp6_device_t *dev)
+{
+	ni_netconfig_t *nc;
+	ni_netdev_t *ifp;
+
+	nc = ni_global_state_handle(0);
+	if (!nc || !(ifp = ni_netdev_by_index(nc, dev->link.ifindex))) {
+		ni_error("%s: Unable to find network interface by index %u",
+			dev->ifname, dev->link.ifindex);
+		return NULL;
+	}
+	return ifp;
+}
 
 /*
  * Cleanup functions
@@ -166,6 +195,11 @@ ni_dhcp6_device_close(ni_dhcp6_device_t *dev)
 void
 ni_dhcp6_device_stop(ni_dhcp6_device_t *dev)
 {
+	/*
+	 * Cancel start timer if any
+	 */
+	ni_dhcp6_device_start_timer_cancel(dev);
+
 	/*
 	 * Reset FSM timers and go to init state
 	 */
@@ -306,7 +340,7 @@ ni_dhcp6_device_uptime(const ni_dhcp6_device_t *dev, unsigned int clamp)
 {
 	struct timeval now;
 	struct timeval delta;
-	long           uptime = 0;
+	unsigned long uptime = 0;
 
 	ni_timer_get_time(&now);
 	if (timerisset(&dev->retrans.start) && timercmp(&now, &dev->retrans.start, >)) {
@@ -385,30 +419,143 @@ ni_dhcp6_device_show_addrs(ni_dhcp6_device_t *dev)
 	}
 }
 
+static void
+ni_dhcp6_device_start_timeout_cb(void *user_data, const ni_timer_t *timer)
+{
+	ni_dhcp6_device_t *dev = user_data;
+
+	if (!dev || dev->start_timer != timer)
+		return;
+	dev->start_timer = NULL;
+
+	ni_timeout_recompute(&dev->start_params);
+	ni_dhcp6_device_start(dev);
+}
+
+static void
+ni_dhcp6_device_start_timer_cancel(ni_dhcp6_device_t *dev)
+{
+	if (dev && dev->start_timer) {
+		ni_timer_cancel(dev->start_timer);
+		dev->start_timer = NULL;
+	}
+}
+
+static inline void
+ni_dhcp6_device_start_timer_init(ni_timeout_param_t *tmo)
+{
+	memset(tmo, 0, sizeof(*tmo));
+	tmo->jitter.max  = NI_DHCP6_MAX_JITTER;
+	tmo->timeout     = NI_DHCP6_SOL_TIMEOUT;
+	tmo->max_timeout = NI_DHCP6_SOL_MAX_RT;
+	tmo->nretries    = NI_DHCP6_UNLIMITED;
+	tmo->increment   = NI_DHCP6_EXP_BACKOFF;
+}
+
+static int
+ni_dhcp6_device_start_timer_arm(ni_dhcp6_device_t *dev)
+{
+	ni_timeout_t timeout;
+
+	timeout = ni_timeout_randomize(dev->start_params.timeout, &dev->start_params.jitter);
+	if (dev->start_timer) {
+		ni_timer_rearm(dev->start_timer, timeout);
+	} else {
+		dev->start_timer = ni_timer_register(timeout, ni_dhcp6_device_start_timeout_cb, dev);
+	}
+	return 1;
+}
+
+static inline int
+ni_dhcp6_device_start_auto_prefix(ni_dhcp6_device_t *dev)
+{
+	struct timeval now;
+	struct timeval end;
+	unsigned int defer;
+	ni_netdev_t *ifp;
+
+	/* auto + prefix:
+	 * wait a bit until ready + RA arrives, but then
+	 * start to request prefix and complete it later.
+	 */
+	if (!(ifp = ni_dhcp6_device_netdev(dev)))
+		return -1;
+
+	ni_dhcp6_device_show_addrs(dev);
+	if (!ni_dhcp6_device_is_ready(dev, ifp))
+		return ni_dhcp6_device_start_timer_arm(dev);
+
+	/* refresh in case kernel forgot to send it
+	 * (we increment timeout between attempts) */
+	ni_dhcp6_device_refresh_mode(dev, ifp);
+
+	/* request prefix after 1/3 defer timeout */
+	ni_timer_get_time(&now);
+	end = dev->start_time;
+	if (!(defer = (dev->config->defer_timeout / 3)))
+		defer = dev->config->acquire_timeout / 3;
+	end.tv_sec += defer;
+	if (timercmp(&end, &now, >) &&
+	    (dev->config->mode & NI_BIT(NI_DHCP6_MODE_AUTO)))
+		return ni_dhcp6_device_start_timer_arm(dev);
+
+	ni_dhcp6_device_start_timer_cancel(dev);
+	return ni_dhcp6_fsm_start(dev);
+}
+
+static inline int
+ni_dhcp6_device_start_auto(ni_dhcp6_device_t *dev)
+{
+	ni_netdev_t *ifp;
+
+	/* auto:
+	 * wait until ready + RA arrived
+	 */
+	if (dev->config->mode & NI_BIT(NI_DHCP6_MODE_PREFIX))
+		return ni_dhcp6_device_start_auto_prefix(dev);
+
+	if (!(ifp = ni_dhcp6_device_netdev(dev)))
+		return -1;
+
+	ni_dhcp6_device_show_addrs(dev);
+	if (!ni_dhcp6_device_is_ready(dev, ifp))
+		return ni_dhcp6_device_start_timer_arm(dev);
+
+	/* refresh in case kernel forgot to send it
+	 * (we increment timeout between attempts) */
+	ni_dhcp6_device_refresh_mode(dev, ifp);
+
+	if (dev->config->mode & NI_BIT(NI_DHCP6_MODE_AUTO))
+		return ni_dhcp6_device_start_timer_arm(dev);
+
+	ni_dhcp6_device_start_timer_cancel(dev);
+	return ni_dhcp6_fsm_start(dev);
+}
+
 int
 ni_dhcp6_device_start(ni_dhcp6_device_t *dev)
 {
-	if (!dev->config) {
+	if (!dev || !dev->config) {
 		ni_error("%s: Cannot start DHCPv6 without config",
 			dev->ifname);
 		return -1;
 	}
 
-	if (dev->config->mode == NI_DHCP6_MODE_AUTO)
-		return 1;
+	if (dev->config->mode & NI_BIT(NI_DHCP6_MODE_AUTO))
+		return ni_dhcp6_device_start_auto(dev);
 
 	ni_dhcp6_device_show_addrs(dev);
 	if (!ni_dhcp6_device_is_ready(dev, NULL))
-		return 1;
-
-	dev->failed = 0;
+		return ni_dhcp6_device_start_timer_arm(dev);
 
+	ni_dhcp6_device_start_timer_cancel(dev);
 	return ni_dhcp6_fsm_start(dev);
 }
 
 int
 ni_dhcp6_device_restart(ni_dhcp6_device_t *dev)
 {
+	ni_stringbuf_t buf = NI_STRINGBUF_INIT_DYNAMIC;
 	char *err = NULL;
 	int rv = -1;
 
@@ -419,15 +566,17 @@ ni_dhcp6_device_restart(ni_dhcp6_device_t *dev)
 
 	ni_debug_dhcp("%s: Restart DHCPv6 acquire request %s in mode %s",
 		dev->ifname, ni_uuid_print(&dev->request->uuid),
-		ni_dhcp6_mode_type_to_name(dev->request->mode));
+		ni_dhcp6_mode_format(&buf, dev->request->mode, NULL));
+	ni_stringbuf_destroy(&buf);
 
 	if ((rv = ni_dhcp6_acquire(dev, dev->request, &err)) >= 0)
 		return rv;
 
 	ni_error("%s: Cannot restart DHCPv6 acquire request %s in mode %s%s%s",
 		dev->ifname, ni_uuid_print(&dev->request->uuid),
-		ni_dhcp6_mode_type_to_name(dev->request->mode),
+		ni_dhcp6_mode_format(&buf, dev->request->mode, NULL),
 		(err ? ": " : ""), (err ? err : ""));
+	ni_stringbuf_destroy(&buf);
 	ni_string_free(&err);
 
 	/* Also discard the request */
@@ -502,28 +651,13 @@ ni_dhcp6_device_find_lladdr(ni_dhcp6_device_t *dev)
 	return rv;
 }
 
-static ni_netdev_t *
-ni_dhcp6_device_netdev(const ni_dhcp6_device_t *dev)
-{
-	ni_netconfig_t *nc;
-	ni_netdev_t *ifp;
-
-	nc = ni_global_state_handle(0);
-	if (!nc || !(ifp = ni_netdev_by_index(nc, dev->link.ifindex))) {
-		ni_error("%s: Unable to find network interface by index %u",
-			dev->ifname, dev->link.ifindex);
-		return NULL;
-	}
-	return ifp;
-}
-
-void
+static ni_bool_t
 ni_dhcp6_device_refresh_mode(ni_dhcp6_device_t *dev, ni_netdev_t *ifp)
 {
 	ni_netconfig_t *nc = ni_global_state_handle(0);
 
 	if (!nc || !dev || (!ifp && !(ifp = ni_dhcp6_device_netdev(dev))))
-		return;
+		return FALSE;
 
 	/*
 	 * Refresh ipv6 link info on NEWPREFIX events in auto mode
@@ -537,22 +671,45 @@ ni_dhcp6_device_refresh_mode(ni_dhcp6_device_t *dev, ni_netdev_t *ifp)
 	 */
 	__ni_device_refresh_ipv6_link_info(nc, ifp);
 
-	ni_dhcp6_device_update_mode(dev, ifp);
+	return ni_dhcp6_device_update_mode(dev, ifp);
 }
 
-void
+ni_bool_t
 ni_dhcp6_device_update_mode(ni_dhcp6_device_t *dev, const ni_netdev_t *ifp)
 {
-	if (!ifp && !(ifp = ni_dhcp6_device_netdev(dev)))
-		return;
+	ni_stringbuf_t old = NI_STRINGBUF_INIT_DYNAMIC;
+	ni_stringbuf_t new = NI_STRINGBUF_INIT_DYNAMIC;
+	unsigned int omode = 0;
 
-	if (ifp->ipv6 && dev->config) {
-		if (ifp->ipv6->radv.managed_addr)
-			dev->config->mode = NI_DHCP6_MODE_MANAGED;
-		else
-		if (ifp->ipv6->radv.other_config)
-			dev->config->mode = NI_DHCP6_MODE_INFO;
+	if (!ifp && !(ifp = ni_dhcp6_device_netdev(dev)))
+		return FALSE;
+
+	if (ni_ipv6_devinfo_ra_received(ifp->ipv6) && dev->config) {
+		omode = dev->config->mode;
+
+		if (ifp->ipv6->radv.managed_addr) {
+			dev->config->mode |= NI_BIT(NI_DHCP6_MODE_MANAGED);
+			dev->config->mode = ni_dhcp6_mode_adjust(dev->config->mode);
+		} else
+		if (ifp->ipv6->radv.other_config) {
+			dev->config->mode |= NI_BIT(NI_DHCP6_MODE_INFO);
+			dev->config->mode = ni_dhcp6_mode_adjust(dev->config->mode);
+		} else {
+			dev->config->mode &= ~NI_BIT(NI_DHCP6_MODE_AUTO);
+		}
+
+		if (omode != dev->config->mode) {
+			ni_dhcp6_mode_format(&old, omode, NULL);
+			ni_dhcp6_mode_format(&new, dev->config->mode, NULL);
+			ni_debug_dhcp("%s: updated dhcp6 mode from %s to %s",
+					dev->ifname, old.string, new.string ?
+					new.string : "disabled");
+			ni_stringbuf_destroy(&old);
+			ni_stringbuf_destroy(&new);
+			return TRUE;
+		}
 	}
+	return FALSE;
 }
 
 ni_bool_t
@@ -614,25 +771,28 @@ static int
 ni_dhcp6_device_transmit_arm_delay(ni_dhcp6_device_t *dev)
 {
 	ni_int_range_t jitter;
-	unsigned long  delay;
+	ni_timeout_t   delay;
 
 	/*
 	 * rfc3315#section-5.5 (17.1.2, 18.1.2, 18.1.5):
 	 *
 	 * Initial delay is a MUST for Solicit, Confirm and InfoRequest.
+	 * "[..]
+	 *  MUST be delayed by a random amount of time between 0 and [..]_MAX_DELAY
+	 * [..]"
+	 *
+	 * We could track the RA receive time that "causes the client to invoke the
+	 * stateful address autoconfiguration" and subtract the delta from delay...
 	 */
 	if (dev->retrans.delay == 0)
 		return FALSE;
 
-	ni_debug_dhcp("%s: setting initial transmit delay of %u [%d .. %d] msec",
-			dev->ifname, dev->retrans.delay,
-			0 - dev->retrans.jitter,
-			0 + dev->retrans.jitter);
+	jitter.min = 0;
+	jitter.max = dev->retrans.delay;
+	delay = ni_timeout_randomize(0, &jitter);
 
-	/* we can use base jitter as is, it's 0.1 msec already */
-	jitter.min = 0 - dev->retrans.jitter;
-	jitter.max = 0 + dev->retrans.jitter;
-	delay = ni_timeout_randomize(dev->retrans.delay, &jitter);
+	ni_debug_dhcp("%s: setting initial transmit delay of 0 .. %u.%03us",
+			dev->ifname, NI_TIMEOUT_SEC(delay), NI_TIMEOUT_MSEC(delay));
 
 	ni_dhcp6_fsm_set_timeout_msec(dev, delay);
 
@@ -661,23 +821,8 @@ ni_dhcp6_device_retransmit_arm(ni_dhcp6_device_t *dev)
 		 * by choosing RAND to be strictly greater than 0.
 		 * [...]"
 		 */
-		dev->retrans.params.jitter = ni_dhcp6_jitter_rebase(
-				dev->retrans.params.timeout,
-				0, /* exception, no negative jitter */
-				0 + dev->retrans.jitter);
-
-		/*
-		 * rfc3315#section-14
-		 *
-		 * "[...]
-		 *  RT for the first message transmission is based on IRT:
-		 * 	RT = IRT + RAND*IRT
-		 *  [...]"
-		 *
-		 * IRT is already initialized in retrans.params.timeout.
-		 */
-		dev->retrans.params.timeout = ni_timeout_arm_msec(&dev->retrans.deadline,
-								  &dev->retrans.params);
+		dev->retrans.params.jitter.min = 0; /* exception, no negative jitter */
+		dev->retrans.params.jitter.max = 0 + dev->retrans.jitter;
 	} else {
 		/*
 		 * rfc3315#section-14
@@ -687,23 +832,31 @@ ni_dhcp6_device_retransmit_arm(ni_dhcp6_device_t *dev)
 		 * between -0.1 and +0.1.
 		 * [...]"
 		 */
-		dev->retrans.params.jitter = ni_dhcp6_jitter_rebase(
-				dev->retrans.params.timeout,
-				0 - dev->retrans.jitter,
-				0 + dev->retrans.jitter);
-
-		/*
-		 * rfc3315#section-14
-		 *
-		 * "[...]RT for the first message transmission is based on IRT:
-		 * 		RT = IRT + RAND*IRT
-		 *  [...]"
-		 *
-		 *  IRT is already initialized in retrans.params.timeout.
-		 */
-		dev->retrans.params.timeout = ni_timeout_arm_msec(&dev->retrans.deadline,
-								  &dev->retrans.params);
+		dev->retrans.params.jitter.min = 0 - dev->retrans.jitter,
+		dev->retrans.params.jitter.max = 0 + dev->retrans.jitter;
 	}
+
+	/*
+	 * rfc3315#section-14
+	 *
+	 * "[...]RT for the first message transmission is based on IRT:
+	 * 		RT = IRT + RAND*IRT
+	 *  [...]"
+	 *
+	 *  IRT is already initialized in retrans.params.timeout.
+	 */
+	dev->retrans.params.timeout += ni_timeout_randomize(dev->retrans.params.timeout,
+							&dev->retrans.params.jitter);
+	ni_timer_get_time(&dev->retrans.deadline);
+	ni_timeval_add_timeout(&dev->retrans.deadline, dev->retrans.params.timeout);
+	ni_debug_dhcp("%s: initialized xid 0x%06x retransmission timeout of %u.%03u [%.3f .. %.3f] sec",
+			dev->ifname, dev->dhcp6.xid,
+			NI_TIMEOUT_SEC(dev->retrans.params.timeout),
+			NI_TIMEOUT_MSEC(dev->retrans.params.timeout),
+			(double)dev->retrans.params.jitter.min/1000,
+			(double)dev->retrans.params.jitter.max/1000);
+
+
 	if (dev->retrans.duration) {
 		/*
 		 * rfc3315#section-14
@@ -716,6 +869,10 @@ ni_dhcp6_device_retransmit_arm(ni_dhcp6_device_t *dev)
 		 * [...]"
 		 */
 		ni_dhcp6_fsm_set_timeout_msec(dev, dev->retrans.duration);
+		ni_debug_dhcp("%s: initialized xid 0x%06x duration %u.%03u sec",
+			dev->ifname, dev->dhcp6.xid,
+			NI_TIMEOUT_SEC(dev->retrans.duration),
+			NI_TIMEOUT_MSEC(dev->retrans.duration));
 	}
 }
 
@@ -727,8 +884,8 @@ ni_dhcp6_device_retransmit_disarm(ni_dhcp6_device_t *dev)
 	ni_timer_get_time(&now);
 
 	if (dev->dhcp6.xid || dev->retrans.params.timeout) {
-		ni_debug_dhcp("%s: disarming retransmission at %s",
-				dev->ifname, ni_dhcp6_print_timeval(&now));
+		ni_debug_dhcp("%s: disarming xid 0x%06x retransmission",
+				dev->ifname, dev->dhcp6.xid);
 	}
 
 	dev->dhcp6.xid = 0;
@@ -738,6 +895,8 @@ ni_dhcp6_device_retransmit_disarm(ni_dhcp6_device_t *dev)
 static ni_bool_t
 ni_dhcp6_device_retransmit_advance(ni_dhcp6_device_t *dev)
 {
+	ni_timeout_t previous = dev->retrans.params.timeout;
+
 	/*
 	 * rfc3315#section-14
 	 *
@@ -753,30 +912,26 @@ ni_dhcp6_device_retransmit_advance(ni_dhcp6_device_t *dev)
 	 *
 	 */
 	if( ni_timeout_recompute(&dev->retrans.params)) {
-		unsigned int old_timeout = dev->retrans.params.timeout;
+		dev->retrans.params.jitter = ni_dhcp6_jitter_rebase(previous,
+						0 - dev->retrans.jitter,
+						0 + dev->retrans.jitter);
+		dev->retrans.params.timeout += ni_timeout_randomize(previous,
+						&dev->retrans.params.jitter);
+		ni_timer_get_time(&dev->retrans.deadline);
+		ni_timeval_add_timeout(&dev->retrans.deadline, dev->retrans.params.timeout);
 
-		/*
-		 * Hmm... should we set this as backoff callback?
-		 */
-		dev->retrans.params.jitter = ni_dhcp6_jitter_rebase(
-				dev->retrans.params.timeout,
-				0 - dev->retrans.jitter,
-				0 + dev->retrans.jitter);
-
-		dev->retrans.params.timeout = ni_timeout_arm_msec(
-				&dev->retrans.deadline,
-				&dev->retrans.params);
-
-		ni_debug_dhcp("%s: increased retransmission timeout from %u to %u [%d .. %d]: %s",
-				dev->ifname, old_timeout,
-				dev->retrans.params.timeout,
-				dev->retrans.params.jitter.min,
-				dev->retrans.params.jitter.max,
-				ni_dhcp6_print_timeval(&dev->retrans.deadline));
+		ni_debug_dhcp("%s: advanced xid 0x%06x retransmission timeout from %u.%03u to %u.%03u [%.3f .. %.3f] sec",
+				dev->ifname, dev->dhcp6.xid,
+				NI_TIMEOUT_SEC(previous),
+				NI_TIMEOUT_MSEC(previous),
+				NI_TIMEOUT_SEC(dev->retrans.params.timeout),
+				NI_TIMEOUT_MSEC(dev->retrans.params.timeout),
+				(double)dev->retrans.params.jitter.min/1000,
+				(double)dev->retrans.params.jitter.max/1000);
 
 		return TRUE;
 	}
-	ni_debug_dhcp("%s: retransmission limit reached", dev->ifname);
+	ni_debug_dhcp("%s: xid 0x%06x retransmission limit reached", dev->ifname, dev->dhcp6.xid);
 	return FALSE;
 }
 
@@ -794,8 +949,10 @@ ni_dhcp6_device_retransmit(ni_dhcp6_device_t *dev)
 	if ((rv = ni_dhcp6_fsm_retransmit(dev)) < 0)
 		return rv;
 
-	ni_debug_dhcp("%s: retransmitted, next deadline at %s", dev->ifname,
-			ni_dhcp6_print_timeval(&dev->retrans.deadline));
+	ni_debug_dhcp("%s: xid 0x%06x retransmitted, next deadline in %ld.%03ld",
+			dev->ifname, dev->dhcp6.xid,
+			dev->retrans.deadline.tv_sec,
+			dev->retrans.deadline.tv_usec/1000);
 	return 0;
 }
 
@@ -849,6 +1006,7 @@ __nondefault(unsigned int req, unsigned int def)
 int
 ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **err)
 {
+	ni_stringbuf_t buf = NI_STRINGBUF_INIT_DYNAMIC;
 	ni_dhcp6_config_t *config;
 	const char *mode;
 	size_t len;
@@ -863,27 +1021,22 @@ ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **e
 		return -NI_ERROR_INVALID_ARGS;
 	}
 
-	switch (req->mode) {
-	case NI_DHCP6_MODE_AUTO:
-	case NI_DHCP6_MODE_INFO:
-	case NI_DHCP6_MODE_MANAGED:
+	config = xcalloc(1, sizeof(*config));
+	config->uuid = req->uuid;
+	config->mode = ni_dhcp6_mode_adjust(req->mode);
+	config->flags= req->flags;
+
+	if ((mode = ni_dhcp6_mode_format(&buf, config->mode, NULL))) {
 		ni_note("%s: Request to acquire DHCPv6 lease with UUID %s in mode %s",
-			dev->ifname, ni_uuid_print(&req->uuid),
-			ni_dhcp6_mode_type_to_name(req->mode));
-		break;
-	default:
-		if ((mode = ni_dhcp6_mode_type_to_name(req->mode)) != NULL) {
-			ni_string_printf(err, "unsupported mode %s", mode);
-		} else {
-			ni_string_printf(err, "invalid mode %u", req->mode);
-		}
+			dev->ifname, ni_uuid_print(&config->uuid), mode);
+		ni_stringbuf_destroy(&buf);
+	} else {
+		ni_string_printf(err, "invalid DHCPv6 request mode 0x%x (0x%x)",
+				req->mode, config->mode);
+		ni_dhcp6_device_config_free(config);
 		return -NI_ERROR_INVALID_ARGS;
 	}
 
-	config = xcalloc(1, sizeof(*config));
-	config->uuid = req->uuid;
-	config->mode = req->mode;
-	config->flags= req->flags;
 	if (req->update == -1U) {
 		config->update = ni_config_addrconf_update(dev->ifname, NI_ADDRCONF_DHCP, AF_INET6);
 	} else {
@@ -892,7 +1045,12 @@ ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **e
 	}
 	config->dry_run	= req->dry_run;
 
+	if (req->address_len <= ni_af_address_prefixlen(AF_INET6))
+		config->address_len = req->address_len;
+
 	ni_timer_get_time(&dev->start_time);
+	ni_dhcp6_device_start_timer_init(&dev->start_params);
+
 	config->start_delay	= __nondefault(req->start_delay,
 					NI_DHCP6_START_DELAY);
 
@@ -920,6 +1078,7 @@ ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **e
 	config->lease_time	= __nondefault(req->lease_time,
 						NI_DHCP6_LEASE_TIME);
 	config->recover_lease	= req->recover_lease;
+	config->refresh_lease  	= req->refresh_lease;
 	config->release_lease	= req->release_lease;
 
 	if (!dev->lease && config->dry_run != NI_DHCP6_RUN_OFFER && config->recover_lease)
@@ -927,6 +1086,7 @@ ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **e
 
 	if (!ni_dhcp6_device_iaid(dev, &dev->iaid)) {
 		ni_string_printf(err, "Unable to generate a device IAID");
+		ni_dhcp6_device_config_free(config);
 		return -NI_ERROR_GENERAL_FAILURE;
 	}
 	if (!ni_dhcp6_config_init_duid(dev, config, req->clientid)) {
@@ -943,19 +1103,32 @@ ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **e
 		}
 	}
 
-	/*
-	 * Hmm... in info mode we don't need any IA's,
-	 *        in auto mode, we don't know yet ...
-	 */
-	if (config->mode != NI_DHCP6_MODE_INFO) {
-		if (req->ia_list == NULL) {
-			ni_dhcp6_ia_t *ia = ni_dhcp6_ia_na_new(dev->iaid);
-			ni_dhcp6_ia_set_default_lifetimes(ia, config->lease_time);
+	/* Copy IA-PD (only) with prefix-hint to the config for later use.
+	 * Another IA's aren't using a hint and will be added according to
+	 * the managed/info mode bit later to request automatically.
+	 *
+	 * Once we support multiple prefixes, we will need to sort requests
+	 * by iaid into a unique list of IAs + per-ia prefix hints. */
+	if ((config->mode & NI_BIT(NI_DHCP6_MODE_PREFIX)) && req->prefix_reqs) {
+		const ni_dhcp6_prefix_req_t *pr;
+		ni_dhcp6_ia_addr_t *ph, *padr;
+		ni_dhcp6_ia_t *ia;
+
+		for (pr = req->prefix_reqs; pr; pr = pr->next) {
+			/* one IA using our iaid + hint for now */
+			if (!(ia = ni_dhcp6_ia_pd_new(dev->iaid)))
+				continue;
+
+			for (ph = pr->hints; ph; ph = ph->next) {
+				if (!ph->plen)
+					continue;
+
+				padr = ni_dhcp6_ia_addr_clone(ph);
+				ni_dhcp6_ia_addr_list_append(&ia->addrs, padr);
+				break; /* one pd hint per ia only */
+			}
 			ni_dhcp6_ia_list_append(&config->ia_list, ia);
-		} else {
-			/* TODO: Merge multiple ia's of same type into one?
-			 *       for tests we take it as is -- at the moment */
-			ni_dhcp6_ia_list_copy(&config->ia_list, req->ia_list, FALSE);
+			break; /* one ia-pd only */
 		}
 	}
 
@@ -999,33 +1172,22 @@ ni_dhcp6_acquire(ni_dhcp6_device_t *dev, const ni_dhcp6_request_t *req, char **e
 
 	ni_dhcp6_device_set_config(dev, config);
 
-	if (config->mode == NI_DHCP6_MODE_AUTO) {
-		/* refresh in case kernel forgot a newlink on RA */
-		ni_dhcp6_device_refresh_mode(dev, NULL);
-	}
-
-	if (config->mode == NI_DHCP6_MODE_AUTO) {
-		unsigned int deadline = 0;
-
-		if (config->defer_timeout) {
-			/*
-			 * set timer to emit lease-deferred signal to wicked
-			 * when there is no IPv6 RA on the network or DHCPv6
-			 * is not used (managed and other-config unset).
-			 */
-			deadline = config->defer_timeout * 1000;
-			ni_dhcp6_fsm_set_timeout_msec(dev, deadline);
-			dev->fsm.fail_on_timeout = 0;
-		} else
-		if (config->acquire_timeout) {
-			/*
-			 * immediatelly set timer to fail after timeout,
-			 * that is to drop config, disarm fsm and stop.
-			 */
-			deadline = config->acquire_timeout * 1000;
-			ni_dhcp6_fsm_set_timeout_msec(dev, deadline);
-			dev->fsm.fail_on_timeout = 1;
-		}
+	if (config->defer_timeout) {
+		/*
+		 * set timer to emit lease-deferred signal to wicked
+		 * when there is no IPv6 RA on the network or DHCPv6
+		 * is not used (managed and other-config unset).
+		 */
+		ni_dhcp6_fsm_set_timeout_sec(dev, config->defer_timeout);
+		dev->fsm.fail_on_timeout = 0;
+	} else
+	if (config->acquire_timeout) {
+		/*
+		 * immediately set timer to fail after timeout,
+		 * that is to drop config, disarm fsm and stop.
+		 */
+		ni_dhcp6_fsm_set_timeout_sec(dev, config->acquire_timeout);
+		dev->fsm.fail_on_timeout = 1;
 	}
 
 	/*
@@ -1062,13 +1224,23 @@ ni_dhcp6_start_release(void *user_data, const ni_timer_t *timer)
 }
 
 int
-ni_dhcp6_release(ni_dhcp6_device_t *dev, const ni_uuid_t *req_uuid)
+ni_dhcp6_drop(ni_dhcp6_device_t *dev, const ni_dhcp6_drop_request_t *req)
 {
 	char *rel_uuid = NULL;
+	const char *action = "drop";
 
-	ni_string_dup(&rel_uuid, ni_uuid_print(req_uuid));
+	if (ni_tristate_is_set(req->release)) {
+		if (ni_tristate_is_enabled(req->release))
+			action = "release";
+	} else {
+		if (dev->config && dev->config->release_lease)
+			action = "release";
+	}
+
+	ni_string_dup(&rel_uuid, ni_uuid_print(&req->uuid));
 	if (dev->lease == NULL || dev->config == NULL) {
-		ni_info("%s: Request to release DHCPv6 lease%s%s: no lease", dev->ifname,
+		ni_info("%s: Request to %s DHCPv6 lease%s%s: no lease",
+			dev->ifname, action,
 			rel_uuid ? " using UUID " : "", rel_uuid ? rel_uuid : "");
 		ni_string_free(&rel_uuid);
 
@@ -1078,13 +1250,20 @@ ni_dhcp6_release(ni_dhcp6_device_t *dev, const ni_uuid_t *req_uuid)
 		return -NI_ERROR_ADDRCONF_NO_LEASE;
 	}
 
-	ni_note("%s: Request to release DHCPv6 lease%s%s: releasing...", dev->ifname,
+	ni_note("%s: Request to %s DHCPv6 lease%s%s: starting...",
+			dev->ifname, action,
 			rel_uuid ? " using UUID " : "", rel_uuid ? rel_uuid : "");
 	ni_string_free(&rel_uuid);
 
-	dev->lease->uuid = *req_uuid;
-	dev->config->uuid = *req_uuid;
+	dev->lease->uuid = req->uuid;
+	dev->config->uuid = req->uuid;
+	if (ni_tristate_is_enabled(req->release))
+		dev->config->release_lease = TRUE;
+	else
+	if (ni_tristate_is_disabled(req->release))
+		dev->config->release_lease = FALSE;
 
+	ni_dhcp6_device_start_timer_cancel(dev);
 	ni_dhcp6_fsm_reset(dev);
 	dev->fsm.state = NI_DHCP6_STATE_RELEASING;
 	dev->fsm.timer = ni_timer_register(0, ni_dhcp6_start_release, dev);
@@ -1118,6 +1297,8 @@ ni_dhcp6_device_event(ni_dhcp6_device_t *dev, ni_netdev_t *ifp, ni_event_t event
 	break;
 
 	case NI_EVENT_LINK_UP:
+		/* retrigger dad on lease commit */
+		dev->link.reconnect = TRUE;
 		ni_debug_dhcp("received link up event");
 		if (dev->config) {
 			ni_dhcp6_device_start(dev);
@@ -1132,7 +1313,7 @@ ni_dhcp6_device_event(ni_dhcp6_device_t *dev, ni_netdev_t *ifp, ni_event_t event
 	break;
 
 	case NI_EVENT_DEVICE_CHANGE:
-		if (dev->config && dev->config->mode == NI_DHCP6_MODE_AUTO) {
+		if (dev->config && (dev->config->mode & NI_BIT(NI_DHCP6_MODE_AUTO))) {
 			ni_dhcp6_device_update_mode(dev, ifp);
 			ni_dhcp6_device_start(dev);
 		}
@@ -1185,7 +1366,7 @@ ni_dhcp6_prefix_event(ni_dhcp6_device_t *dev, ni_netdev_t *ifp, ni_event_t event
 {
 	switch (event) {
 	case NI_EVENT_PREFIX_UPDATE:
-		if (dev->config && dev->config->mode == NI_DHCP6_MODE_AUTO) {
+		if (dev->config && (dev->config->mode & NI_BIT(NI_DHCP6_MODE_AUTO))) {
 			/* refresh in case kernel forgot a newlink on RA */
 			ni_dhcp6_device_refresh_mode(dev, ifp);
 			ni_server_trace_interface_prefix_events(ifp, event, pi);
@@ -1254,10 +1435,9 @@ ni_dhcp6_device_transmit(ni_dhcp6_device_t *dev)
 		ni_timer_get_time(&now);
 		ni_debug_verbose(NI_LOG_DEBUG, NI_TRACE_DHCP,
 				"%s: %s message #%u, xid 0x%x sent with %zd of %zu"
-				" byte to %s at %s",
+				" byte to %s",
 				dev->ifname, name, dev->retrans.count, xid, rv, cnt,
-				ni_sockaddr_print(&dev->mcast.dest),
-				ni_dhcp6_print_timeval(&now));
+				ni_sockaddr_print(&dev->mcast.dest));
 
 		ni_buffer_clear(&dev->message);
 		return 0;
@@ -1409,50 +1589,6 @@ ni_dhcp6_config_set_request_options(const char *ifname, ni_uint_array_t *cfg, co
 	}
 }
 
-ni_string_array_t *
-ni_dhcp6_get_ia_addrs(struct ni_dhcp6_ia *ia_list, ni_var_array_t *p_lft, ni_var_array_t *v_lft)
-{
-	ni_string_array_t *addrs = NULL;
-	const ni_dhcp6_ia_t *ia = NULL;
-
-	addrs = xcalloc(1, sizeof(ni_string_array_t));
-
-	for (ia = ia_list; ia; ia = ia->next) {
-		const ni_dhcp6_ia_addr_t *iaddr = NULL;
-		ni_sockaddr_t addr;
-		const char *addr_str = NULL;
-		for (iaddr = ia->addrs; iaddr; iaddr = iaddr->next) {
-			ni_sockaddr_set_ipv6(&addr, iaddr->addr, 0);
-			switch (ia->type) {
-			case NI_DHCP6_OPTION_IA_TA:
-			case NI_DHCP6_OPTION_IA_NA:
-				addr_str = ni_sockaddr_print(&addr);
-				ni_string_array_append(addrs, addr_str);
-				break;
-
-			case NI_DHCP6_OPTION_IA_PD:
-				addr_str = ni_sockaddr_prefix_print(&addr, iaddr->plen);
-				ni_string_array_append(addrs, addr_str);
-				break;
-
-			default:
-				break;
-			}
-
-			if (p_lft)
-				ni_var_array_set_uint(p_lft,
-						addr_str,
-						iaddr->preferred_lft);
-			if (v_lft)
-				ni_var_array_set_uint(v_lft,
-						addr_str,
-						iaddr->valid_lft);
-		}
-	}
-
-	return addrs;
-}
-
 /*
  * Create/delete a dhcp6 request object
  */
@@ -1465,7 +1601,7 @@ ni_dhcp6_request_new(void)
 
 	/* Apply defaults */
 	req->enabled = TRUE; /* used by wickedd */
-	req->mode = NI_DHCP6_MODE_AUTO;
+	req->mode = NI_BIT(NI_DHCP6_MODE_AUTO);
 	req->rapid_commit = TRUE;
 
 	/* By default, we try to obtain all sorts of config from the server */
@@ -1483,7 +1619,7 @@ ni_dhcp6_request_free(ni_dhcp6_request_t *req)
 	if(req) {
 		ni_string_free(&req->hostname);
 		ni_string_free(&req->clientid);
-		ni_dhcp6_ia_list_destroy(&req->ia_list);
+		ni_dhcp6_prefix_req_list_destroy(&req->prefix_reqs);
 		ni_string_array_destroy(&req->request_options);
 		/*
 		 * req->vendor_class
@@ -1493,6 +1629,59 @@ ni_dhcp6_request_free(ni_dhcp6_request_t *req)
 	}
 }
 
+void
+ni_dhcp6_drop_request_init(ni_dhcp6_drop_request_t *req)
+{
+	ni_uuid_init(&req->uuid);
+	req->release = NI_TRISTATE_DEFAULT;
+}
+
+ni_dhcp6_prefix_req_t *
+ni_dhcp6_prefix_req_new(void)
+{
+	ni_dhcp6_prefix_req_t *req;
+
+	req = calloc(1, sizeof(*req));
+	return req;
+}
+
+void
+ni_dhcp6_prefix_req_free(ni_dhcp6_prefix_req_t *req)
+{
+	if (req) {
+		ni_dhcp6_ia_addr_list_destroy(&req->hints);
+#if 0
+		ni_netdev_ref_destroy(&req->device);
+#endif
+		free(req);
+	}
+}
+
+ni_bool_t
+ni_dhcp6_prefix_req_list_append(ni_dhcp6_prefix_req_t **list, ni_dhcp6_prefix_req_t *req)
+{
+	if (list && req) {
+		while (*list)
+			list = &(*list)->next;
+		*list = req;
+		return TRUE;
+	}
+	return FALSE;
+}
+
+void
+ni_dhcp6_prefix_req_list_destroy(ni_dhcp6_prefix_req_t **list)
+{
+	ni_dhcp6_prefix_req_t *req;
+
+	if (list) {
+		while ((req = *list)) {
+			*list = req->next;
+			ni_dhcp6_prefix_req_free(req);
+		}
+	}
+}
+
 ni_bool_t
 ni_dhcp6_supported(const ni_netdev_t *ifp)
 {
@@ -1501,6 +1690,10 @@ ni_dhcp6_supported(const ni_netdev_t *ifp)
 	 * we've simply did not tested it on other links ...
 	 */
 	switch (ifp->link.hwaddr.type) {
+	case ARPHRD_PPP:
+		break;
+	case ARPHRD_NONE:
+		break;
 	case ARPHRD_ETHER:
 	case ARPHRD_INFINIBAND:
 		if (ifp->link.masterdev.index) {
