@@ -266,7 +266,6 @@ __ni_netdev_process_events(ni_netconfig_t *nc, ni_netdev_t *dev, unsigned int ol
 int
 __ni_rtevent_newlink(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struct nlmsghdr *h)
 {
-	char namebuf[IF_NAMESIZE+1] = {'\0'};
 	ni_netdev_t *dev, *old;
 	struct ifinfomsg *ifi;
 	struct nlattr *nla;
@@ -280,7 +279,7 @@ __ni_rtevent_newlink(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struc
 		return 0;
 
 	old = ni_netdev_by_index(nc, ifi->ifi_index);
-	ifname = if_indextoname(ifi->ifi_index, namebuf);
+	ni_netdev_index_to_name(&ifname, ifi->ifi_index);
 	if (!ifname) {
 		/*
 		 * device (index) does not exists any more;
@@ -307,25 +306,28 @@ __ni_rtevent_newlink(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struc
 		}
 		dev = old;
 		old_flags = old->link.ifflags;
+		ni_string_free(&ifname);
 	} else {
 		if (!(dev = ni_netdev_new(ifname, ifi->ifi_index))) {
 			ni_warn("%s[%u]: unable to allocate memory for device",
 					ifname, ifi->ifi_index);
+			ni_string_free(&ifname);
 			return -1;
 		}
 		dev->created = 1;
 		ni_netconfig_device_append(nc, dev);
+		ni_string_free(&ifname);
 	}
 
 	if (__ni_netdev_process_newlink(dev, h, ifi, nc) < 0) {
-		ni_error("Problem parsing RTM_NEWLINK message for %s", ifname);
+		ni_error("Problem parsing RTM_NEWLINK message for %s", dev->name);
 		return -1;
 	}
 
-	if ((ifname = dev->name)) {
+	if (dev->name) {
 		ni_netdev_t *conflict;
 
-		conflict = ni_netdev_by_name(nc, ifname);
+		conflict = ni_netdev_by_name(nc, dev->name);
 		if (conflict && conflict->link.ifindex != (unsigned int)ifi->ifi_index) {
 			/*
 			 * As the events often provide an already obsolete name [2 events,
@@ -340,9 +342,7 @@ __ni_rtevent_newlink(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struc
 			 * Just update the name of the conflicting device in advance too
 			 * and when the interface does not exist any more, emit events.
 			 */
-			char *current = if_indextoname(conflict->link.ifindex, namebuf);
-			if (current) {
-				ni_string_dup(&conflict->name, current);
+			if (ni_netdev_index_to_name(&conflict->name, conflict->link.ifindex)) {
 				__ni_netdev_event(nc, conflict, NI_EVENT_DEVICE_RENAME);
 			} else {
 				unsigned int ifflags = conflict->link.ifflags;
@@ -394,7 +394,7 @@ __ni_rtevent_dellink(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struc
 	} else {
 		unsigned int old_flags = dev->link.ifflags;
 
-		dev->link.ifflags = __ni_netdev_translate_ifflags(ifi->ifi_flags, old_flags);
+		dev->link.ifflags = __ni_netdev_translate_ifflags(dev->name, ifi->ifi_flags, old_flags);
 		dev->deleted = 1;
 		__ni_netdev_process_events(nc, dev, old_flags);
 		ni_client_state_drop(dev->link.ifindex);
@@ -512,7 +512,7 @@ __ni_rtevent_deladdr(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struc
 	if (dev == NULL)
 		return 0;
 
-	if (__ni_rtnl_parse_newaddr(dev->link.ifflags, h, ifa, &tmp) < 0) {
+	if (__ni_rtnl_parse_newaddr(dev->name, dev->link.ifflags, h, ifa, &tmp) < 0) {
 		ni_error("Problem parsing %s message for %s", dev->name,
 				ni_rtnl_msg_type_to_name(h->nlmsg_type, NULL));
 		return -1;
@@ -520,7 +520,7 @@ __ni_rtevent_deladdr(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, struc
 
 	/* Remove the address when we track it */
 	if ((ap = ni_address_list_find(dev->addrs, &tmp.local_addr)) != NULL)
-		__ni_address_list_remove(&dev->addrs, ap);
+		ni_address_list_delete(&dev->addrs, ap);
 
 	/* Tentative IPv6 addresses are not exposed via NEWADDR events,
 	 * but in manuall address lookup / dump only.
@@ -813,17 +813,17 @@ __ni_rtevent_process_nd_radv_opts(ni_netdev_t *dev, const struct nd_opt_hdr *opt
 		size_t opt_len;
 
 		if (len < 2) {
-			ni_error("%s: nd user option length too short", dev->name);
+			ni_error("%s: ND user option length too short", dev->name);
 			return -1;
 		}
 
 		opt_len = (opt->nd_opt_len << 3);
 		if (opt_len == 0) {
-			ni_error("%s: zero length nd user option", dev->name);
+			ni_error("%s: zero length ND user option", dev->name);
 			return -1;
 		}
 		else if (opt_len > len) {
-			ni_error("%s: nd user option length exceeds total length",
+			ni_error("%s: ND user option length exceeds total length",
 				dev->name);
 			return -1;
 		}
@@ -848,7 +848,7 @@ __ni_rtevent_process_nd_radv_opts(ni_netdev_t *dev, const struct nd_opt_hdr *opt
 		default:
 			/* kernels up to at least 3.4 do not provide other */
 			ni_debug_verbose(NI_LOG_DEBUG1, NI_TRACE_IPV6|NI_TRACE_EVENTS,
-					"%s: unhandled nd user option %d",
+					"%s: unhandled ND user option %d",
 					dev->name, opt->nd_opt_type);
 		break;
 		}
@@ -871,7 +871,7 @@ __ni_rtevent_nduseropt(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, str
 
 	dev = ni_netdev_by_index(nc, msg->nduseropt_ifindex);
 	if (!dev) {
-		ni_debug_events("ipv6 nd user option event for unknown device index: %u",
+		ni_debug_events("ipv6 ND user option event for unknown device index: %u",
 				msg->nduseropt_ifindex);
 		return 0;
 	}
@@ -879,7 +879,7 @@ __ni_rtevent_nduseropt(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, str
 	if (msg->nduseropt_icmp_type != ND_ROUTER_ADVERT ||
 	    msg->nduseropt_icmp_code != 0 ||
 	    msg->nduseropt_family    != AF_INET6) {
-		ni_debug_events("%s: unknown rtnetlink nd user option message"
+		ni_debug_events("%s: unknown rtnetlink ND user option message"
 				" type %d, code %d, family %d", dev->name,
 				msg->nduseropt_icmp_type,
 				msg->nduseropt_icmp_code,
@@ -888,7 +888,7 @@ __ni_rtevent_nduseropt(ni_netconfig_t *nc, const struct sockaddr_nl *nladdr, str
 	}
 
 	if (!nlmsg_valid_hdr(h, sizeof(struct nduseroptmsg) + msg->nduseropt_opts_len)) {
-		ni_debug_events("%s: invalid rtnetlink nd user radv option length %d",
+		ni_debug_events("%s: invalid rtnetlink ND user radv option length %d",
 				dev->name, msg->nduseropt_opts_len);
 		return -1;
 	}
@@ -1129,22 +1129,26 @@ __ni_rtevent_sock_open(void)
 static ni_bool_t
 __ni_rtevent_restart(ni_socket_t *sock)
 {
-	ni_rtevent_handle_t *handle = sock->user_data;
-	if (handle) {
-		if ((__ni_rtevent_sock = __ni_rtevent_sock_open())) {
-			const ni_uint_array_t *groups = &handle->groups;
-			unsigned int i;
+	ni_rtevent_handle_t *handle;
+	ni_bool_t ret = FALSE;
 
-			handle = __ni_rtevent_sock->user_data;
-			for (i = 0; i < groups->count; ++i) {
-				__ni_rtevent_join_group(handle, groups->data[i]);
-			}
-			ni_socket_activate(__ni_rtevent_sock);
-			return TRUE;
+	ni_assert(sock == __ni_rtevent_sock);
+
+	handle = sock->user_data;
+	if ((__ni_rtevent_sock = __ni_rtevent_sock_open())) {
+		const ni_uint_array_t *groups = &handle->groups;
+		unsigned int i;
+
+		handle = __ni_rtevent_sock->user_data;
+		for (i = 0; i < groups->count; ++i) {
+			__ni_rtevent_join_group(handle, groups->data[i]);
 		}
-		ni_socket_release(sock);
+		ni_socket_activate(__ni_rtevent_sock);
+		ret = TRUE;
 	}
-	return FALSE;
+
+	ni_socket_close(sock);
+	return ret;
 }
 
 /*
@@ -1321,13 +1325,13 @@ ni_server_enable_interface_nduseropt_events(void (*ifnduseropt_handler)(ni_netde
 	ni_rtevent_handle_t *handle;
 
 	if (!__ni_rtevent_sock || ni_global.interface_nduseropt_event) {
-		ni_error("Interface ND user opt event handler already set");
+		ni_error("Interface ND user option event handler already set");
 		return -1;
 	}
 
 	handle = __ni_rtevent_sock->user_data;
 	if (!__ni_rtevent_join_group(handle, RTNLGRP_ND_USEROPT)) {
-		ni_error("Cannot add rtnetlink nd user opt event membership: %m");
+		ni_error("Cannot add rtnetlink ND user option event membership: %m");
 		return -1;
 	}
 	ni_global.interface_nduseropt_event = ifnduseropt_handler;
