@@ -11,7 +11,7 @@
 
 #include <wicked/netinfo.h>
 #include <wicked/logging.h>
-#include <wicked/socket.h>
+#include <wicked/time.h>
 #include <wicked/ipv6.h>
 #include <errno.h>
 
@@ -62,7 +62,26 @@ enum {
 	NI_IPV6_DEVCONF_MLDV1_UNSOLICITED_REPORT_INTERVAL,
 	NI_IPV6_DEVCONF_MLDV2_UNSOLICITED_REPORT_INTERVAL,
 	NI_IPV6_DEVCONF_SUPPRESS_FRAG_NDISC,
-	__NI_IPV6_DEVCONF_MAX
+	NI_IPV6_DEVCONF_ACCEPT_RA_FROM_LOCAL,
+	NI_IPV6_DEVCONF_USE_OPTIMISTIC,
+	NI_IPV6_DEVCONF_ACCEPT_RA_MTU,
+	NI_IPV6_DEVCONF_STABLE_SECRET,
+	NI_IPV6_DEVCONF_USE_OIF_ADDRS_ONLY,
+	NI_IPV6_DEVCONF_ACCEPT_RA_MIN_HOP_LIMIT,
+	NI_IPV6_DEVCONF_IGNORE_ROUTES_WITH_LINKDOWN,
+	NI_IPV6_DEVCONF_DROP_UNICAST_IN_L2_MULTICAST,
+	NI_IPV6_DEVCONF_DROP_UNSOLICITED_NA,
+	NI_IPV6_DEVCONF_KEEP_ADDR_ON_DOWN,
+	NI_IPV6_DEVCONF_RTR_SOLICIT_MAX_INTERVAL,
+	NI_IPV6_DEVCONF_SEG6_ENABLED,
+	NI_IPV6_DEVCONF_SEG6_REQUIRE_HMAC,
+	NI_IPV6_DEVCONF_ENHANCED_DAD,
+	NI_IPV6_DEVCONF_ADDR_GEN_MODE,
+	NI_IPV6_DEVCONF_DISABLE_POLICY,
+	NI_IPV6_DEVCONF_ACCEPT_RA_RT_INFO_MIN_PLEN,
+	NI_IPV6_DEVCONF_NDISC_TCLASS,
+
+	NI_IPV6_DEVCONF_MAX
 };
 
 /*
@@ -102,7 +121,26 @@ static const ni_intmap_t	__ipv6_devconf_sysctl_name_map[] = {
 	{ "mldv1_unsolicited_report_interval",	NI_IPV6_DEVCONF_MLDV1_UNSOLICITED_REPORT_INTERVAL},
 	{ "mldv2_unsolicited_report_interval",	NI_IPV6_DEVCONF_MLDV2_UNSOLICITED_REPORT_INTERVAL},
 	{ "suppress_frag_ndisc",		NI_IPV6_DEVCONF_SUPPRESS_FRAG_NDISC	},
-	{ NULL,					__NI_IPV6_DEVCONF_MAX			},
+	{ "accept_ra_from_local",		NI_IPV6_DEVCONF_ACCEPT_RA_FROM_LOCAL	},
+	{ "use_optimistic",			NI_IPV6_DEVCONF_USE_OPTIMISTIC		},
+	{ "accept_ra_mtu",			NI_IPV6_DEVCONF_ACCEPT_RA_MTU		},
+	{ "stable_secret",			NI_IPV6_DEVCONF_STABLE_SECRET		},
+	{ "use_oif_addrs_only",			NI_IPV6_DEVCONF_USE_OIF_ADDRS_ONLY	},
+	{ "accept_ra_min_hop_limit",		NI_IPV6_DEVCONF_ACCEPT_RA_MIN_HOP_LIMIT	},
+	{ "ignore_routes_with_linkdown",	NI_IPV6_DEVCONF_IGNORE_ROUTES_WITH_LINKDOWN},
+	{ "drop_unicast_in_l2_multicast",	NI_IPV6_DEVCONF_DROP_UNICAST_IN_L2_MULTICAST},
+	{ "drop_unsolicited_na",		NI_IPV6_DEVCONF_DROP_UNSOLICITED_NA	},
+	{ "keep_addr_on_down",			NI_IPV6_DEVCONF_KEEP_ADDR_ON_DOWN	},
+	{ "router_solicitation_max_interval",	NI_IPV6_DEVCONF_RTR_SOLICIT_MAX_INTERVAL},
+	{ "seg6_enabled",			NI_IPV6_DEVCONF_SEG6_ENABLED		},
+	{ "seg6_require_hmac",			NI_IPV6_DEVCONF_SEG6_REQUIRE_HMAC	},
+	{ "enhanced_dad",			NI_IPV6_DEVCONF_ENHANCED_DAD		},
+	{ "addr_gen_mode",			NI_IPV6_DEVCONF_ADDR_GEN_MODE		},
+	{ "disable_policy",			NI_IPV6_DEVCONF_DISABLE_POLICY		},
+	{ "accept_ra_rt_info_min_plen",		NI_IPV6_DEVCONF_ACCEPT_RA_RT_INFO_MIN_PLEN},
+	{ "ndisc_tclass",			NI_IPV6_DEVCONF_NDISC_TCLASS		},
+
+	{ NULL,					NI_IPV6_DEVCONF_MAX			}
 };
 
 /*
@@ -128,6 +166,8 @@ ni_ipv6_devconf_reset(ni_ipv6_devconf_t *conf)
 	conf->accept_ra = NI_IPV6_ACCEPT_RA_DEFAULT;
 	conf->accept_dad = NI_IPV6_ACCEPT_DAD_DEFAULT;
 	conf->accept_redirects = NI_TRISTATE_DEFAULT;
+	conf->addr_gen_mode = NI_IPV6_ADDR_GEN_MODE_DEFAULT;
+	conf->stable_secret = in6addr_any;
 }
 
 /*
@@ -153,6 +193,24 @@ ni_netdev_get_ipv6(ni_netdev_t *dev)
 	return dev->ipv6;
 }
 
+ni_bool_t
+ni_netdev_ipv6_is_ready(const ni_netdev_t *dev)
+{
+	return dev && ni_ipv6_devinfo_is_ready(dev->ipv6);
+}
+
+ni_bool_t
+ni_netdev_ipv6_ra_received(const ni_netdev_t *dev)
+{
+	return dev && ni_ipv6_devinfo_ra_received(dev->ipv6);
+}
+
+ni_bool_t
+ni_netdev_ipv6_ra_requested(const ni_netdev_t *dev)
+{
+	return dev && ni_ipv6_devinfo_ra_requested(dev->ipv6);
+}
+
 void
 ni_netdev_set_ipv6(ni_netdev_t *dev, ni_ipv6_devconf_t *conf)
 {
@@ -184,64 +242,22 @@ ni_ipv6_devinfo_free(ni_ipv6_devinfo_t *ipv6)
 	free(ipv6);
 }
 
-/*
- * Discover current IPv6 device settings
- */
-int
-ni_system_ipv6_devinfo_get(ni_netdev_t *dev, ni_ipv6_devinfo_t *ipv6)
+ni_bool_t
+ni_ipv6_devinfo_is_ready(const ni_ipv6_devinfo_t *ipv6)
 {
-	if (ipv6 == NULL)
-		ipv6 = ni_netdev_get_ipv6(dev);
+	return ipv6 && (ipv6->flags & NI_BIT(NI_IPV6_READY));
+}
 
-	if (!ni_ipv6_supported()) {
-		ni_ipv6_devconf_reset(&ipv6->conf);
-		ni_ipv6_ra_info_reset(&ipv6->radv);
-		ipv6->conf.enabled = NI_TRISTATE_DISABLE;
-		return 0;
-	}
+ni_bool_t
+ni_ipv6_devinfo_ra_received(const ni_ipv6_devinfo_t *ipv6)
+{
+	return ipv6 && (ipv6->flags & NI_BIT(NI_IPV6_RA_RCVD));
+}
 
-	/*
-	 * dhcpcd does something very odd when shutting down an interface;
-	 * in addition to removing all IPv4 addresses, it also removes any
-	 * IPv6 addresses. The kernel seems to take this as "disable IPv6
-	 * on this interface", and subsequently, /proc/sys/ipv6/conf/<ifname>
-	 * is gone.
-	 * When we bring the interface back up, everything is fine; but until
-	 * then we need to ignore this glitch.
-	 */
-	if (ni_sysctl_ipv6_ifconfig_is_present(dev->name)) {
-		int val;
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "disable_ipv6", &val) >= 0)
-			ni_tristate_set(&ipv6->conf.enabled, !val);
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "forwarding", &val) >= 0)
-			ni_tristate_set(&ipv6->conf.forwarding, !!val);
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "autoconf", &val) >= 0)
-			ni_tristate_set(&ipv6->conf.autoconf, !!val);
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "use_tempaddr", &val) >= 0)
-			ipv6->conf.privacy = val < -1 ? -1 : (val > 2 ? 2 : val);
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "accept_ra", &val) >= 0)
-			ipv6->conf.accept_ra = val < 0 ? 0 : val > 2 ? 2 : val;
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "accept_dad", &val) >= 0)
-			ipv6->conf.accept_dad = val < 0 ? 0 : val > 2 ? 2 : val;
-
-		if (ni_sysctl_ipv6_ifconfig_get_int(dev->name, "accept_redirects", &val) >= 0)
-			ni_tristate_set(&ipv6->conf.accept_redirects, !!val);
-
-	} else {
-		ni_warn("%s: cannot get ipv6 device attributes", dev->name);
-
-		/* Reset to defaults */
-		ni_ipv6_devconf_reset(&ipv6->conf);
-		ni_ipv6_ra_info_reset(&ipv6->radv);
-	}
-
-	return 0;
+ni_bool_t
+ni_ipv6_devinfo_ra_requested(const ni_ipv6_devinfo_t *ipv6)
+{
+	return ipv6 && (ipv6->flags & NI_BIT(NI_IPV6_RS_SENT));
 }
 
 /*
@@ -271,12 +287,13 @@ __change_int(const char *ifname, const char *attr, int value)
 static ni_bool_t
 __tristate_changed(ni_tristate_t cfg, ni_tristate_t sys)
 {
-	return ni_tristate_is_set(cfg) && cfg != sys;
+	return ni_tristate_is_set(sys) && ni_tristate_is_set(cfg) && cfg != sys;
 }
 
 int
 ni_system_ipv6_devinfo_set(ni_netdev_t *dev, const ni_ipv6_devconf_t *conf)
 {
+	struct in6_addr stable_secret = in6addr_any;
 	ni_ipv6_devinfo_t *ipv6;
 	int ret;
 
@@ -359,6 +376,27 @@ ni_system_ipv6_devinfo_set(ni_netdev_t *dev, const ni_ipv6_devconf_t *conf)
 			ipv6->conf.accept_redirects = conf->accept_redirects;
 	}
 
+	if (__tristate_changed(conf->addr_gen_mode, ipv6->conf.addr_gen_mode)) {
+		ret = __change_int(dev->name, "addr_gen_mode", conf->addr_gen_mode);
+		if (ret < 0)
+			return ret;
+		if (ret == 0)
+			ipv6->conf.addr_gen_mode = conf->addr_gen_mode;
+	}
+
+	/* netlink omits stable_secret, but because it usually provides *
+	 * other sysctls, our sysfs get function (above) isn't called.  *
+	 * try avoid i/o errors due to not readable stable_secret.      */
+	if (ipv6->conf.addr_gen_mode == NI_IPV6_ADDR_GEN_MODE_STABLE_PRIVACY)
+		ni_sysctl_ipv6_ifconfig_get_ipv6(dev->name, "stable_secret", &stable_secret);
+	if (memcmp(&in6addr_any, &conf->stable_secret, sizeof(struct in6_addr)) ||
+	    (ipv6->conf.addr_gen_mode == NI_IPV6_ADDR_GEN_MODE_STABLE_PRIVACY &&
+	     memcmp(&stable_secret, &conf->stable_secret, sizeof(struct in6_addr)))) {
+		ret = ni_sysctl_ipv6_ifconfig_set_ipv6(dev->name, "stable_secret",
+				conf->stable_secret);
+		if (ret < 0)
+			return ret;
+	}
 	return 0;
 }
 
@@ -688,6 +726,19 @@ ni_ipv6_devconf_accept_dad_to_name(int accept_dad)
 	return ni_format_uint_mapped(accept_dad, __accept_dad_names);
 }
 
+const char *
+ni_ipv6_devconf_addr_gen_mode_to_name(int addr_gen_mode)
+{
+	static const ni_intmap_t	addr_gen_mode_map[] = {
+		{ "eui64",		NI_IPV6_ADDR_GEN_MODE_EUI64		},
+		{ "none",		NI_IPV6_ADDR_GEN_MODE_NONE		},
+		{ "stable-privacy",	NI_IPV6_ADDR_GEN_MODE_STABLE_PRIVACY	},
+		{ "random",		NI_IPV6_ADDR_GEN_MODE_RANDOM		},
+		{ NULL,			NI_IPV6_ADDR_GEN_MODE_DEFAULT		}
+	};
+	return ni_format_uint_mapped(addr_gen_mode, addr_gen_mode_map);
+}
+
 static inline const char *
 ni_ipv6_devconf_flag_to_sysctl_name(unsigned int flag)
 {
@@ -715,7 +766,7 @@ __ni_ipv6_devconf_process_flag(ni_netdev_t *dev, unsigned int flag, int value)
 		ipv6->conf.accept_ra = value < 0 ? 0 : value > 2 ? 2 : value;
 		break;
 	case NI_IPV6_DEVCONF_ACCEPT_DAD:
-		ipv6->conf.accept_dad = value < 0 ? 0 : value > 2 ? 2 : value;
+		ipv6->conf.accept_dad = value < -1 ? -1 : value > 2 ? 2 : value;
 		break;
 	case NI_IPV6_DEVCONF_AUTOCONF:
 		ipv6->conf.autoconf = !!value;
@@ -723,6 +774,13 @@ __ni_ipv6_devconf_process_flag(ni_netdev_t *dev, unsigned int flag, int value)
 	case NI_IPV6_DEVCONF_USE_TEMPADDR:
 		ipv6->conf.privacy = value < -1 ? -1 : value > 2 ? 2 : value;
 		break;
+	case NI_IPV6_DEVCONF_ADDR_GEN_MODE:
+		ipv6->conf.addr_gen_mode = value < 0 ? 0 : value;
+		break;
+	case NI_IPV6_DEVCONF_STABLE_SECRET:
+		/* omitted via netlink using an int32_t array */
+		unused = 2;
+		break;
 	default:
 		/* TODO: handle more (all) of them */
 		unused = 1;
@@ -730,7 +788,7 @@ __ni_ipv6_devconf_process_flag(ni_netdev_t *dev, unsigned int flag, int value)
 	}
 
 	level = NI_LOG_DEBUG1 + unused;
-	if (ni_debug_guard(level, NI_TRACE_EVENTS|NI_TRACE_IPV6)) {
+	if (unused < 2 && ni_debug_guard(level, NI_TRACE_EVENTS|NI_TRACE_IPV6)) {
 		const char *name;
 
 		name = ni_ipv6_devconf_flag_to_sysctl_name(flag);
