@@ -34,6 +34,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <inttypes.h>
+#include <netlink/netlink.h>
 
 #include <wicked/netinfo.h>
 #include <wicked/addrconf.h>
@@ -188,17 +189,29 @@ ni_addrconf_lease_routes_data_to_xml(const ni_addrconf_lease_t *lease, xml_node_
 	ni_route_t *rp;
 	unsigned int count = 0;
 	unsigned int i;
-
 	(void)ifname;
-	/* A very limitted view */
+
+	/*
+	 * A very limitted view as needed by dhcp4 (+auto6) leases.
+	 * We don't write or read static lease (config) at all and
+	 * don't support all the routing details, but only unicast
+	 * routes in table main with required details.
+	 */
 	for (tab = lease->routes; tab; tab = tab->next) {
-		if (tab->tid != 254) /* RT_TABLE_MAIN for now */
+		if (tab->tid == RT_TABLE_UNSPEC)
 			continue;
 
 		for (i = 0; i < tab->routes.count; ++i) {
 			if (!(rp = tab->routes.data[i]))
 				continue;
 
+			if (rp->family != lease->family)
+				continue;
+			if (rp->type != RTN_UNICAST)
+				continue;
+			if (rp->table != RT_TABLE_MAIN)
+				continue;
+
 			route = xml_node_new("route", NULL);
 			if (ni_sockaddr_is_specified(&rp->destination)) {
 				xml_node_new_element("destination", route,
@@ -213,6 +226,14 @@ ni_addrconf_lease_routes_data_to_xml(const ni_addrconf_lease_t *lease, xml_node_
 				xml_node_new_element("gateway", hop,
 					ni_sockaddr_print(&nh->gateway));
 			}
+			if (rp->priority > 0) {
+				xml_node_new_element_uint("priority", route,
+					rp->priority);
+			}
+			if (ni_sockaddr_is_specified(&rp->pref_src)) {
+				xml_node_new_element("pref-source", route,
+					ni_sockaddr_print(&rp->pref_src));
+			}
 			if (route->children) {
 				xml_node_add_child(node, route);
 				count++;
@@ -256,6 +277,7 @@ ni_addrconf_lease_nis_data_to_xml(const ni_addrconf_lease_t *lease, xml_node_t *
 	unsigned int i, j;
 	ni_nis_info_t *nis;
 	xml_node_t *data;
+	const char *name;
 
 	(void)ifname;
 
@@ -266,46 +288,53 @@ ni_addrconf_lease_nis_data_to_xml(const ni_addrconf_lease_t *lease, xml_node_t *
 	/* Default domain */
 	data = xml_node_new("default", NULL);
 	if (!ni_string_empty(nis->domainname)) {
-		count++;
-		xml_node_new_element("domain", data, nis->domainname);
-	}
-	if (nis->default_binding == NI_NISCONF_BROADCAST ||
-	    nis->default_binding == NI_NISCONF_STATIC) {
-		/* no SLP here */
-		count++;
-		xml_node_new_element("binding", data,
-			ni_nis_binding_type_to_name(nis->default_binding));
-	}
-	/* Only in when static binding? */
-	for (i = 0; i < nis->default_servers.count; ++i) {
-		const char *server = nis->default_servers.data[i];
-		if (ni_string_empty(server))
-			continue;
-		count++;
-		xml_node_new_element("server", data, server);
+		if (xml_node_new_element("domain", data, nis->domainname))
+			count++;
+
+		if (nis->default_binding == NI_NISCONF_BROADCAST ||
+		    nis->default_binding == NI_NISCONF_STATIC) {
+			/* no SLP here */
+			name = ni_nis_binding_type_to_name(nis->default_binding);
+			if (name && xml_node_new_element("binding", data, name))
+				count++;
+		}
+		/* Only in when static binding? */
+		for (i = 0; i < nis->default_servers.count; ++i) {
+			const char *server = nis->default_servers.data[i];
+			if (ni_string_empty(server))
+				continue;
+			if (xml_node_new_element("server", data, server))
+				count++;
+		}
 	}
 	if (count) {
 		xml_node_add_child(node, data);
+	} else {
+		xml_node_free(data);
 	}
 
-	/* Further domains */
+	/* Further domains with servers */
 	for (i = 0; i < nis->domains.count; ++i) {
 		ni_nis_domain_t *dom = nis->domains.data[i];
 		if (!dom || ni_string_empty(dom->domainname))
 			continue;
 
-		count++;
-		data = xml_node_new("domain", node);
-		xml_node_new_element("domain", data, dom->domainname);
-		if (ni_nis_binding_type_to_name(nis->default_binding)) {
-			xml_node_new_element("binding", data,
-				ni_nis_binding_type_to_name(nis->default_binding));
-		}
+		if (!(data = xml_node_new("domain", node)))
+			continue;
+
+		if (xml_node_new_element("domain", data, dom->domainname))
+			count++;
+
+		name = ni_nis_binding_type_to_name(dom->binding);
+		if (name && xml_node_new_element("binding", data, name))
+			count++;
+
 		for (j = 0; j < dom->servers.count; ++j) {
 			const char *server = dom->servers.data[j];
 			if (ni_string_empty(server))
 				continue;
-			xml_node_new_element("server", data, server);
+			if (xml_node_new_element("server", data, server))
+				count++;
 		}
 	}
 
@@ -482,6 +511,7 @@ ni_addrconf_lease_opts_data_to_xml(const ni_addrconf_lease_t *lease, xml_node_t
 static int
 __ni_addrconf_lease_info_to_xml(const ni_addrconf_lease_t *lease, xml_node_t *node)
 {
+	struct timeval acquired;
 	char buf[32] = { '\0' };
 
 	xml_node_new_element("family", node, ni_addrfamily_type_to_name(lease->family));
@@ -492,7 +522,8 @@ __ni_addrconf_lease_info_to_xml(const ni_addrconf_lease_t *lease, xml_node_t *no
 		xml_node_new_element("uuid", node, ni_uuid_print(&lease->uuid));
 	xml_node_new_element("state", node, ni_addrconf_state_to_name(lease->state));
 
-	snprintf(buf, sizeof(buf), "%"PRId64, (int64_t)lease->acquired.tv_sec);
+	ni_time_timer_to_real(&lease->acquired, &acquired);
+	snprintf(buf, sizeof(buf), "%"PRId64, (int64_t)acquired.tv_sec);
 	xml_node_new_element("acquired", node, buf);
 
 	snprintf(buf, sizeof(buf), "0x%08x", lease->update);
@@ -616,6 +647,7 @@ __ni_addrconf_lease_info_from_xml(ni_addrconf_lease_t *lease, const xml_node_t *
 	if (!lease || !node)
 		return -1;
 
+	ni_timer_get_time(&lease->acquired); /* pre-init */
 	for (child = node->children; child; child = child->next) {
 		if (ni_string_eq(child->name, "family")) {
 			if ((value = ni_addrfamily_name_to_type(child->cdata)) == -1)
@@ -645,13 +677,15 @@ __ni_addrconf_lease_info_from_xml(ni_addrconf_lease_t *lease, const xml_node_t *
 			update = TRUE;
 		}
 		if (ni_string_eq(child->name, "acquired")) {
-			int64_t acquired;
+			struct timeval acquired;
+			int64_t sec;
 
-			if (ni_parse_int64(child->cdata, &acquired, 10))
+			if (ni_parse_int64(child->cdata, &sec, 10))
 				return -1;
 
-			lease->acquired.tv_sec = acquired;
-			lease->acquired.tv_usec = 0;
+			acquired.tv_sec = sec;
+			acquired.tv_usec = 0;
+			ni_time_real_to_timer(&acquired, &lease->acquired);
 		}
 	}
 	if (!update)
@@ -729,7 +763,7 @@ __ni_addrconf_lease_addr_from_xml(ni_address_t **ap_list, unsigned int family, c
 	    (family == AF_INET6 && plen > 128))
 		return -1;
 
-	if (!(ap = ni_address_new(family, plen, &addr, NULL)))
+	if (!(ap = ni_address_create(family, plen, &addr, NULL)))
 		return -1;
 
 	if ((child = xml_node_get_child(node, "peer"))) {
@@ -793,14 +827,12 @@ static int
 __ni_addrconf_lease_route_nh_from_xml(ni_route_t *rp, const xml_node_t *node)
 {
 	const xml_node_t *child;
-	ni_route_nexthop_t *nh = NULL;
+	ni_route_nexthop_t *nh = &rp->nh;
 	ni_sockaddr_t addr;
 
 	for (child = node->children; child; child = child->next) {
 		if (ni_string_eq(child->name, "gateway") && child->cdata) {
-			if (ni_sockaddr_parse(&addr, child->cdata, AF_UNSPEC) != 0)
-				return -1;
-			if (rp->family != addr.ss_family)
+			if (ni_sockaddr_parse(&addr, child->cdata, rp->family))
 				return -1;
 			if (nh == NULL) {
 				nh = ni_route_nexthop_new();
@@ -819,24 +851,28 @@ static int
 __ni_addrconf_lease_route_from_xml(ni_route_t *rp, const xml_node_t *node)
 {
 	const xml_node_t *child;
-	ni_sockaddr_t addr;
-	unsigned int plen;
 
 	for (child = node->children; child; child = child->next) {
 		if (ni_string_eq(child->name, "destination") && child->cdata) {
-			if (!ni_sockaddr_prefix_parse(child->cdata, &addr, &plen))
+			if (!ni_sockaddr_prefix_parse(child->cdata,
+					&rp->destination, &rp->prefixlen))
 				return -1;
-			if (rp->family != addr.ss_family)
+			if (rp->family != rp->destination.ss_family)
 				return -1;
-			if ((rp->family == AF_INET  && plen > 32) ||
-			    (rp->family == AF_INET6 && plen > 128))
+			if (rp->prefixlen > ni_af_address_prefixlen(rp->family))
 				return -1;
-			rp->destination = addr;
-			rp->prefixlen = plen;
 		} else
 		if (ni_string_eq(child->name, "nexthop")) {
 			if (__ni_addrconf_lease_route_nh_from_xml(rp, child) != 0)
 				return -1;
+		} else
+		if (ni_string_eq(child->name, "priority")) {
+			if (ni_parse_uint(child->cdata, &rp->priority, 10))
+				return -1;
+		} else
+		if (ni_string_eq(child->name, "pref-source")) {
+			if (ni_sockaddr_parse(&rp->pref_src, child->cdata, rp->family))
+				return -1;
 		}
 	}
 	return 0;
@@ -847,15 +883,25 @@ ni_addrconf_lease_routes_data_from_xml(ni_addrconf_lease_t *lease, const xml_nod
 {
 	const xml_node_t *child;
 	ni_route_t *rp;
-
 	(void)ifname;
+
+	/*
+	 * A very limitted view as needed by dhcp4 (+auto6) leases.
+	 * We don't write or read static lease (config) at all and
+	 * don't support all the routing details, but only unicast
+	 * routes in table main with required details.
+	 */
 	for (child = node->children; child; child = child->next) {
 		if (ni_string_eq(child->name, "route")) {
 			rp = ni_route_new();
 			if (!rp)
 				return -1;
+
+			rp->type   = RTN_UNICAST;
+			rp->table  = RT_TABLE_MAIN;
 			rp->family = lease->family;
-			rp->table = ni_route_guess_table(rp);
+			/* route without destination node is a default route */
+			rp->destination.ss_family = lease->family;
 			if (__ni_addrconf_lease_route_from_xml(rp, child) != 0) {
 				ni_route_free(rp);
 			} else
@@ -1313,6 +1359,7 @@ ni_addrconf_lease_file_write(const char *ifname, ni_addrconf_lease_t *lease)
 	ni_debug_dhcp("Writing lease to temporary file for '%s'", filename);
 	xml_node_print(xml, fp);
 	fclose(fp);
+	fp = NULL;
 	xml_node_free(xml);
 
 	if ((ret = rename(tempname, filename)) != 0) {
