@@ -24,11 +24,8 @@
 #include <net/ethernet.h>
 #include <stdarg.h>
 
-#if defined(HAVE_DCB_ATTR_IEEE_MAXRATE) && defined(HAVE_LINUX_DCBNL_H)
-#  include <linux/dcbnl.h>
-#else
-#  include "linux/dcbnl.h"
-#endif
+#include <linux/dcbnl.h>
+
 #include "buffer.h"
 #include "util_priv.h"
 #include "debug.h"
@@ -73,7 +70,7 @@ struct ni_lldp_agent {
 
 struct ni_lldp_peer {
 	ni_lldp_peer_t *	next;
-	time_t			expires;
+	struct timeval		expires;
 	ni_lldp_t *		data;
 	unsigned int		raw_id_len;
 	unsigned char		raw_id[0];
@@ -476,7 +473,8 @@ ni_lldp_agent_start(ni_netdev_t *dev, ni_lldp_t *lldp, ni_dcbx_state_t *dcbx)
 		if (ni_capture_devinfo_init(&devinfo, dev->name, &dev->link) < 0)
 			return -1;
 
-		capture = ni_capture_open(&devinfo, &protinfo, ni_lldp_receive);
+		capture = ni_capture_open(&devinfo, &protinfo, ni_lldp_receive, "lldp");
+		ni_capture_devinfo_destroy(&devinfo);
 	}
 	agent->capture = capture;
 
@@ -536,7 +534,7 @@ ni_lldp_agent_send(ni_lldp_agent_t *agent)
 	if (agent->txCredit > 0) {
 		ni_buffer_t *bp = &agent->sendbuf;
 
-		ni_debug_lldp("%s: sending LLDP packet (PDU len=%u)", agent->dev->name, ni_buffer_count(bp));
+		ni_debug_lldp("%s: sending LLDP packet (PDU len=%zu)", agent->dev->name, ni_buffer_count(bp));
 		/* ni_debug_lldp(PDU=%s", ni_print_hex(ni_buffer_head(bp), ni_buffer_count(bp))); */
 		ni_capture_send(agent->capture, &agent->sendbuf, NULL);
 		agent->txCredit--;
@@ -639,15 +637,15 @@ ni_lldp_agent_update(ni_lldp_agent_t *agent, ni_lldp_t *lldp, const void *raw_id
 {
 	ni_lldp_peer_t **pos, *peer, *found = NULL;
 	unsigned int npeers = 0;
-	time_t now;
+	struct timeval now;
 
-	now = time(NULL);
+	ni_timer_get_time(&now);
 
 	pos = &agent->peers;
 
 	/* First, expire any old entries. Note that peers are sorted by order
 	 * of increasing expiry timeout. */
-	while ((peer = *pos) != NULL && peer->expires <= now)
+	while ((peer = *pos) != NULL && timercmp(&peer->expires, &now, <=))
 		ni_lldp_peer_unlink_and_free(pos);
 
 	while ((peer = *pos) != NULL) {
@@ -683,12 +681,13 @@ ni_lldp_agent_update(ni_lldp_agent_t *agent, ni_lldp_t *lldp, const void *raw_id
 	}
 
 	/* Update/init the peer info */
-	found->expires = now + lldp->ttl;
+	found->expires = now;
+	found->expires.tv_sec += lldp->ttl;
 	found->data = lldp;
 
 	/* Insert in order of increasing timeout */
 	pos = &agent->peers;
-	while ((peer = *pos) != NULL && peer->expires < found->expires)
+	while ((peer = *pos) != NULL && timercmp(&peer->expires, &found->expires, <))
 		*pos = peer->next;
 
 	found->next = *pos;
@@ -731,7 +730,7 @@ ni_lldp_receive(ni_socket_t *sock)
 
 	/* FIXME: we need to store the MAC address we received this packet from.
 	 * This is needed for DCBX tie-breaking among other things. */
-	if (ni_capture_recv(capture, &buf, &from, "lldp") >= 0) {
+	if (ni_capture_recv(capture, &buf, &from) >= 0) {
 		ni_lldp_agent_t *agent = ni_capture_get_user_data(capture);
 		ni_buffer_t raw_id_buf;
 		const void *raw_id;
@@ -934,7 +933,7 @@ ni_lldp_tlv_get(ni_buffer_t *bp, ni_buffer_t *vbuf)
 
 	type = head >> 9;
 	len = head & 0x1FF;
-	if (len > ni_buffer_count(bp))
+	if ((size_t)len > ni_buffer_count(bp))
 		return -1;
 
 	data = ni_buffer_pull_head(bp, len);
@@ -985,10 +984,10 @@ ni_lldp_tlv_get_uint32(ni_buffer_t *bp, uint32_t *var)
 static int
 ni_lldp_tlv_get_string(ni_buffer_t *bp, char **var)
 {
-	unsigned int len = ni_buffer_count(bp);
+	size_t len = ni_buffer_count(bp);
 	char *string;
 
-	if (!(string = malloc(len + 1)))
+	if (!len || !(string = malloc(len + 1)))
 		return -1;
 	memcpy(string, ni_buffer_head(bp), len);
 	string[len] = '\0';
@@ -1041,7 +1040,7 @@ ni_lldp_tlv_get_mac(ni_buffer_t *bp, ni_hwaddr_t *mac)
 	void *data;
 
 	if (!(data = ni_buffer_pull_head(bp, ETH_ALEN))) {
-		ni_debug_lldp("%s: bad MAC address length %u", __func__, ni_buffer_count(bp));
+		ni_debug_lldp("%s: bad MAC address length %zu", __func__, ni_buffer_count(bp));
 		return -1;
 	}
 
@@ -1086,7 +1085,7 @@ ni_lldp_tlv_get_netaddr(ni_buffer_t *bp, ni_sockaddr_t *ap)
 		return -1;
 	}
 
-	if (ni_buffer_count(bp) < len) {
+	if (ni_buffer_count(bp) < (size_t)len) {
 		ni_debug_lldp("%s: truncated network address (af %d, len %u)", __func__, af, len);
 		return -1;
 	}
