@@ -7,13 +7,14 @@
 #include "config.h"
 #endif
 
-#include <net/if_arp.h>
-
 #include <wicked/netinfo.h>
 #include <wicked/addrconf.h>
 #include <wicked/lldp.h>
 #include "kernel.h"
 
+#include <linux/if_arp.h>
+
+
 /*
  * Tristate is basically a bool + "unset, use defaults"
  */
@@ -50,7 +51,7 @@ ni_tristate_by_name(const char *name, ni_tristate_t *tristate)
 /*
  * Map interface link layer types to strings and vice versa
  */
-static const ni_intmap_t	__linktype_names[] = {
+static const ni_intmap_t	ni_linktype_name_map[] = {
 	{ "unknown",		NI_IFTYPE_UNKNOWN },
 	{ "loopback",		NI_IFTYPE_LOOPBACK },
 	{ "ethernet",		NI_IFTYPE_ETHERNET },
@@ -88,7 +89,7 @@ ni_linktype_name_to_type(const char *name)
 {
 	unsigned int value;
 
-	if (ni_parse_uint_mapped(name, __linktype_names, &value) < 0)
+	if (ni_parse_uint_mapped(name, ni_linktype_name_map, &value) < 0)
 		return -1;
 	return value;
 }
@@ -96,15 +97,17 @@ ni_linktype_name_to_type(const char *name)
 const char *
 ni_linktype_type_to_name(unsigned int type)
 {
-	return ni_format_uint_mapped(type, __linktype_names);
+	return ni_format_uint_mapped(type, ni_linktype_name_map);
 }
 
 /*
- * Map kernel kind to type
+ * Map kernel kind to wicked iftype
  *
- * Note: this is an initial, incomplete, one-direction match!
+ * Note: the kind string is what the kernel is using,
+ *       but it's not yet / 1:1 the final wicked type,
+ *       see e.g. tun & tap or ovs kind and types.
  */
-static const ni_intmap_t	__linkinfo_kind_names[] = {
+static const ni_intmap_t	ni_linkinfo_kind_map[] = {
 	{ "bridge",		NI_IFTYPE_BRIDGE },
 	{ "bond",		NI_IFTYPE_BOND },
 	{ "team",		NI_IFTYPE_TEAM },
@@ -119,20 +122,27 @@ static const ni_intmap_t	__linkinfo_kind_names[] = {
 	{ "gre",		NI_IFTYPE_GRE },
 	{ "openvswitch",	NI_IFTYPE_OVS_UNSPEC }, /* new in 4.4 kernels */
 
-	{ NULL }
+	{ NULL,			NI_IFTYPE_UNKNOWN }
 };
 
 ni_bool_t
-__ni_linkinfo_kind_to_type(const char *name, ni_iftype_t *iftype)
+ni_linkinfo_kind_to_type(const char *kind, ni_iftype_t *type)
 {
 	unsigned int value;
 
-	if (!iftype || ni_parse_uint_mapped(name, __linkinfo_kind_names, &value) < 0)
+	if (!type || ni_parse_uint_mapped(kind, ni_linkinfo_kind_map, &value) < 0)
 		return FALSE;
-	*iftype = value;
+
+	*type = value;
 	return TRUE;
 }
 
+const char *
+ni_linkinfo_type_to_kind(ni_iftype_t type)
+{
+	return ni_format_uint_mapped(type, ni_linkinfo_kind_map);
+}
+
 /*
  * Map addrconf name to type constant and vice versa
  */
@@ -226,16 +236,16 @@ ni_addrconf_flag_bit_set(unsigned int *mask, unsigned int flag, ni_bool_t enable
 {
 	if (mask) {
 		if (enable)
-			*mask |=  (1U << flag);
+			*mask |=  NI_BIT(flag);
 		else
-			*mask &= ~(1U << flag);
+			*mask &= ~NI_BIT(flag);
 	}
 }
 
 ni_bool_t
 ni_addrconf_flag_bit_is_set(unsigned int flags, unsigned int flag)
 {
-	return flags & (1 << flag);
+	return flags & NI_BIT(flag);
 }
 
 const char *
@@ -288,9 +298,9 @@ ni_addrconf_update_set(unsigned int *mask, unsigned int flag, ni_bool_t enable)
 {
 	if (mask) {
 		if (enable)
-			*mask |= (1 << flag);
+			*mask |= NI_BIT(flag);
 		else
-			*mask &= ~(1 << flag);
+			*mask &= ~NI_BIT(flag);
 	}
 }
 
@@ -394,26 +404,98 @@ ni_addrfamily_type_to_name(unsigned int type)
 /*
  * Map DHCP6 configuration modes
  */
-static const ni_intmap_t	__dhcp6_modes[] = {
+static const ni_intmap_t	ni_dhcp6_mode_names[] = {
 	{ "auto",		NI_DHCP6_MODE_AUTO	},
 	{ "info",		NI_DHCP6_MODE_INFO	},
+	{ "config",		NI_DHCP6_MODE_INFO	},
 	{ "managed",		NI_DHCP6_MODE_MANAGED	},
+	{ "address",		NI_DHCP6_MODE_MANAGED	},
+	{ "prefix",		NI_DHCP6_MODE_PREFIX	},
 
 	{ NULL,			NI_DHCP6_MODE_AUTO	}
 };
 
-const char *
-ni_dhcp6_mode_type_to_name(unsigned int type)
+const ni_intmap_t *
+ni_dhcp6_mode_map(void)
 {
-	return ni_format_uint_mapped(type, __dhcp6_modes);
+	return ni_dhcp6_mode_names;
 }
 
-int
-ni_dhcp6_mode_name_to_type(const char *name, unsigned int *type)
+const char *
+ni_dhcp6_mode_flag_to_name(ni_dhcp6_mode_t flag)
 {
-	return ni_parse_uint_mapped(name, __dhcp6_modes, type);
+	return ni_format_uint_mapped(flag, ni_dhcp6_mode_names);
 }
 
+ni_bool_t
+ni_dhcp6_mode_name_to_flag(const char *name, ni_dhcp6_mode_t *flag)
+{
+	return ni_parse_uint_mapped(name, ni_dhcp6_mode_names, flag) == 0;
+}
+
+unsigned int
+ni_dhcp6_mode_adjust(unsigned int mode)
+{
+	/* prefix + managed include info as well */
+	if  (mode & NI_BIT(NI_DHCP6_MODE_MANAGED)) {
+		mode &= ~NI_BIT(NI_DHCP6_MODE_INFO);
+		mode &= ~NI_BIT(NI_DHCP6_MODE_AUTO);
+	} else
+	if (mode & NI_BIT(NI_DHCP6_MODE_INFO)) {
+		mode &= ~NI_BIT(NI_DHCP6_MODE_AUTO);
+	}
+	if (mode & NI_BIT(NI_DHCP6_MODE_PREFIX)) {
+		mode &= ~NI_BIT(NI_DHCP6_MODE_INFO);
+	}
+
+	/* use auto when nothing specific given  */
+	if (!mode)
+		mode = NI_BIT(NI_DHCP6_MODE_AUTO);
+	else
+		mode &= NI_BIT(NI_DHCP6_MODE_AUTO) |
+			NI_BIT(NI_DHCP6_MODE_INFO) |
+			NI_BIT(NI_DHCP6_MODE_MANAGED) |
+			NI_BIT(NI_DHCP6_MODE_PREFIX);
+	return mode;
+}
+
+ni_bool_t
+ni_dhcp6_mode_parse(unsigned int *mode, const char *string)
+{
+	ni_string_array_t flags = NI_STRING_ARRAY_INIT;
+	unsigned int i, flag;
+
+	if (!mode)
+		return FALSE;
+
+	*mode = 0;
+
+	ni_string_split(&flags, string, ",|+ ", 0);
+	for (i = 0; i < flags.count; ++i) {
+		if (ni_dhcp6_mode_name_to_flag(flags.data[i], &flag))
+			*mode |= NI_BIT(flag);
+		else
+			goto failure;
+	}
+
+	*mode = ni_dhcp6_mode_adjust(*mode);
+
+	ni_string_array_destroy(&flags);
+	return TRUE;
+
+failure:
+	*mode = 0;
+	ni_string_array_destroy(&flags);
+	return FALSE;
+}
+
+const char *
+ni_dhcp6_mode_format(ni_stringbuf_t *buff, unsigned int mask, const char *sep)
+{
+	return ni_format_bitmap(buff, ni_dhcp6_mode_names, mask, sep ? sep : ",");
+}
+
+
 /* Map DHCP4 user-class formats */
 static const ni_intmap_t	__dhcp4_user_class_formats[] = {
 	{ "rfc3004",		NI_DHCP4_USER_CLASS_RFC3004	},
@@ -514,7 +596,7 @@ static const ni_intmap_t	__arphrd_names[] = {
  __ARPMAP(IEEE80211_PRISM,	ieee80211-prism),
  __ARPMAP(IEEE80211_RADIOTAP,	ieee80211-radiotap),
  __ARPMAP(IEEE802154,		ieee802154),
- __ARPMAP(IEEE802154_PHY,	ieee802154-monitor),
+ __ARPMAP(IEEE802154_MONITOR,	ieee802154-monitor),
  __ARPMAP(PHONET,		phonet),
  __ARPMAP(PHONET_PIPE,		phonet-pipe),
  __ARPMAP(CAIF,			caif),
@@ -618,7 +700,7 @@ ni_linkflags_bit_to_name(unsigned int bit)
 {
 	if (bit >= 32)
 		return NULL;
-	return ni_format_uint_mapped(1 << bit, __ni_linkifflag_names);
+	return ni_format_uint_mapped(NI_BIT(bit), __ni_linkifflag_names);
 }
 
 const char *
@@ -734,7 +816,7 @@ ni_lldp_system_capability_name_to_type(const char *name)
 }
 
 const char *
-ni_lldp_system_capability_type_to_name(unsigned int type)
+ni_lldp_system_capability_type_to_name(ni_lldp_destination_t type)
 {
 	return ni_format_uint_maybe_mapped(type, __ni_lldp_systemcap_names);
 }
