@@ -44,6 +44,39 @@
  * Constructor for network interface.
  * Takes interface name and ifindex.
  */
+ni_bool_t
+ni_linkinfo_init(ni_linkinfo_t *link)
+{
+	if (link) {
+		memset(link, 0, sizeof(*link));
+
+		ni_link_address_init(&link->hwaddr);
+		ni_link_address_init(&link->hwpeer);
+
+		ni_netdev_ref_init(&link->lowerdev);
+		ni_netdev_ref_init(&link->masterdev);
+
+		return TRUE;
+	}
+	return FALSE;
+}
+
+void
+ni_linkinfo_destroy(ni_linkinfo_t *link)
+{
+	if (link) {
+		ni_string_free(&link->kind);
+		ni_string_free(&link->alias);
+		ni_string_free(&link->qdisc);
+
+		ni_netdev_ref_destroy(&link->lowerdev);
+		ni_netdev_ref_destroy(&link->masterdev);
+		ni_netdev_port_info_destroy(&link->port);
+
+		ni_linkinfo_init(link);
+	}
+}
+
 ni_netdev_t *
 ni_netdev_new(const char *name, unsigned int index)
 {
@@ -54,13 +87,11 @@ ni_netdev_new(const char *name, unsigned int index)
 		return NULL;
 
 	dev->users = 1;
-	dev->link.type = NI_IFTYPE_UNKNOWN;
-	dev->link.hwaddr.type = ARPHRD_VOID;
-	dev->link.hwpeer.type = ARPHRD_VOID;
-	dev->link.ifindex = index;
 
-	if (name)
-		dev->name = xstrdup(name);
+	ni_linkinfo_init(&dev->link);
+
+	dev->link.ifindex = index;
+	ni_string_dup(&dev->name, name);
 
 	return dev;
 }
@@ -80,66 +111,136 @@ ni_netdev_clear_routes(ni_netdev_t *dev)
 	ni_route_tables_destroy(&dev->routes);
 }
 
-void
-ni_netdev_slaveinfo_destroy(ni_slaveinfo_t *slave)
+ni_bool_t
+ni_netdev_port_info_data_init(ni_netdev_port_info_t *info, ni_iftype_t type)
 {
-	switch (slave->type) {
-	case NI_IFTYPE_BOND:
-		ni_bonding_slave_info_free(slave->bond);
-		break;
-	default:
-		break;
+	if (info) {
+		switch (type) {
+		case NI_IFTYPE_BOND:
+			if (!(info->bond = ni_bonding_port_info_new()))
+				return FALSE;
+			break;
+
+		case NI_IFTYPE_TEAM:
+			if (!(info->team = ni_team_port_info_new()))
+				return FALSE;
+			break;
+
+		case NI_IFTYPE_BRIDGE:
+			if (!(info->bridge = ni_bridge_port_info_new()))
+				return FALSE;
+			break;
+
+		case NI_IFTYPE_OVS_BRIDGE:
+			if (!(info->ovsbr = ni_ovs_bridge_port_info_new()))
+				return FALSE;
+			break;
+
+		default:
+			info->bond = NULL; /* union ptr */
+			break;
+		}
+
+		info->type = type;
+		return TRUE;
 	}
-	free(slave->kind);
-	memset(slave, 0, sizeof(*slave));
+	return FALSE;
 }
 
-static void
-ni_netdev_free(ni_netdev_t *dev)
+void
+ni_netdev_port_info_data_destroy(ni_netdev_port_info_t *info)
 {
-	/* Clear out linkinfo */
-	ni_string_free(&dev->link.qdisc);
-	ni_string_free(&dev->link.kind);
-	ni_string_free(&dev->link.alias);
-	ni_netdev_ref_destroy(&dev->link.lowerdev);
-	ni_netdev_ref_destroy(&dev->link.masterdev);
-	ni_netdev_slaveinfo_destroy(&dev->link.slave);
-	ni_netdev_set_link_stats(dev, NULL);
+	if (info) {
+		switch (info->type) {
+		case NI_IFTYPE_BOND:
+			ni_bonding_port_info_free(info->bond);
+			break;
+
+		case NI_IFTYPE_TEAM:
+			ni_team_port_info_free(info->team);
+			break;
+
+		case NI_IFTYPE_BRIDGE:
+			ni_bridge_port_info_free(info->bridge);
+			break;
+
+		case NI_IFTYPE_OVS_BRIDGE:
+			ni_ovs_bridge_port_info_free(info->ovsbr);
+			break;
+
+		default:
+			break;
+		}
+		info->bond = NULL; /* union ptr */
+		info->type = NI_IFTYPE_UNKNOWN;
+	}
+}
+
+ni_bool_t
+ni_netdev_port_info_init(ni_netdev_port_info_t *info, const char *kind)
+{
+	if (info) {
+		ni_netdev_port_info_data_init(info, NI_IFTYPE_UNKNOWN);
+		return ni_string_dup(&info->kind, kind);
+	}
+	return FALSE;
+}
+
+void
+ni_netdev_port_info_destroy(ni_netdev_port_info_t *info)
+{
+	if (info) {
+		ni_netdev_port_info_data_destroy(info);
+		ni_string_free(&info->kind);
+	}
+}
+
+void
+ni_netdev_reset(ni_netdev_t *dev)
+{
+	if (!dev)
+		return;
+
+	/* Clear and Reset all (kernel) property data,
+	 * except of list, ref counter & marker flags */
+	ni_string_free(&dev->name);
+	ni_linkinfo_destroy(&dev->link);
+
+	ni_netdev_set_client_state(dev, NULL);
 
-	/* Clear out addresses, routes, ... */
 	ni_netdev_clear_addresses(dev);
 	ni_netdev_clear_routes(dev);
-	ni_netdev_set_ethernet(dev, NULL);
-	ni_netdev_set_infiniband(dev, NULL);
-	ni_netdev_set_bonding(dev, NULL);
-	ni_netdev_set_team(dev, NULL);
-	ni_netdev_set_bridge(dev, NULL);
-	ni_netdev_set_ovs_bridge(dev, NULL);
-	ni_netdev_set_vlan(dev, NULL);
-	ni_netdev_set_vxlan(dev, NULL);
-	ni_netdev_set_macvlan(dev, NULL);
-	ni_netdev_set_ipip(dev, NULL);
-	ni_netdev_set_sit(dev, NULL);
-	ni_netdev_set_gre(dev, NULL);
-	ni_netdev_set_wireless(dev, NULL);
-	ni_netdev_set_openvpn(dev, NULL);
-	ni_netdev_set_ppp(dev, NULL);
-	ni_netdev_set_dcb(dev, NULL);
-	ni_netdev_set_lldp(dev, NULL);
-	ni_netdev_set_client_state(dev, NULL);
 
 	ni_netdev_set_ipv4(dev, NULL);
 	ni_netdev_set_ipv6(dev, NULL);
 	ni_netdev_set_auto6(dev, NULL);
 
-	ni_netdev_set_pci(dev, NULL);
-	ni_netdev_set_ethtool(dev, NULL);
-	ni_netdev_clear_event_filters(dev);
-
 	ni_addrconf_lease_list_destroy(&dev->leases);
 
-	ni_string_free(&dev->name);
-	free(dev);
+	ni_netdev_set_team(dev, NULL);
+	ni_netdev_set_bonding(dev, NULL);
+	ni_netdev_set_bridge(dev, NULL);
+	ni_netdev_set_ovs_bridge(dev, NULL);
+	ni_netdev_set_ethernet(dev, NULL);
+	ni_netdev_set_infiniband(dev, NULL);
+	ni_netdev_set_vlan(dev, NULL);
+	ni_netdev_set_vxlan(dev, NULL);
+	ni_netdev_set_macvlan(dev, NULL);
+	ni_netdev_set_wireless(dev, NULL);
+	ni_netdev_set_openvpn(dev, NULL);
+	ni_netdev_set_tuntap(dev, NULL);
+
+	ni_netdev_set_sit(dev, NULL);
+	ni_netdev_set_ipip(dev, NULL);
+	ni_netdev_set_gre(dev, NULL);
+	ni_netdev_set_ppp(dev, NULL);
+	ni_netdev_set_lldp(dev, NULL);
+	ni_netdev_set_dcb(dev, NULL);
+
+	ni_netdev_set_pci(dev, NULL);
+	ni_netdev_set_ethtool(dev, NULL);
+
+	ni_netdev_clear_event_filters(dev);
 }
 
 /*
@@ -162,10 +263,12 @@ ni_netdev_put(ni_netdev_t *dev)
 		ni_assert(dev->users);
 		dev->users--;
 
-		if (dev->users == 0)
-			ni_netdev_free(dev);
-		else
+		if (dev->users == 0) {
+			ni_netdev_reset(dev);
+			free(dev);
+		} else {
 			return dev->users;
+		}
 	}
 	return 0;
 }
@@ -194,29 +297,6 @@ ni_netdev_device_is_ready(ni_netdev_t *dev)
 	return dev ? dev->link.ifflags & NI_IFF_DEVICE_READY : FALSE;
 }
 
-ni_tristate_t
-ni_netdev_guess_link_required(const ni_netdev_t *dev)
-{
-	ni_tristate_t link_required = NI_TRISTATE_DEFAULT;
-
-	switch (dev->link.type) {
-	case NI_IFTYPE_OVS_SYSTEM:
-	case NI_IFTYPE_TUN:
-	case NI_IFTYPE_TAP:
-		ni_tristate_set(&link_required, FALSE);
-		break;
-
-	case NI_IFTYPE_BRIDGE:
-		if (dev->bridge && dev->bridge->stp && !dev->bridge->ports.count)
-			ni_tristate_set(&link_required, FALSE);
-		break;
-
-	default:
-		break;
-	}
-	return link_required;
-}
-
 /*
  * Get the interface's VLAN information
  */
@@ -508,7 +588,7 @@ ni_netdev_get_wireless(ni_netdev_t *dev)
 	if (dev->link.type != NI_IFTYPE_WIRELESS)
 		return NULL;
 	if (!dev->wireless)
-		dev->wireless = ni_wireless_new(dev);
+		dev->wireless = ni_wireless_new();
 	return dev->wireless;
 }
 
@@ -578,17 +658,6 @@ ni_netdev_set_dcb(ni_netdev_t *dev, ni_dcb_t *dcb)
 	dev->dcb = dcb;
 }
 
-/*
- * Set the interface's link stats
- */
-void
-ni_netdev_set_link_stats(ni_netdev_t *dev, ni_link_stats_t *stats)
-{
-	if (dev->link.stats)
-		free(dev->link.stats);
-	dev->link.stats = stats;
-}
-
 /*
  * Set the PCI device info
  */
@@ -602,8 +671,8 @@ ni_netdev_set_pci(ni_netdev_t *dev, ni_pci_dev_t *pci_dev)
 
 /*
  * Set the interface's client_state structure.
- * This information is not intepreted by the server at all, but
- * we retain it for the client.
+ * This information is not interpreted by the server at all,#
+ * but we retain it for the client.
  */
 void
 ni_netdev_set_client_state(ni_netdev_t *dev, ni_client_state_t *client_state)
@@ -623,7 +692,7 @@ ni_netdev_get_client_state(ni_netdev_t *dev)
 		return NULL;
 
 	if (!dev->client_state)
-		dev->client_state = ni_client_state_new(0);
+		dev->client_state = ni_client_state_new();
 
 	return dev->client_state;
 }
@@ -636,6 +705,7 @@ ni_netdev_load_client_state(ni_netdev_t *dev)
 	ni_client_state_init(&cs);
 	if (dev && ni_client_state_load(&cs, dev->link.ifindex)) {
 		ni_netdev_set_client_state(dev, ni_client_state_clone(&cs));
+		ni_client_state_reset(&cs);
 		ni_debug_ifconfig("loading %s structure from a file for %s",
 			NI_CLIENT_STATE_XML_NODE, dev->name);
 		return TRUE;
@@ -647,21 +717,12 @@ ni_netdev_load_client_state(ni_netdev_t *dev)
 void
 ni_netdev_discover_client_state(ni_netdev_t *dev)
 {
-	ni_fsm_state_t state = NI_FSM_STATE_DEVICE_EXISTS;
 	ni_client_state_t *cs;
 
 	if (!dev)
 		return;
 
-	if (ni_netdev_device_is_up(dev))
-		state = NI_FSM_STATE_DEVICE_UP;
-	if (ni_netdev_link_is_up(dev))
-		state = NI_FSM_STATE_LINK_UP;
-	if (ni_netdev_network_is_up(dev))
-		state = NI_FSM_STATE_LINK_UP;
-
-	cs = ni_client_state_new(state);
-
+	cs = ni_client_state_new();
 	ni_netdev_set_client_state(dev, cs);
 }
 
@@ -734,7 +795,7 @@ ni_netdev_get_event_uuid(ni_netdev_t *dev, ni_event_t ev)
 	for (pos = &dev->event_filter; (efp = *pos) != NULL; pos = &efp->next) {
 		if (efp->event_mask & (1 << ev)) {
 			static ni_uuid_t ret_uuid;
-			
+
 			ret_uuid = efp->uuid;
 			*pos = efp->next;
 			__ni_event_filter_free(efp);
@@ -994,15 +1055,15 @@ __ni_netdev_list_destroy(ni_netdev_t **list)
 }
 
 void
-__ni_netdev_list_append(ni_netdev_t **list, ni_netdev_t *new_ifp)
+__ni_netdev_list_append(ni_netdev_t **list, ni_netdev_t *new_dev)
 {
 	ni_netdev_t *dev;
 
 	while ((dev = *list) != NULL)
 		list = &dev->next;
 
-	new_ifp->next = NULL;
-	*list = new_ifp;
+	new_dev->next = NULL;
+	*list = new_dev;
 }
 
 ni_bool_t
