@@ -7,13 +7,8 @@
 #include "config.h"
 #endif
 
-#include <signal.h>
-#include <limits.h>
-#include <errno.h>
-
 #include <wicked/netinfo.h>
 #include <wicked/route.h>
-#include <wicked/addrconf.h>
 #include <wicked/team.h>
 #include <wicked/bridge.h>
 #include <wicked/bonding.h>
@@ -22,8 +17,6 @@
 #include <wicked/vlan.h>
 #include <wicked/openvpn.h>
 #include <wicked/socket.h>
-#include <wicked/resolver.h>
-#include <wicked/nis.h>
 #include "netinfo_priv.h"
 #include "util_priv.h"
 #include "dbus-server.h"
@@ -31,11 +24,16 @@
 #include "xml-schema.h"
 #include "sysfs.h"
 #include "modem-manager.h"
-#include "dhcp6/options.h"
-#include "dhcp.h"
+
+#include <signal.h>
+#include <limits.h>
+#include <errno.h>
+
+#include <net/if.h>
+
 #include <gcrypt.h>
 
-extern void		ni_addrconf_updater_free(ni_addrconf_updater_t **);
+#define NI_NETDEV_REF_ARRAY_CHUNK	16
 
 typedef struct ni_netconfig_filter {
 	unsigned int		family;
@@ -74,7 +72,7 @@ static int
 __ni_init_gcrypt(void)
 {
 /*
- * gcry_check_version checks for minmum version
+ * gcry_check_version checks for minimum version
  * we want consider sufficient and returns NULL
  * on failures.
  *
@@ -582,13 +580,15 @@ ni_netconfig_device_append(ni_netconfig_t *nc, ni_netdev_t *dev)
 }
 
 static inline void
-ni_netconfig_device_unbind_slave_index(ni_netconfig_t *nc, unsigned int master)
+ni_netconfig_device_unbind_ports(ni_netconfig_t *nc, unsigned int master)
 {
 	ni_netdev_t *dev;
 
 	for (dev = nc->interfaces; dev; dev = dev->next) {
-		if (dev->link.masterdev.index == master)
+		if (dev->link.masterdev.index == master) {
 			ni_netdev_ref_destroy(&dev->link.masterdev);
+			ni_netdev_port_info_destroy(&dev->link.port);
+		}
 	}
 }
 
@@ -600,7 +600,7 @@ ni_netconfig_device_remove(ni_netconfig_t *nc, ni_netdev_t *dev)
 	for (pos = &nc->interfaces; (cur = *pos) != NULL; pos = &cur->next) {
 		if (cur == dev) {
 			*pos = cur->next;
-			ni_netconfig_device_unbind_slave_index(nc, cur->link.ifindex);
+			ni_netconfig_device_unbind_ports(nc, cur->link.ifindex);
 			ni_netdev_put(cur);
 			return;
 		}
@@ -772,13 +772,13 @@ ni_netconfig_rule_del(ni_netconfig_t *nc, const ni_rule_t *rule, ni_rule_t **pde
 			continue;
 
 		if (pdel) {
-			*pdel = ni_rule_array_remove(rules, i);
+			*pdel = ni_rule_array_remove_at(rules, i);
 			if (!*pdel) {
 				ni_error("%s: unable to remove policy rule", __func__);
 				return -1;
 			}
 		} else {
-			if (!ni_rule_array_delete(rules, i)) {
+			if (!ni_rule_array_delete_at(rules, i)) {
 				ni_error("%s: unable to remove policy rule", __func__);
 				return -1;
 			}
@@ -816,6 +816,9 @@ ni_netdev_by_name(ni_netconfig_t *nc, const char *name)
 {
 	ni_netdev_t *dev;
 
+	if (!nc && !(nc = ni_global_state_handle(0)))
+		return NULL;
+
 	for (dev = nc->interfaces; dev; dev = dev->next) {
 		if (dev->name && ni_string_eq(dev->name, name))
 			return dev;
@@ -832,6 +835,9 @@ ni_netdev_by_index(ni_netconfig_t *nc, unsigned int ifindex)
 {
 	ni_netdev_t *dev;
 
+	if (!nc && !(nc = ni_global_state_handle(0)))
+		return NULL;
+
 	for (dev = nc->interfaces; dev; dev = dev->next) {
 		if (dev->link.ifindex == ifindex)
 			return dev;
@@ -851,6 +857,9 @@ ni_netdev_by_hwaddr(ni_netconfig_t *nc, const ni_hwaddr_t *lla)
 	if (!lla || !lla->len)
 		return NULL;
 
+	if (!nc && !(nc = ni_global_state_handle(0)))
+		return NULL;
+
 	for (dev = nc->interfaces; dev; dev = dev->next) {
 		if (ni_link_address_equal(&dev->link.hwaddr, lla))
 			return dev;
@@ -859,6 +868,25 @@ ni_netdev_by_hwaddr(ni_netconfig_t *nc, const ni_hwaddr_t *lla)
 	return NULL;
 }
 
+/*
+ * Find network interface by iftype
+ */
+ni_netdev_t *
+ni_netdev_by_iftype(ni_netconfig_t *nc, ni_iftype_t iftype)
+{
+	ni_netdev_t *dev;
+
+	if (!nc && !(nc = ni_global_state_handle(0)))
+		return NULL;
+
+	for (dev = nc->interfaces; dev; dev = dev->next) {
+		if (dev->link.type == iftype)
+			return dev;
+	}
+
+	return NULL;
+}
+
 /*
  * Find VLAN interface by its tag
  */
@@ -869,6 +897,10 @@ ni_netdev_by_vlan_name_and_tag(ni_netconfig_t *nc, const char *parent_name, uint
 
 	if (!parent_name || !tag)
 		return NULL;
+
+	if (!nc && !(nc = ni_global_state_handle(0)))
+		return NULL;
+
 	for (dev = nc->interfaces; dev; dev = dev->next) {
 		if (dev->link.type == NI_IFTYPE_VLAN
 		 && dev->vlan
@@ -890,6 +922,23 @@ ni_netdev_name_to_index(const char *name)
 	return if_nametoindex(name);
 }
 
+const char *
+ni_netdev_index_to_name(char **ifname, unsigned int ifindex)
+{
+	char ifnamebuf[IFNAMSIZ] = {'\0'};
+
+	if (!ifname || !ifindex)
+		return NULL;
+
+	if (!if_indextoname(ifindex, ifnamebuf))
+		return NULL;
+
+	if (!ni_string_dup(ifname, ifnamebuf))
+		return NULL;
+
+	return *ifname;
+}
+
 /*
  * Create a unique interface name
  */
@@ -908,21 +957,142 @@ ni_netdev_make_name(ni_netconfig_t *nc, const char *stem, unsigned int first)
 	return NULL;
 }
 
+/*
+ * Get port count and optionally an array of port netdev
+ * references for the given netdev (index).
+ */
+static unsigned int
+ni_netdev_get_ovsbr_ports_by_index(unsigned int index, ni_netdev_ref_array_t *ports,
+		ni_netconfig_t *nc)
+{
+	unsigned int count = ports ? ports->count : 0;
+	ni_ovs_bridge_port_info_t *pinfo;
+	ni_netdev_t *dev;
+
+	if (!index || (!nc && !(nc = ni_global_state_handle(0))))
+		return count;
+
+	for (dev = nc->interfaces; dev; dev = dev->next) {
+		if (dev->link.port.type != NI_IFTYPE_OVS_BRIDGE)
+			continue;
+
+		if (!(pinfo = dev->link.port.ovsbr))
+			continue;
+
+		if (index != pinfo->bridge.index)
+			continue;
+
+		if (ports)
+			ni_netdev_ref_array_append(ports, dev->name,
+					dev->link.ifindex);
+		else
+			count++;
+	}
+
+	return ports ? ports->count - count : count;
+}
+
+static unsigned int
+ni_netdev_get_ports_by_index(unsigned int index, ni_netdev_ref_array_t *ports,
+		ni_netconfig_t *nc)
+{
+	unsigned int count = ports ? ports->count : 0;
+	ni_netdev_t *dev;
+
+	if (!index || (!nc && !(nc = ni_global_state_handle(0))))
+		return count;
+
+	for (dev = nc->interfaces; dev; dev = dev->next) {
+		if (dev->link.masterdev.index != index)
+			continue;
+
+		if (ports)
+			ni_netdev_ref_array_append(ports, dev->name,
+					dev->link.ifindex);
+		else
+			count++;
+	}
+
+	return ports ? ports->count - count : count;
+}
+
+extern unsigned int
+ni_netdev_get_ports(const ni_netdev_t *dev, ni_netdev_ref_array_t *ports,
+		ni_netconfig_t *nc)
+{
+	unsigned int index;
+
+	if (!dev)
+		return ports ? ports->count : 0;
+
+	index = dev->link.ifindex;
+	if (dev->link.type == NI_IFTYPE_OVS_BRIDGE)
+		return ni_netdev_get_ovsbr_ports_by_index(index, ports, nc);
+	else
+		return ni_netdev_get_ports_by_index(index, ports, nc);
+}
+
+/*
+ * Resolve port's master reference to netdev
+ */
+const ni_netdev_ref_t *
+ni_netdev_get_master_ref(const ni_netdev_t *dev)
+{
+	if (!dev)
+		return NULL;
+
+	if (dev->link.port.type == NI_IFTYPE_OVS_BRIDGE) {
+		ni_ovs_bridge_port_info_t *ovsbr;
+
+		if (!(ovsbr = dev->link.port.ovsbr))
+			return NULL;
+
+		return &ovsbr->bridge;
+	}
+
+	return &dev->link.masterdev;
+}
+
+ni_netdev_t *
+ni_netdev_resolve_master(const ni_netdev_t *dev, ni_netconfig_t *nc)
+{
+	const ni_netdev_ref_t *ref;
+
+	if ((ref = ni_netdev_get_master_ref(dev)))
+		return ni_netdev_ref_resolve(ref, nc);
+
+	return NULL;
+}
+
 /*
  * netdev reference
  */
 ni_bool_t
-ni_netdev_ref_init(ni_netdev_ref_t *ref, const char *ifname, unsigned int ifindex)
+ni_netdev_ref_init(ni_netdev_ref_t *ref)
 {
 	if (ref) {
 		memset(ref, 0, sizeof(*ref));
-		ni_string_dup(&ref->name, ifname);
-		ref->index = ifindex;
+		ref->ns_id = NI_NETNSID_DEFAULT;
 		return TRUE;
 	}
 	return FALSE;
 }
 
+ni_bool_t
+ni_netdev_ref_copy(ni_netdev_ref_t *dst, const ni_netdev_ref_t *src)
+{
+	if (dst && src) {
+		ni_netdev_ref_destroy(dst);
+		if (ni_string_dup(&dst->name, src->name)) {
+			dst->index = src->index;
+			dst->ns_id = src->ns_id;
+		}
+		return TRUE;
+	}
+	return FALSE;
+}
+
+
 ni_bool_t
 ni_netdev_ref_set(ni_netdev_ref_t *ref, const char *ifname, unsigned int ifindex)
 {
@@ -949,6 +1119,16 @@ ni_netdev_ref_set_ifindex(ni_netdev_ref_t *ref, unsigned int ifindex)
 	return FALSE;
 }
 
+ni_bool_t
+ni_netdev_ref_set_netnsid(ni_netdev_ref_t *ref, unsigned int ns_id)
+{
+	if (ref) {
+		ref->ns_id = ns_id;
+		return TRUE;
+	}
+	return FALSE;
+}
+
 ni_netdev_t *
 ni_netdev_ref_resolve(const ni_netdev_ref_t *ref, ni_netconfig_t *nc)
 {
@@ -957,6 +1137,9 @@ ni_netdev_ref_resolve(const ni_netdev_ref_t *ref, ni_netconfig_t *nc)
 	if (!ref || (!nc && !(nc = ni_global_state_handle(0))))
 		return NULL;
 
+	if (ref->ns_id != NI_NETNSID_DEFAULT)
+		return NULL;
+
 	if (ref->index && (dev = ni_netdev_by_index(nc, ref->index)))
 		return dev;
 
@@ -974,6 +1157,9 @@ ni_netdev_ref_bind_ifname(ni_netdev_ref_t *ref, ni_netconfig_t *nc)
 	if (!ref || (!nc && !(nc = ni_global_state_handle(0))))
 		return NULL;
 
+	if (ref->ns_id != NI_NETNSID_DEFAULT)
+		return NULL;
+
 	dev = ni_netdev_by_index(nc, ref->index);
 	if (dev == NULL)
 		return NULL;
@@ -991,6 +1177,9 @@ ni_netdev_ref_bind_ifindex(ni_netdev_ref_t *ref, ni_netconfig_t *nc)
 	if (!ref || (!nc && !(nc = ni_global_state_handle(0))))
 		return NULL;
 
+	if (ref->ns_id != NI_NETNSID_DEFAULT)
+		return NULL;
+
 	dev = ni_netdev_by_name(nc, ref->name);
 	if (dev == NULL)
 		return NULL;
@@ -1003,71 +1192,183 @@ void
 ni_netdev_ref_destroy(ni_netdev_ref_t *ref)
 {
 	if (ref) {
+		ref->ns_id = NI_NETNSID_DEFAULT;
 		ref->index = 0;
 		ni_string_free(&ref->name);
 	}
 }
 
+ni_bool_t
+ni_netdev_ref_array_init(ni_netdev_ref_array_t *array)
+{
+	if (array) {
+		memset(array, 0, sizeof(*array));
+		return TRUE;
+	}
+	return FALSE;
+}
+
+const ni_netdev_ref_t *
+ni_netdev_ref_array_at(const ni_netdev_ref_array_t *array, unsigned int i)
+{
+	if (!array || i >= array->count)
+		return NULL;
+	return &array->data[i];
+}
+
+const ni_netdev_ref_t *
+ni_netdev_ref_array_find_index(const ni_netdev_ref_array_t *array, unsigned int index)
+{
+	const ni_netdev_ref_t *ref;
+	unsigned int i;
+
+	if (!array)
+		return NULL;
+
+	for (i = 0; i < array->count; ++i) {
+		ref = &array->data[i];
+		if (ref->index == index)
+			return ref;
+	}
+	return NULL;
+}
+
+const ni_netdev_ref_t *
+ni_netdev_ref_array_find_name(const ni_netdev_ref_array_t *array, const char *name)
+{
+	const ni_netdev_ref_t *ref;
+	unsigned int i;
+
+	if (!array)
+		return NULL;
+
+	for (i = 0; i < array->count; ++i) {
+		ref = &array->data[i];
+		if (ni_string_eq(ref->name, name))
+			return ref;
+	}
+	return NULL;
+}
+
+static ni_bool_t
+ni_netdev_ref_array_realloc(ni_netdev_ref_array_t *array, unsigned int count)
+{
+	ni_netdev_ref_t *newdata;
+	size_t           newsize;
+	unsigned int     i;
+
+	if ((UINT_MAX - array->count) <= count)
+		return FALSE;
+
+	newsize = array->count + count;
+	if ((SIZE_MAX / sizeof(*newdata)) < newsize)
+		return FALSE;
+
+	newdata = realloc(array->data, newsize * sizeof(*newdata));
+	if (!newdata)
+		return FALSE;
+
+	array->data = newdata;
+	for (i = array->count; i < newsize; ++i)
+		ni_netdev_ref_init(&array->data[i]);
+	return TRUE;
+}
+
+const ni_netdev_ref_t *
+ni_netdev_ref_array_append(ni_netdev_ref_array_t *array, const char *name, unsigned int index)
+{
+	ni_netdev_ref_t *item;
+
+	if (!array || ((array->count % NI_NETDEV_REF_ARRAY_CHUNK) == 0 &&
+	    !ni_netdev_ref_array_realloc(array, NI_NETDEV_REF_ARRAY_CHUNK)))
+		return NULL;
+
+	item = &array->data[array->count++];
+	ni_netdev_ref_set(item, name, index);
+	return item;
+}
+
+void
+ni_netdev_ref_array_destroy(ni_netdev_ref_array_t *array)
+{
+	ni_netdev_ref_t *item;
+
+	if (array) {
+		while (array->count) {
+			array->count--;
+			item = &array->data[array->count];
+			ni_netdev_ref_destroy(item);
+		}
+		free(array->data);
+		array->data = NULL;
+	}
+}
+
 /*
  * Handle netdev request port config
  */
-static void
-ni_netdev_port_req_init(ni_netdev_port_req_t *port)
+ni_bool_t
+ni_netdev_port_config_init(ni_netdev_port_config_t *conf, ni_iftype_t type)
 {
-	switch (port->type) {
-        case NI_IFTYPE_TEAM:
-		ni_team_port_config_init(&port->team);
-		break;
+	if (conf) {
+		memset(conf, 0, sizeof(*conf));
 
-	case NI_IFTYPE_OVS_BRIDGE:
-		ni_ovs_bridge_port_config_init(&port->ovsbr);
-		break;
+		switch (type) {
+		case NI_IFTYPE_BOND:
+			if (!(conf->bond = ni_bonding_port_config_new()))
+				return FALSE;
+			break;
 
-	case NI_IFTYPE_BOND:
-	case NI_IFTYPE_BRIDGE:
-	default:
-		break;
-	}
-}
-
-ni_netdev_port_req_t *
-ni_netdev_port_req_new(ni_iftype_t master)
-{
-	ni_netdev_port_req_t *port;
-
-	switch (master) {
-	case NI_IFTYPE_TEAM:
-	case NI_IFTYPE_BOND:
-	case NI_IFTYPE_BRIDGE:
-	case NI_IFTYPE_OVS_BRIDGE:
-		port = xcalloc(1, sizeof(*port));
-		port->type = master;
-		ni_netdev_port_req_init(port);
-		return port;
-	default:
-		return NULL;
-	}
-}
-
-void
-ni_netdev_port_req_free(ni_netdev_port_req_t *port)
-{
-	if (port) {
-		switch (port->type) {
 		case NI_IFTYPE_TEAM:
-			ni_team_port_config_destroy(&port->team);
+			if (!(conf->team = ni_team_port_config_new()))
+				return FALSE;
+			break;
+
+		case NI_IFTYPE_BRIDGE:
+			if (!(conf->bridge = ni_bridge_port_config_new()))
+				return FALSE;
 			break;
 
 		case NI_IFTYPE_OVS_BRIDGE:
-			ni_ovs_bridge_port_config_destroy(&port->ovsbr);
+			if (!(conf->ovsbr = ni_ovs_bridge_port_config_new()))
+				return FALSE;
 			break;
 
-		case NI_IFTYPE_BOND:
-		case NI_IFTYPE_BRIDGE:
 		default:
 			break;
 		}
-		free(port);
+		conf->type = type;
+
+		return TRUE;
+	}
+	return FALSE;
+}
+void
+ni_netdev_port_config_destroy(ni_netdev_port_config_t *conf)
+{
+	if (conf) {
+		switch (conf->type) {
+		case NI_IFTYPE_BOND:
+			ni_bonding_port_config_free(conf->bond);
+			break;
+
+		case NI_IFTYPE_TEAM:
+			ni_team_port_config_free(conf->team);
+			break;
+
+		case NI_IFTYPE_BRIDGE:
+			ni_bridge_port_config_free(conf->bridge);
+			break;
+
+		case NI_IFTYPE_OVS_BRIDGE:
+			ni_ovs_bridge_port_config_free(conf->ovsbr);
+			break;
+
+		default:
+			break;
+		}
+
+		memset(conf, 0, sizeof(*conf));
 	}
 }
 
@@ -1080,6 +1381,8 @@ ni_netdev_req_new(void)
 	ni_netdev_req_t *req;
 
 	req = xcalloc(1, sizeof(*req));
+	if (req)
+		ni_netdev_ref_init(&req->master);
 	return req;
 }
 
@@ -1088,156 +1391,7 @@ ni_netdev_req_free(ni_netdev_req_t *req)
 {
 	ni_string_free(&req->alias);
 	ni_netdev_ref_destroy(&req->master);
-	ni_netdev_port_req_free(req->port);
+	ni_netdev_port_config_destroy(&req->port);
 	free(req);
 }
 
-/*
- * Address configuration state (aka leases)
- */
-ni_addrconf_lease_t *
-ni_addrconf_lease_new(int type, int family)
-{
-	ni_addrconf_lease_t *lease;
-
-	lease = calloc(1, sizeof(*lease));
-	if (lease) {
-		lease->seqno = __ni_global_seqno++;
-		lease->type = type;
-		lease->family = family;
-		ni_config_addrconf_update_mask(lease->type, lease->family);
-	}
-	return lease;
-}
-
-void
-ni_addrconf_lease_free(ni_addrconf_lease_t *lease)
-{
-	if (lease)
-		ni_addrconf_lease_destroy(lease);
-	free(lease);
-}
-
-static void
-ni_addrconf_lease_dhcp4_destroy(struct ni_addrconf_lease_dhcp4 *dhcp4)
-{
-	if (dhcp4) {
-		ni_string_free(&dhcp4->boot_sname);
-		ni_string_free(&dhcp4->boot_file);
-		ni_string_free(&dhcp4->root_path);
-		ni_string_free(&dhcp4->message);
-
-		ni_dhcp_option_list_destroy(&dhcp4->options);
-	}
-}
-
-static void
-ni_addrconf_lease_dhcp6_destroy(struct ni_addrconf_lease_dhcp6 *dhcp6)
-{
-	if (dhcp6) {
-		ni_dhcp6_status_destroy(&dhcp6->status);
-		ni_dhcp6_ia_list_destroy(&dhcp6->ia_list);
-
-		ni_string_free(&dhcp6->boot_url);
-		ni_string_array_destroy(&dhcp6->boot_params);
-
-		ni_dhcp_option_list_destroy(&dhcp6->options);
-	}
-}
-
-void
-ni_addrconf_lease_destroy(ni_addrconf_lease_t *lease)
-{
-	ni_addrconf_updater_free(&lease->updater);
-	if (lease->old) {
-		ni_addrconf_lease_free(lease->old);
-		lease->old = NULL;
-	}
-
-	ni_string_free(&lease->owner);
-	ni_string_free(&lease->hostname);
-
-	ni_address_list_destroy(&lease->addrs);
-	ni_route_tables_destroy(&lease->routes);
-
-	if (lease->rules) {
-		ni_rule_array_free(lease->rules);
-		lease->rules = NULL;
-	}
-
-	if (lease->nis) {
-		ni_nis_info_free(lease->nis);
-		lease->nis = NULL;
-	}
-	if (lease->resolver) {
-		ni_resolver_info_free(lease->resolver);
-		lease->resolver = NULL;
-	}
-
-	ni_string_array_destroy(&lease->ntp_servers);
-	ni_string_array_destroy(&lease->nds_servers);
-	ni_string_array_destroy(&lease->nds_context);
-	ni_string_free(&lease->nds_tree);
-	ni_string_array_destroy(&lease->netbios_name_servers);
-	ni_string_array_destroy(&lease->netbios_dd_servers);
-	ni_string_free(&lease->netbios_scope);
-	ni_string_array_destroy(&lease->slp_servers);
-	ni_string_array_destroy(&lease->slp_scopes);
-	ni_string_array_destroy(&lease->sip_servers);
-	ni_string_array_destroy(&lease->lpr_servers);
-	ni_string_array_destroy(&lease->log_servers);
-
-	ni_string_free(&lease->posix_tz_string);
-	ni_string_free(&lease->posix_tz_dbname);
-
-	switch (lease->type) {
-	case NI_ADDRCONF_DHCP:
-
-		switch (lease->family) {
-		case AF_INET:
-			ni_addrconf_lease_dhcp4_destroy(&lease->dhcp4);
-			break;
-
-		case AF_INET6:
-			ni_addrconf_lease_dhcp6_destroy(&lease->dhcp6);
-			break;
-
-		default: ;
-		}
-
-		break;
-
-	default: ;
-	}
-}
-
-void
-ni_addrconf_lease_list_destroy(ni_addrconf_lease_t **list)
-{
-	ni_addrconf_lease_t *lease;
-
-	while ((lease = *list) != NULL) {
-		*list = lease->next;
-		ni_addrconf_lease_free(lease);
-	}
-}
-
-unsigned int
-ni_addrconf_lease_get_priority(const ni_addrconf_lease_t *lease)
-{
-	if (!lease)
-		return 0;
-
-	switch (lease->type) {
-	case NI_ADDRCONF_STATIC:
-		return 2;
-
-	case NI_ADDRCONF_DHCP:
-	case NI_ADDRCONF_INTRINSIC:
-		return 1;
-
-	case NI_ADDRCONF_AUTOCONF:
-	default:
-		return 0;
-	}
-}
