@@ -2,7 +2,7 @@
  *	Handling of ip routing.
  *
  *	Copyright (C) 2009-2012 Olaf Kirch <okir@suse.de>
- *	Copyright (C) 2012-2016 SUSE LINUX GmbH, Nuernberg, Germany.
+ *	Copyright (C) 2012-2023 SUSE LLC
  *
  *	This program is free software; you can redistribute it and/or modify
  *	it under the terms of the GNU General Public License as published by
@@ -18,13 +18,23 @@
  *	along with this program. If not, see <http://www.gnu.org/licenses/>.
  *
  *	Authors:
- *		Olaf Kirch <okir@suse.de>
- *		Marius Tomaschewski <mt@suse.de>
+ *		Olaf Kirch
+ *		Karol Mroz
+ *		Marius Tomaschewski
  */
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
+#include <wicked/logging.h>
+#include <wicked/netinfo.h>
+#include <wicked/route.h>
+
+#include "refcount_priv.h"
+#include "array_priv.h"
+#include "util_priv.h"
+#include "debug.h"
+
 #include <stdlib.h>
 #include <limits.h>
 #include <netlink/netlink.h>
@@ -32,11 +42,6 @@
 #include <linux/ipv6_route.h>
 #endif
 
-#include <wicked/logging.h>
-#include <wicked/netinfo.h>
-#include <wicked/route.h>
-#include "util_priv.h"
-#include "debug.h"
 
 #define NI_ROUTE_ARRAY_CHUNK		16
 #define NI_RULE_ARRAY_CHUNK		4
@@ -154,17 +159,32 @@ static const ni_intmap_t	ni_route_mxlock_bits[] = {
 /*
  * ni_route functions
  */
-ni_route_t *
-ni_route_new(void)
+static inline ni_bool_t
+ni_route_init(ni_route_t *route)
 {
-	ni_route_t *rp;
-
-	rp = xcalloc(1, sizeof(ni_route_t));
-	if (rp)
-		rp->users = 1;
-	return rp;
+	if (route) {
+		memset(route, 0, sizeof(*route));
+		ni_netdev_ref_init(&route->nh.device);
+		return TRUE;
+	}
+	return FALSE;
 }
 
+static inline void
+ni_route_destroy(ni_route_t *rp)
+{
+	ni_route_nexthop_list_destroy(&rp->nh.next);
+	ni_route_nexthop_destroy(&rp->nh);
+	memset(rp, 0, sizeof(*rp));
+}
+
+extern ni_define_refcounted_new(ni_route);
+extern ni_define_refcounted_ref(ni_route);
+extern ni_define_refcounted_hold(ni_route);
+extern ni_define_refcounted_free(ni_route);
+extern ni_define_refcounted_drop(ni_route);
+extern ni_define_refcounted_move(ni_route);
+
 ni_route_t *
 ni_route_create(unsigned int prefixlen, const ni_sockaddr_t *dest,
 		const ni_sockaddr_t *gw, unsigned int table,
@@ -222,12 +242,9 @@ ni_route_create(unsigned int prefixlen, const ni_sockaddr_t *dest,
 	else
 		rp->table = ni_route_guess_table(rp);
 
-	if (list) {
-		if (!ni_route_tables_add_route(list, rp)) {
-			ni_route_free(rp);
-			rp = NULL;
-		}
-	}
+	if (list && !ni_route_tables_add_route(list, rp))
+		ni_route_drop(&rp);
+
 	return rp;
 }
 
@@ -295,39 +312,6 @@ ni_route_copy(ni_route_t *rp, const ni_route_t *src)
 	return ni_route_replace_hops(rp, &src->nh);
 }
 
-ni_route_t *
-ni_route_ref(ni_route_t *rp)
-{
-	if (!rp)
-		return NULL;
-
-	ni_assert(rp->users);
-	rp->users++;
-	return rp;
-}
-
-static inline void
-do_route_free(ni_route_t *rp)
-{
-	ni_route_nexthop_list_destroy(&rp->nh.next);
-	ni_route_nexthop_destroy(&rp->nh);
-
-	free(rp);
-}
-
-void
-ni_route_free(ni_route_t *rp)
-{
-	if (!rp)
-		return;
-
-	ni_assert(rp->users);
-	rp->users--;
-	if (rp->users == 0) {
-		do_route_free(rp);
-	}
-}
-
 ni_bool_t
 ni_route_update_options(ni_route_t *rp, const ni_route_t *src)
 {
@@ -470,7 +454,7 @@ ni_route_expand_hops(ni_route_array_t *routes, const ni_route_t *rp)
 
 		ni_route_free(r);
 		while (routes->count > count) {
-			if (!ni_route_array_delete(routes, routes->count - 1))
+			if (!ni_route_array_delete_at(routes, routes->count - 1))
 				break;
 		}
 		return 0;
@@ -1280,7 +1264,12 @@ ni_route_via_gateway(const ni_route_t *rp)
 ni_route_nexthop_t *
 ni_route_nexthop_new(void)
 {
-	return xcalloc(1, sizeof(ni_route_nexthop_t));
+	ni_route_nexthop_t *hop;
+
+	hop = calloc(1, sizeof(ni_route_nexthop_t));
+	if (hop)
+		ni_netdev_ref_init(&hop->device);
+	return hop;
 }
 
 void
@@ -1289,6 +1278,7 @@ ni_route_nexthop_destroy(ni_route_nexthop_t *hop)
 	if (hop) {
 		ni_netdev_ref_destroy(&hop->device);
 		memset(hop, 0, sizeof(*hop));
+		ni_netdev_ref_init(&hop->device);
 	}
 }
 
@@ -1309,8 +1299,7 @@ ni_route_nexthop_copy(ni_route_nexthop_t *dst, const ni_route_nexthop_t *src)
 		dst->weight  = src->weight;
 		dst->flags   = src->flags;
 		dst->realm   = src->realm;
-		dst->device.index = src->device.index;
-		return ni_string_dup(&dst->device.name, src->device.name);
+		return ni_netdev_ref_copy(&dst->device, &src->device);
 	}
 	return FALSE;
 }
@@ -1482,80 +1471,15 @@ ni_route_array_free(ni_route_array_t *nra)
 		free(nra);
 	}
 }
-
-void
-ni_route_array_init(ni_route_array_t *nra)
-{
-	memset(nra, 0, sizeof(*nra));
-}
-
-void
-ni_route_array_destroy(ni_route_array_t *nra)
-{
-	if (nra) {
-		while (nra->count) {
-			nra->count--;
-			ni_route_free(nra->data[nra->count]);
-		}
-		free(nra->data);
-		nra->data = NULL;
-	}
-}
-
-static ni_bool_t
-ni_route_array_realloc(ni_route_array_t *nra, unsigned int newsize)
-{
-	ni_route_t **newdata;
-	unsigned int i;
-
-	if ((UINT_MAX - NI_ROUTE_ARRAY_CHUNK) <= newsize)
-		return FALSE;
-
-	newsize = (newsize + NI_ROUTE_ARRAY_CHUNK);
-	newdata = xrealloc(nra->data, newsize * sizeof(ni_route_t *));
-	if (!newdata)
-		return FALSE;
-
-	nra->data = newdata;
-	for (i = nra->count; i < newsize; ++i) {
-		nra->data[i] = NULL;
-	}
-	return TRUE;
-}
-
-ni_bool_t
-ni_route_array_append(ni_route_array_t *nra, ni_route_t *rp)
-{
-	if (!nra || !rp)
-		return FALSE;
-
-	if ((nra->count % NI_ROUTE_ARRAY_CHUNK) == 0 &&
-	    !ni_route_array_realloc(nra, nra->count))
-		return FALSE;
-
-	nra->data[nra->count++] = rp;
-	return TRUE;
-}
-
-ni_route_t *
-ni_route_array_remove(ni_route_array_t *nra, unsigned int index)
-{
-	ni_route_t *rp;
-
-	if(!nra || index >= nra->count)
-		return NULL;
-
-	rp = nra->data[index];
-	nra->count--;
-	if (index < nra->count) {
-		memmove(&nra->data[index], &nra->data[index + 1],
-			(nra->count - index) * sizeof(ni_route_t *));
-	}
-	nra->data[nra->count] = NULL;
-
-	/* Don't bother with shrinking the array. It's not worth the trouble */
-	return rp;
-}
+extern				ni_define_ptr_array_init(ni_route);
+extern				ni_define_ptr_array_destroy(ni_route);
+static				ni_define_ptr_array_realloc(ni_route, NI_ROUTE_ARRAY_CHUNK);
+extern				ni_define_ptr_array_append(ni_route);
+extern				ni_define_ptr_array_delete_at(ni_route);
+extern				ni_define_ptr_array_remove_at(ni_route);
+extern				ni_define_ptr_array_at(ni_route);
+static				ni_define_ptr_array_qsort_cmp_fn(ni_route);
+extern				ni_define_ptr_array_qsort(ni_route);
 
 ni_route_t *
 ni_route_array_remove_ref(ni_route_array_t *nra, const ni_route_t *rp)
@@ -1567,23 +1491,11 @@ ni_route_array_remove_ref(ni_route_array_t *nra, const ni_route_t *rp)
 
 	for (i = 0; i < nra->count; i++) {
 		if (rp == nra->data[i])
-			return ni_route_array_remove(nra, i);
+			return ni_route_array_remove_at(nra, i);
 	}
 	return NULL;
 }
 
-ni_bool_t
-ni_route_array_delete(ni_route_array_t *nra, unsigned int index)
-{
-	ni_route_t *rp;
-
-	if ((rp = ni_route_array_remove(nra, index))) {
-		ni_route_free(rp);
-		return TRUE;
-	}
-	return FALSE;
-}
-
 ni_bool_t
 ni_route_array_delete_ref(ni_route_array_t *nra, const ni_route_t *rp)
 {
@@ -1596,18 +1508,10 @@ ni_route_array_delete_ref(ni_route_array_t *nra, const ni_route_t *rp)
 	return FALSE;
 }
 
-ni_route_t *
-ni_route_array_get(ni_route_array_t *nra, unsigned int index)
-{
-	if (!nra || index >= nra->count)
-		return NULL;
-	return nra->data[index];
-}
-
 ni_route_t *
 ni_route_array_ref(ni_route_array_t *nra, unsigned int index)
 {
-	return ni_route_ref(ni_route_array_get(nra, index));
+	return ni_route_ref(ni_route_array_at(nra, index));
 }
 
 ni_route_t *
@@ -1663,25 +1567,6 @@ ni_route_sort_cmp_rev(const ni_route_t *r1, const ni_route_t *r2)
 	return 0 - ni_route_sort_cmp(r1, r2);
 }
 
-static int
-ni_route_qsort_r_cmp(const void *_r1, const void *_r2, void *_cmp)
-{
-	const ni_route_t *r1 = *(const ni_route_t **)_r1;
-	const ni_route_t *r2 = *(const ni_route_t **)_r2;
-	ni_route_cmp_fn *rt_cmp = (ni_route_cmp_fn *)_cmp;
-	return rt_cmp(r1, r2);
-}
-
-void
-ni_route_array_qsort(ni_route_array_t *nra, ni_route_cmp_fn *cmp_fn)
-{
-	if (!nra || !nra->count || !cmp_fn)
-		return;
-
-	qsort_r(&nra->data[0], nra->count, sizeof(nra->data[0]),
-			ni_route_qsort_r_cmp, cmp_fn);
-}
-
 void
 ni_route_array_sort(ni_route_array_t *nra)
 {
@@ -1737,6 +1622,29 @@ ni_route_table_clear(ni_route_table_t *tab)
 /*
  * ni_route_tables list functions
  */
+void
+ni_route_tables_copy(ni_route_table_t **dst,  const ni_route_table_t *src)
+{
+	const ni_route_table_t *srt;
+	const ni_route_t *srp;
+	ni_route_t *rp;
+	unsigned int i;
+
+	if (!dst)
+		return;
+
+	for (srt = src; srt; srt = srt->next) {
+		for (i = 0; i < srt->routes.count; ++i) {
+			if (!(srp = srt->routes.data[i]))
+				continue;
+
+			rp = ni_route_clone(srp);
+			if (!ni_route_tables_add_route(dst, rp))
+				ni_route_free(rp);
+		}
+	}
+}
+
 ni_bool_t
 ni_route_tables_add_route(ni_route_table_t **list, ni_route_t *rp)
 {
@@ -1862,31 +1770,35 @@ ni_route_tables_destroy(ni_route_table_t **list)
 /*
  * routing policy rules
  */
-ni_rule_t *
-ni_rule_new(void)
+static inline ni_bool_t
+ni_rule_init(ni_rule_t *rule)
 {
-	ni_rule_t *rule;
-
-	rule = xcalloc(1, sizeof(*rule));
 	if (rule) {
-		rule->refcount = 1;
-
+		memset(rule, 0, sizeof(*rule));
 		rule->suppress_prefixlen = -1U;
 		rule->suppress_ifgroup = -1U;
+		ni_netdev_ref_init(&rule->iif);
+		ni_netdev_ref_init(&rule->oif);
+		return TRUE;
 	}
-	return rule;
+	return FALSE;
 }
 
-ni_rule_t *
-ni_rule_ref(ni_rule_t *rule)
+inline static void
+ni_rule_destroy(ni_rule_t *rule)
 {
-	if (rule) {
-		ni_assert(rule->refcount);
-		rule->refcount++;
-	}
-	return rule;
+	ni_netdev_ref_destroy(&rule->iif);
+	ni_netdev_ref_destroy(&rule->oif);
+	memset(rule, 0, sizeof(*rule));
 }
 
+extern ni_define_refcounted_new(ni_rule);
+extern ni_define_refcounted_ref(ni_rule);
+extern ni_define_refcounted_hold(ni_rule);
+extern ni_define_refcounted_free(ni_rule);
+extern ni_define_refcounted_drop(ni_rule);
+extern ni_define_refcounted_move(ni_rule);
+
 ni_bool_t
 ni_rule_copy(ni_rule_t *dst, const ni_rule_t *src)
 {
@@ -1912,12 +1824,10 @@ ni_rule_copy(ni_rule_t *dst, const ni_rule_t *src)
 	C(fwmask);
 	C(suppress_prefixlen);
 	C(suppress_ifgroup);
-	C(iif.index);
-	C(oif.index);
 #undef C
-	if (!ni_string_dup(&dst->iif.name, src->iif.name))
+	if (!ni_netdev_ref_copy(&dst->iif, &src->iif))
 		return FALSE;
-	if (!ni_string_dup(&dst->oif.name, src->oif.name))
+	if (!ni_netdev_ref_copy(&dst->oif, &src->oif))
 		return FALSE;
 	return TRUE;
 }
@@ -1937,25 +1847,6 @@ ni_rule_clone(const ni_rule_t *src)
 	return NULL;
 }
 
-static void
-do_rule_free(ni_rule_t *rule)
-{
-	ni_netdev_ref_destroy(&rule->iif);
-	ni_netdev_ref_destroy(&rule->oif);
-	free(rule);
-}
-
-void
-ni_rule_free(ni_rule_t *rule)
-{
-	if (rule) {
-		ni_assert(rule->refcount);
-		rule->refcount--;
-		if (rule->refcount == 0)
-			do_rule_free(rule);
-	}
-}
-
 static int
 do_rule_cmp_show(int ret, const char *what)
 {
@@ -2197,7 +2088,7 @@ ni_rule_print(ni_stringbuf_t *out, const ni_rule_t *rule)
 		ni_stringbuf_printf(out, " pref auto");
 
 	if (rule->flags & NI_BIT(NI_RULE_INVERT))
-		ni_stringbuf_printf(out, " not", rule->pref);
+		ni_stringbuf_printf(out, " not");
 
 	if (rule->src.len)
 		ni_stringbuf_printf(out, " from %s/%u",
@@ -2289,25 +2180,6 @@ ni_rule_print(ni_stringbuf_t *out, const ni_rule_t *rule)
 	return out ? out->string : NULL;
 }
 
-void
-ni_rule_array_init(ni_rule_array_t *rules)
-{
-	memset(rules, 0, sizeof(*rules));
-}
-
-void
-ni_rule_array_destroy(ni_rule_array_t *rules)
-{
-	if (rules) {
-		while (rules->count) {
-			rules->count--;
-			ni_rule_free(rules->data[rules->count]);
-		}
-		free(rules->data);
-		rules->data = NULL;
-	}
-}
-
 ni_rule_array_t *
 ni_rule_array_new(void)
 {
@@ -2321,117 +2193,41 @@ ni_rule_array_free(ni_rule_array_t *rules)
 	free(rules);
 }
 
-unsigned int
-ni_rule_array_index(const ni_rule_array_t *rules, const ni_rule_t *rule)
+ni_rule_array_t *
+ni_rule_array_clone(const ni_rule_array_t *orig)
 {
-	unsigned int i;
-	ni_rule_t *r;
+	ni_rule_array_t *clone;
 
-	if (rules) {
-		for (i = 0; i < rules->count; ++i) {
-			r = rules->data[i];
-			if (r == rule)
-				return i;
-		}
-	}
-	return -1U;
-}
-
-static ni_bool_t
-ni_rule_array_realloc(ni_rule_array_t *rules, unsigned int newsize)
-{
-	ni_rule_t **newdata;
-	unsigned int i;
-
-	if ((UINT_MAX - NI_RULE_ARRAY_CHUNK) <= newsize)
-		return FALSE;
-
-	newsize = (newsize + NI_RULE_ARRAY_CHUNK);
-	newdata = xrealloc(rules->data, newsize * sizeof(ni_rule_t *));
-	if (!newdata)
-		return FALSE;
-
-	rules->data = newdata;
-	for (i = rules->count; i < newsize; ++i)
-		rules->data[i] = NULL;
-
-	return TRUE;
-}
-
-ni_bool_t
-ni_rule_array_append(ni_rule_array_t *rules, ni_rule_t *rule)
-{
-	if (!rules || !rule)
-		return FALSE;
-
-	if ((rules->count % NI_RULE_ARRAY_CHUNK) == 0 &&
-	    !ni_rule_array_realloc(rules, rules->count))
-		return FALSE;
-
-	rules->data[rules->count++] = rule;
-	return TRUE;
-}
-
-ni_bool_t
-ni_rule_array_insert(ni_rule_array_t *rules, unsigned int index, ni_rule_t *rule)
-{
-	if (!rules || !rule)
-		return FALSE;
-
-	if (index >= rules->count)
-		return ni_rule_array_append(rules, rule);
-
-	if ((rules->count % NI_RULE_ARRAY_CHUNK) == 0 &&
-	    !ni_rule_array_realloc(rules, rules->count))
-		return FALSE;
-
-	memmove(&rules->data[index + 1], &rules->data[index],
-		(rules->count - index) * sizeof(ni_rule_t *));
-	rules->data[index] = rule;
-	rules->count++;
-	return TRUE;
-}
-
-ni_bool_t
-ni_rule_array_delete(ni_rule_array_t *rules, unsigned int index)
-{
-	ni_rule_t *rule;
-
-	if ((rule = ni_rule_array_remove(rules, index))) {
-		ni_rule_free(rule);
-		return TRUE;
-	}
-	return FALSE;
-}
-
-ni_rule_t *
-ni_rule_array_remove(ni_rule_array_t *rules, unsigned int index)
-{
-	ni_rule_t *rule;
-
-	if (!rules || index >= rules->count)
+	if (!orig || !(clone = ni_rule_array_new()))
 		return NULL;
 
-	rule = rules->data[index];
-	rules->count--;
-	if (index < rules->count) {
-		memmove(&rules->data[index], &rules->data[index + 1],
-			(rules->count - index) * sizeof(ni_rule_t *));
-	}
-	rules->data[rules->count] = NULL;
-
-	/* Don't bother with shrinking the array. It's not worth the trouble */
-	return rule;
+	ni_rule_array_copy(clone, orig);
+	return clone;
 }
 
-ni_rule_t *
-ni_rule_array_get(ni_rule_array_t *rules, unsigned int index)
+
+void
+ni_rule_array_copy(ni_rule_array_t *dst, const ni_rule_array_t *src)
 {
-	if (!rules || index >= rules->count)
-		return NULL;
-	return rules->data[index];
+	unsigned int i;
+
+	if (!src || !dst)
+		return;
+
+	for (i = 0; i < src->count; ++i)
+		ni_rule_array_append(dst, ni_rule_clone(src->data[i]));
 }
 
+extern ni_define_ptr_array_init(ni_rule);
+extern ni_define_ptr_array_destroy(ni_rule);
+extern ni_define_ptr_array_index(ni_rule);
+static ni_define_ptr_array_realloc(ni_rule, NI_RULE_ARRAY_CHUNK);
+extern ni_define_ptr_array_append(ni_rule);
+extern ni_define_ptr_array_insert(ni_rule);
+extern ni_define_ptr_array_delete_at(ni_rule);
+extern ni_define_ptr_array_remove_at(ni_rule);
+extern ni_define_ptr_array_at(ni_rule);
+
 ni_rule_t *
 ni_rule_array_find_match(const ni_rule_array_t *rules, const ni_rule_t *rule,
 		ni_bool_t (*match)(const ni_rule_t *, const ni_rule_t *))
