@@ -1,7 +1,25 @@
 /*
- * Routines for reading from sysfs files
+ *	Routines for reading from sysfs files
  *
- * Copyright (C) 2009-2012 Olaf Kirch <okir@suse.de>
+ *	Copyright (C) 2009-2012 Olaf Kirch <okir@suse.de>
+ *	Copyright (C) 2012-2022 SUSE LLC
+ *
+ *	This program is free software; you can redistribute it and/or modify
+ *	it under the terms of the GNU General Public License as published by
+ *	the Free Software Foundation; either version 2 of the License, or
+ *	(at your option) any later version.
+ *
+ *	This program is distributed in the hope that it will be useful,
+ *	but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *	GNU General Public License for more details.
+ *
+ *	You should have received a copy of the GNU General Public License
+ *	along with this program. If not, see <http://www.gnu.org/licenses/>.
+ *
+ *	Authors:
+ *		Olaf Kirch
+ *		Marius Tomaschewski
  */
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -18,11 +36,18 @@
 #include "sysfs.h"
 #include "ibft.h"
 
-#ifndef NI_PATH_SYSFS
-#define NI_PATH_SYSFS			"/sys"
+#ifndef NI_SYSFS_PATH
+#define NI_SYSFS_PATH			"/sys"
 #endif
 
-#define _PATH_SYS_CLASS_NET		"/sys/class/net"
+#define NI_SYSFS_BUS_PATH              NI_SYSFS_PATH"/bus"
+#define NI_SYSFS_BUS_USB_PATH          NI_SYSFS_BUS_PATH"/usb"
+#define NI_SYSFS_BUS_USB_DEVICES_PATH  NI_SYSFS_BUS_USB_PATH"/devices"
+#define NI_SYSFS_BUS_PCI_PATH          NI_SYSFS_BUS_PATH"/pci"
+#define NI_SYSFS_BUS_PCI_DEVICES_PATH  NI_SYSFS_BUS_PCI_PATH"/devices"
+
+#define NI_SYSFS_CLASS_PATH		NI_SYSFS_PATH"/class"
+#define NI_SYSFS_CLASS_NET_PATH		NI_SYSFS_CLASS_PATH"/net"
 
 /* #include <linux/if_bridge.h> */
 #ifndef SYSFS_BRIDGE_ATTR
@@ -38,12 +63,13 @@
 #define SYSFS_BRIDGE_PORT_LINK		"bridge"
 #endif
 
-/* iBFT related constants */
-#define NI_SYSFS_FIRMWARE_IBFT_PATH     "/sys/firmware/ibft"
-#define NI_SYSFS_IBFT_INI_PREFIX        "initiator"
-#define NI_SYSFS_IBFT_NIC_PREFIX        "ethernet"
-#define NI_SYSFS_IBFT_TGT_PREFIX        "target"
+#define NI_SYSFS_FIRMWARE_PATH		NI_SYSFS_PATH"/firmware"
 
+/* iBFT related constants */
+#define NI_SYSFS_FIRMWARE_IBFT_PATH	NI_SYSFS_FIRMWARE_PATH"/ibft"
+#define NI_SYSFS_IBFT_INI_PREFIX	"initiator"
+#define NI_SYSFS_IBFT_NIC_PREFIX	"ethernet"
+#define NI_SYSFS_IBFT_TGT_PREFIX	"target"
 
 static const char *	__ni_sysfs_netif_attrpath(const char *ifname, const char *attr);
 static const char *	__ni_sysfs_netif_get_attr(const char *ifname, const char *attr);
@@ -173,7 +199,7 @@ ni_sysfs_netif_printf(const char *ifname, const char *attr_name, const char *fmt
 ni_bool_t
 ni_sysfs_is_read_only(void)
 {
-	return ni_fs_is_read_only(NI_PATH_SYSFS);
+	return ni_fs_is_read_only(NI_SYSFS_PATH);
 }
 
 ni_bool_t
@@ -246,7 +272,7 @@ __ni_sysfs_netif_attrpath(const char *ifname, const char *attr_name)
 	static char pathbuf[PATH_MAX];
 
 	snprintf(pathbuf, sizeof(pathbuf), "%s/%s/%s",
-			_PATH_SYS_CLASS_NET, ifname, attr_name);
+			NI_SYSFS_CLASS_NET_PATH, ifname, attr_name);
 	return pathbuf;
 }
 
@@ -324,7 +350,7 @@ ni_sysfs_bonding_get_attr(const char *ifname, const char *attr_name, char **resu
 {
 	static char pathbuf[PATH_MAX];
 
-	snprintf(pathbuf, sizeof(pathbuf), "%s/%s/bonding/%s", _PATH_SYS_CLASS_NET, ifname, attr_name);
+	snprintf(pathbuf, sizeof(pathbuf), "%s/%s/bonding/%s", NI_SYSFS_CLASS_NET_PATH, ifname, attr_name);
 	return __ni_sysfs_read_string(pathbuf, result);
 }
 
@@ -333,7 +359,7 @@ ni_sysfs_bonding_set_attr(const char *ifname, const char *attr_name, const char
 {
 	static char pathbuf[PATH_MAX];
 
-	snprintf(pathbuf, sizeof(pathbuf), "%s/%s/bonding/%s", _PATH_SYS_CLASS_NET, ifname, attr_name);
+	snprintf(pathbuf, sizeof(pathbuf), "%s/%s/bonding/%s", NI_SYSFS_CLASS_NET_PATH, ifname, attr_name);
 	return __ni_sysfs_printf(pathbuf, "%s", attr_value);
 }
 
@@ -346,7 +372,7 @@ ni_sysfs_bonding_set_list_attr(const char *ifname, const char *attr_name, const
 	int rv = -1;
 
 	snprintf(pathbuf, sizeof(pathbuf), "%s/%s/bonding/%s",
-			_PATH_SYS_CLASS_NET, ifname, attr_name);
+			NI_SYSFS_CLASS_NET_PATH, ifname, attr_name);
 
 	ni_string_array_init(&current);
 	if (__ni_sysfs_read_list(pathbuf, &current) < 0)
@@ -491,14 +517,14 @@ ni_sysfs_bridge_get_port_names(const char *ifname, ni_string_array_t *names)
 }
 
 void
-ni_sysfs_bridge_port_get_config(const char *ifname, ni_bridge_port_t *port)
+ni_sysfs_bridge_port_get_config(const char *ifname, ni_bridge_port_config_t *port)
 {
 	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/priority", &port->priority);
 	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/path_cost", &port->path_cost);
 }
 
 int
-ni_sysfs_bridge_port_update_config(const char *ifname, const ni_bridge_port_t *port)
+ni_sysfs_bridge_port_update_config(const char *ifname, const ni_bridge_port_config_t *port)
 {
 	int rv = 0;
 
@@ -513,28 +539,6 @@ ni_sysfs_bridge_port_update_config(const char *ifname, const ni_bridge_port_t *p
 	return rv;
 }
 
-void
-ni_sysfs_bridge_port_get_status(const char *ifname, ni_bridge_port_status_t *ps)
-{
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/priority", &ps->priority);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/path_cost", &ps->path_cost);
-
-	ni_sysfs_netif_get_int(ifname, SYSFS_BRIDGE_PORT_ATTR "/state", &ps->state);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/port_no", &ps->port_no);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/port_id", &ps->port_no);
-	ni_sysfs_netif_get_string(ifname, SYSFS_BRIDGE_PORT_ATTR "/designated_root", &ps->designated_root);
-	ni_sysfs_netif_get_string(ifname, SYSFS_BRIDGE_PORT_ATTR "/designated_bridge", &ps->designated_bridge);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/designated_port", &ps->designated_port);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/designated_cost", &ps->designated_cost);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/change_ack", &ps->change_ack);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/hairpin_mode", &ps->hairpin_mode);
-	ni_sysfs_netif_get_uint(ifname, SYSFS_BRIDGE_PORT_ATTR "/config_pending", &ps->config_pending);
-
-	ni_sysfs_netif_get_ulong(ifname, SYSFS_BRIDGE_PORT_ATTR "/hold_timer", &ps->hold_timer);
-	ni_sysfs_netif_get_ulong(ifname, SYSFS_BRIDGE_PORT_ATTR "/message_age_timer", &ps->message_age_timer);
-	ni_sysfs_netif_get_ulong(ifname, SYSFS_BRIDGE_PORT_ATTR "/forward_delay_timer", &ps->forward_delay_timer);
-}
-
 /*
  * Get/set IPv4 sysctls
  */
@@ -551,53 +555,9 @@ __ni_sysctl_ipv4_ifconfig_path(const char *ifname, const char *ctl_name)
 }
 
 int
-ni_sysctl_ipv4_ifconfig_is_present(const char *ifname)
+ni_sysctl_ipv4_ifconfig_set(const char *ifname, const char *ctl_name, const char *newval)
 {
-	const char *pathname = __ni_sysctl_ipv4_ifconfig_path(ifname, NULL);
-
-	return access(pathname, F_OK) == 0;
-}
-
-int
-ni_sysctl_ipv4_ifconfig_get_int(const char *ifname, const char *ctl_name, int *value)
-{
-	const char *pathname = __ni_sysctl_ipv4_ifconfig_path(ifname, ctl_name);
-	char *result = NULL;
-	int ret;
-
-	*value = 0;
-	if (__ni_sysfs_read_string(pathname, &result) < 0) {
-		ni_error("%s: unable to read file: %m", pathname);
-		return -1;
-	}
-	if (result == NULL) {
-		ni_error("%s: empty file", pathname);
-		return -1;
-	}
-	ret = ni_parse_int(result, value, 0);
-	ni_string_free(&result);
-	return ret;
-}
-
-int
-ni_sysctl_ipv4_ifconfig_get_uint(const char *ifname, const char *ctl_name, unsigned int *value)
-{
-	const char *pathname = __ni_sysctl_ipv4_ifconfig_path(ifname, ctl_name);
-	char *result = NULL;
-	int ret;
-
-	*value = 0;
-	if (__ni_sysfs_read_string(pathname, &result) < 0) {
-		ni_error("%s: unable to read file: %m", pathname);
-		return -1;
-	}
-	if (result == NULL) {
-		ni_error("%s: empty file", pathname);
-		return -1;
-	}
-	ret = ni_parse_uint(result, value, 0);
-	ni_string_free(&result);
-	return ret;
+	return __ni_sysfs_printf(__ni_sysctl_ipv4_ifconfig_path(ifname, ctl_name), "%s", newval ? newval : "");
 }
 
 int
@@ -636,45 +596,39 @@ ni_sysctl_ipv6_ifconfig_is_present(const char *ifname)
 }
 
 int
-ni_sysctl_ipv6_ifconfig_get_int(const char *ifname, const char *ctl_name, int *value)
+ni_sysctl_ipv6_ifconfig_get(const char *ifname, const char *ctl_name, char **result)
 {
 	const char *pathname = __ni_sysctl_ipv6_ifconfig_path(ifname, ctl_name);
-	char *result = NULL;
-	int ret;
 
-	*value = 0;
-	if (__ni_sysfs_read_string(pathname, &result) < 0) {
+	if (!result || __ni_sysfs_read_string(pathname, result) < 0 || !*result) {
 		ni_error("%s: unable to read file: %m", pathname);
 		return -1;
 	}
-	if (result == NULL) {
-		ni_error("%s: empty file", pathname);
+	return 0;
+}
+
+int
+ni_sysctl_ipv6_ifconfig_get_ipv6(const char *ifname, const char *ctl_name, struct in6_addr *ipv6)
+{
+	char *result = NULL;
+	ni_sockaddr_t addr;
+	int ret;
+
+	*ipv6 = in6addr_any;
+	if (ni_sysctl_ipv6_ifconfig_get(ifname, ctl_name, &result) < 0)
 		return -1;
-	}
-	ret = ni_parse_int(result, value, 0);
+
+	if ((ret = ni_sockaddr_parse(&addr, result, AF_INET6)) == 0)
+		*ipv6 = addr.six.sin6_addr;
+
 	ni_string_free(&result);
 	return ret;
 }
 
 int
-ni_sysctl_ipv6_ifconfig_get_uint(const char *ifname, const char *ctl_name, unsigned int *value)
+ni_sysctl_ipv6_ifconfig_set(const char *ifname, const char *ctl_name, const char *newval)
 {
-	const char *pathname = __ni_sysctl_ipv6_ifconfig_path(ifname, ctl_name);
-	char *result = NULL;
-	int ret;
-
-	*value = 0;
-	if (__ni_sysfs_read_string(pathname, &result) < 0) {
-		ni_error("%s: unable to read file: %m", pathname);
-		return -1;
-	}
-	if (result == NULL) {
-		ni_error("%s: empty file", pathname);
-		return -1;
-	}
-	ret = ni_parse_uint(result, value, 0);
-	ni_string_free(&result);
-	return ret;
+	return __ni_sysfs_printf(__ni_sysctl_ipv6_ifconfig_path(ifname, ctl_name), "%s", newval ? newval : "");
 }
 
 int
@@ -689,6 +643,19 @@ ni_sysctl_ipv6_ifconfig_set_uint(const char *ifname, const char *ctl_name, unsig
 	return __ni_sysfs_printf(__ni_sysctl_ipv6_ifconfig_path(ifname, ctl_name), "%u", newval);
 }
 
+int
+ni_sysctl_ipv6_ifconfig_set_ipv6(const char *ifname, const char *ctl_name, const struct in6_addr newval)
+{
+	ni_sockaddr_t addr;
+	char abuf[128] = {'\0'};
+
+	ni_sockaddr_set_ipv6(&addr, newval, 0);
+	if (!ni_sockaddr_format(&addr, abuf, sizeof(abuf)))
+		return -1;
+
+	return ni_sysctl_ipv6_ifconfig_set(ifname, ctl_name, abuf);
+}
+
 /*
  * Print a value to a sysfs file
  */
@@ -928,7 +895,7 @@ __ni_sysfs_ibft_nic_parse(const char *base, const char *node)
 	if (__ni_sysfs_ibft_nic_get_string(base, node,
 				"subnet-mask", &temp) == 0 && temp) {
 		/* The ibft module in 3.0.x kernels prints the ibft prefix
-		   lenght as ipv4 netmask; I guess nobody ever used IPv6 */
+		   length as ipv4 netmask; I guess nobody ever used IPv6 */
 		ni_sockaddr_t mask;
 		if (ni_sockaddr_parse(&mask, temp, AF_UNSPEC) != 0)
 			goto error;
@@ -1028,7 +995,7 @@ ni_sysfs_netdev_get_pci(const char *ifname)
 	ni_pci_dev_t *pci = NULL;
 	const char *attr;
 
-	snprintf(pathbuf, sizeof(pathbuf), "%s/%s", _PATH_SYS_CLASS_NET, ifname);
+	snprintf(pathbuf, sizeof(pathbuf), "%s/%s", NI_SYSFS_CLASS_NET_PATH, ifname);
 	if (readlink(pathbuf, device_link, sizeof(device_link)) < 0)
 		return NULL;
 
@@ -1055,3 +1022,305 @@ failed:
 		ni_pci_dev_free(pci);
 	return NULL;
 }
+
+/*
+ * USB/PCI bus device scan
+ */
+static ni_bool_t
+ni_sysfs_bus_device_attrs_match(const char *path, const ni_var_array_t *vars)
+{
+	const ni_var_t *var;
+	char *value = NULL;
+	char *attr = NULL;
+	unsigned int i;
+
+	if (!vars || !vars->count)
+		return TRUE;
+
+	for (i = 0; i < vars->count; ++i) {
+		var = &vars->data[i];
+
+		if (ni_string_empty(var->name) ||
+		    !ni_string_printf(&attr, "%s/%s", path, var->name))
+			continue;
+
+		if (__ni_sysfs_read_string(attr, &value) != 0 ||
+		    !ni_string_eq(var->value, value)) {
+			ni_string_free(&value);
+			ni_string_free(&attr);
+			return FALSE;
+		}
+
+		ni_string_free(&value);
+		ni_string_free(&attr);
+	}
+	return TRUE;
+}
+
+static int
+ni_sysfs_bus_device_path_scan(ni_string_array_t *paths,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *base)
+{
+	ni_string_array_t names = NI_STRING_ARRAY_INIT;
+	char *path = NULL;
+	const char *name;
+	unsigned int cnt;
+	unsigned int i;
+	int ret = -1;
+
+	if (!paths || ni_string_empty(base) ||
+	    (ret = ni_scandir(base, pattern, &names)) <= 0)
+		return ret;
+
+	cnt = paths->count;
+	for (i = 0; i < names.count; ++i) {
+		name = names.data[i];
+
+		if (!ni_string_printf(&path, "%s/%s", base, name))
+			continue;
+
+		if (ni_sysfs_bus_device_attrs_match(path, vars))
+			ni_string_array_append(paths, path);
+
+		ni_string_free(&path);
+	}
+
+	ni_string_array_destroy(&names);
+	return paths->count - cnt;
+}
+
+static ni_bool_t
+ni_sysfs_get_net_ifname_ifindex(int *ifindex, const char *netpath, const char *ifname)
+{
+	char *path = NULL;
+	char *temp = NULL;
+
+	if (!ifindex || ni_string_empty(netpath) || ni_string_empty(ifname))
+		return FALSE;
+
+	if (!ni_string_printf(&path, "%s/%s/ifindex", netpath, ifname))
+		return FALSE;
+
+	/* retrieve "$path/$ifname/<ifindex>" attribute */
+	if (__ni_sysfs_read_string(path, &temp) ||
+	    ni_parse_int(temp, ifindex, 10)     || *ifindex <= 0) {
+		ni_string_free(&temp);
+		ni_string_free(&path);
+		return FALSE;
+	} else {
+		ni_string_free(&temp);
+		ni_string_free(&path);
+		return TRUE;
+	}
+}
+
+int
+ni_sysfs_bus_usb_device_path_scan(ni_string_array_t *paths,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *root)
+{
+	char *base = NULL;
+	int ret = -1;
+
+	if (ni_string_empty(root))
+		ni_string_printf(&base, "%s", NI_SYSFS_BUS_USB_DEVICES_PATH);
+	else
+		ni_string_printf(&base, "%s/%s", root, NI_SYSFS_BUS_USB_DEVICES_PATH);
+
+	ret = ni_sysfs_bus_device_path_scan(paths, vars, pattern, base);
+	ni_string_free(&base);
+	return ret;
+}
+
+static int
+ni_sysfs_bus_usb_device_sub_net_scan(ni_string_array_t *npaths,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *root)
+{
+	ni_string_array_t paths = NI_STRING_ARRAY_INIT;
+	ni_string_array_t sdirs = NI_STRING_ARRAY_INIT;
+	char *spattern = NULL;
+	char *npath = NULL;
+	const char *path;
+	const char *sdir;
+	unsigned int i, j;
+	unsigned int cnt;
+	int ret = -1;
+
+	if (!npaths ||
+	    (ret = ni_sysfs_bus_usb_device_path_scan(&paths, vars, pattern, root)) <= 0)
+		return ret;
+
+	cnt = npaths->count;
+	for (i = 0; i < paths.count; ++i) {
+		path = paths.data[i];
+
+		if (!ni_string_printf(&spattern, "%s:*", ni_basename(path)))
+			continue;
+
+		if ((ret = ni_scandir(path, spattern, &sdirs)) <= 0) {
+			ni_string_free(&spattern);
+			continue;
+		}
+		ni_string_free(&spattern);
+
+		for (j = 0; j < sdirs.count; ++j) {
+			sdir = sdirs.data[j];
+
+			if (!ni_string_printf(&npath, "%s/%s/net", path, sdir))
+				continue;
+
+			if (ni_isdir(npath))
+				ni_string_array_append(npaths, npath);
+
+			ni_string_free(&npath);
+		}
+		ni_string_array_destroy(&sdirs);
+	}
+	ni_string_array_destroy(&paths);
+	return npaths->count - cnt;
+}
+
+int
+ni_sysfs_bus_usb_device_netdev_scan(ni_netdev_ref_array_t *devs,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *root)
+{
+	ni_string_array_t paths = NI_STRING_ARRAY_INIT;
+	ni_string_array_t names = NI_STRING_ARRAY_INIT;
+	unsigned int i, j;
+	unsigned int cnt;
+	const char *path;
+	const char *name;
+	int ret = -1;
+	int index;
+
+	if (!devs ||
+	    (ret = ni_sysfs_bus_usb_device_sub_net_scan(&paths, vars, pattern, root)) <= 0)
+		return ret;
+
+	cnt = devs->count;
+	for (i = 0; i < paths.count; ++i) {
+		path = paths.data[i];
+
+		/* get $path/net/<name> sub-dir */
+		if (ni_scandir(path, NULL, &names) <= 0)
+			continue;
+
+		for (j = 0; j < names.count; ++j) {
+			name = names.data[j];
+
+			if (!ni_sysfs_get_net_ifname_ifindex(&index, path, name))
+				continue;
+
+			if (ni_netdev_ref_array_find_index(devs, index))
+				continue;
+
+			ni_netdev_ref_array_append(devs, name, index);
+		}
+		ni_string_array_destroy(&names);
+	}
+	ni_string_array_destroy(&paths);
+	return devs->count - cnt;
+}
+
+int
+ni_sysfs_bus_pci_device_path_scan(ni_string_array_t *paths,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *root)
+{
+	char *base = NULL;
+	int ret = -1;
+
+	if (ni_string_empty(root))
+		ni_string_printf(&base, "%s", NI_SYSFS_BUS_PCI_DEVICES_PATH);
+	else
+		ni_string_printf(&base, "%s/%s", root, NI_SYSFS_BUS_PCI_DEVICES_PATH);
+
+	ret = ni_sysfs_bus_device_path_scan(paths, vars, pattern, base);
+	ni_string_free(&base);
+	return ret;
+}
+
+static int
+ni_sysfs_bus_pci_device_net_scan(ni_string_array_t *npaths,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *root)
+{
+	ni_string_array_t paths = NI_STRING_ARRAY_INIT;
+	char *npath = NULL;
+	const char *path;
+	unsigned int cnt;
+	unsigned int i;
+	int ret = -1;
+
+	if (!npaths ||
+	    (ret = ni_sysfs_bus_pci_device_path_scan(&paths, vars, pattern, root)) <= 0)
+		return ret;
+
+	cnt = npaths->count;
+	for (i = 0; i < paths.count; ++i) {
+		path = paths.data[i];
+
+		if (!ni_string_printf(&npath, "%s/net", path))
+			continue;
+
+		if (ni_isdir(npath))
+			ni_string_array_append(npaths, npath);
+
+		ni_string_free(&npath);
+	}
+	ni_string_array_destroy(&paths);
+	return npaths->count - cnt;
+}
+
+int
+ni_sysfs_bus_pci_device_netdev_scan(ni_netdev_ref_array_t *devs,
+				const ni_var_array_t *vars,
+				const char *pattern,
+				const char *root)
+{
+	ni_string_array_t paths = NI_STRING_ARRAY_INIT;
+	ni_string_array_t names = NI_STRING_ARRAY_INIT;
+	unsigned int i, j;
+	unsigned int cnt;
+	const char *path;
+	const char *name;
+	int ret = -1;
+	int index;
+
+	if (!devs ||
+	    (ret = ni_sysfs_bus_pci_device_net_scan(&paths, vars, pattern, root)) <= 0)
+		return ret;
+
+	cnt = devs->count;
+	for (i = 0; i < paths.count; ++i) {
+		path = paths.data[i];
+
+		/* get $path/net/<name> sub-dir */
+		if (ni_scandir(path, NULL, &names) <= 0)
+			continue;
+
+		for (j = 0; j < names.count; ++j) {
+			name = names.data[j];
+
+			if (!ni_sysfs_get_net_ifname_ifindex(&index, path, name))
+				continue;
+
+			if (ni_netdev_ref_array_find_index(devs, index))
+				continue;
+
+			ni_netdev_ref_array_append(devs, name, index);
+		}
+		ni_string_array_destroy(&names);
+	}
+	ni_string_array_destroy(&paths);
+	return devs->count - cnt;
+}
