@@ -21,6 +21,8 @@
 #include "util_priv.h"
 #include "process.h"
 #include "appconfig.h"
+#include "extension.h"
+#include "addrconf.h"
 #include "buffer.h"
 #include "debug.h"
 
@@ -159,6 +161,7 @@ ni_updater_source_new(void)
 	src = xcalloc(1, sizeof(*src));
 	if (src) {
 		src->refcount = 1;
+		ni_netdev_ref_init(&src->device);
 	}
 	return src;
 }
@@ -385,6 +388,7 @@ ni_updater_job_new(ni_updater_job_t **list, const ni_addrconf_lease_t *lease,
 
 	job->nr = job_nr++; /* for debugging purposes only */
 	job->refcount = 1;
+	ni_netdev_ref_init(&job->device);
 	if (!ni_netdev_ref_set(&job->device, ifname, ifindex)) {
 		free(job);
 		return NULL;
@@ -589,12 +593,12 @@ ni_system_updaters_init(void)
 
 		updater->enabled = TRUE;
 		updater->format = ni_updater_format_type(ex->format);
-		updater->proc_backup = ni_extension_script_find(ex, "backup");
-		updater->proc_restore = ni_extension_script_find(ex, "restore");
-		updater->proc_install = ni_extension_script_find(ex, "install");
-		updater->proc_remove = ni_extension_script_find(ex, "remove");
+		updater->proc_backup = ni_extension_find_script(ex, "backup");
+		updater->proc_restore = ni_extension_find_script(ex, "restore");
+		updater->proc_install = ni_extension_find_script(ex, "install");
+		updater->proc_remove = ni_extension_find_script(ex, "remove");
 		if (kind == NI_ADDRCONF_UPDATER_GENERIC) {
-			if ((updater->proc_batch = ni_extension_script_find(ex, "batch"))) {
+			if ((updater->proc_batch = ni_extension_find_script(ex, "batch"))) {
 				if (!ni_system_updater_generic_batch_test(updater))
 					updater->proc_batch = NULL;
 			}
@@ -1528,7 +1532,7 @@ ni_system_updater_hostname_lookup_call(ni_updater_t *updater, ni_updater_job_t *
 				ni_addrconf_type_to_name(job->lease->type),
 				ni_addrconf_state_to_name(job->lease->state),
 				ni_updater_name(job->kind),
-				ni_basename(shellcmd->command), pi->pid);
+				ni_basename(pi->process->command), pi->pid);
 	} else {
 		ni_process_free(pi);
 	}
