@@ -14,6 +14,7 @@
 #include <dirent.h>
 #include <ctype.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include <signal.h>
 #include <fcntl.h>
@@ -25,14 +26,22 @@
 #include <wicked/logging.h>
 #include <wicked/netinfo.h> /* only for CONFIG_WICKED_STATEDIR */
 #include <wicked/fsm.h> /* for NI_IFWORKER_INFINITE_TIMEOUT */
+#include "slist_priv.h"
 #include "util_priv.h"
 
 #define NI_STRING_ARRAY_CHUNK	16
 #define NI_UINT_ARRAY_CHUNK	16
 #define NI_VAR_ARRAY_CHUNK	16
 
-#define NI_STRINGBUF_CHUNK	64
+#define NI_STRINGBUF_CHUNK	64	/* important: always a (2^n) */
 
+#define NI_BYTE_ARRAY_CHUNK	0xff	/* important: (2^n)-1 with all bits set */
+#define NI_BYTE_ARRAY_SIZE(len)	((len) | NI_BYTE_ARRAY_CHUNK)
+
+#define NI_BITMAP_SEPARATOR_FORMAT	", "
+#define NI_BITMAP_SEPARATOR_PARSE	" ,|\t\n"
+#define NI_BITMASK_SEPARATOR_FORMAT	" | "
+#define NI_BITMASK_SEPARATOR_PARSE	NI_BITMAP_SEPARATOR_PARSE
 
 static int		__ni_pidfile_write(const char *, unsigned int, pid_t, int);
 static const char *	__ni_build_backup_path(const char *, const char *);
@@ -125,7 +134,7 @@ ni_string_array_set(ni_string_array_t *nsa, unsigned int pos, const char *str)
 }
 
 const char *
-ni_string_array_at(ni_string_array_t *nsa, unsigned int pos)
+ni_string_array_at(const ni_string_array_t *nsa, unsigned int pos)
 {
 	if (!nsa || pos >= nsa->count)
 		return NULL;
@@ -188,6 +197,23 @@ ni_string_array_index(const ni_string_array_t *nsa, const char *str)
 	return -1;
 }
 
+unsigned int
+ni_string_array_find(const ni_string_array_t *nsa, unsigned int pos, const char *item,
+		ni_bool_t (*match)(const char *a, const char *b), const char **ret)
+{
+	if (!nsa || !match)
+		return -1U;
+
+	for (; pos < nsa->count; ++pos) {
+		if (match(nsa->data[pos], item)) {
+			if (ret)
+				*ret = item;
+			return pos;
+		}
+	}
+	return -1U;
+}
+
 /*
  * Remove string at index @pos
  */
@@ -460,6 +486,167 @@ ni_uint_array_set(ni_uint_array_t *nua, unsigned int index, unsigned int num)
 	return TRUE;
 }
 
+void
+ni_byte_array_init(ni_byte_array_t *array)
+{
+	memset(array, 0, sizeof(*array));
+}
+
+void
+ni_byte_array_destroy(ni_byte_array_t *array)
+{
+	if (array) {
+		free(array->data);
+		array->data = NULL;
+		array->len = 0;
+	}
+}
+
+ni_byte_array_t *
+ni_byte_array_new(void)
+{
+	return calloc(1, sizeof(ni_byte_array_t));
+}
+
+void
+ni_byte_array_free(ni_byte_array_t *array)
+{
+	if (array) {
+		ni_byte_array_destroy(array);
+		free(array);
+	}
+}
+
+size_t
+ni_byte_array_size(const ni_byte_array_t *array)
+{
+	if (!array || !array->data)
+		return 0;
+	return NI_BYTE_ARRAY_SIZE(array->len);
+}
+
+static ni_bool_t
+ni_byte_array_grow(ni_byte_array_t *array, size_t add)
+{
+	unsigned char *data;
+	size_t size;
+
+	if (!array || (SIZE_MAX - array->len) < add)
+		return FALSE;
+
+	size = NI_BYTE_ARRAY_SIZE(array->len);
+	if (array->data && (size - array->len) >= add)
+		return TRUE;
+
+	size = NI_BYTE_ARRAY_SIZE(array->len + add);
+	data = realloc(array->data, size);
+	if (!data)
+		return FALSE;
+
+	array->data = data;
+	memset(array->data + array->len, 0, size - array->len);
+
+	return TRUE;
+}
+
+size_t
+ni_byte_array_pad(ni_byte_array_t *array, size_t len)
+{
+	if (!len || !ni_byte_array_grow(array, len))
+		return 0;
+
+	array->len += len;
+	return len;
+}
+
+size_t
+ni_byte_array_put(ni_byte_array_t *array, const unsigned char *ptr, size_t len)
+{
+	if (!ptr || !len || !ni_byte_array_grow(array, len))
+		return 0;
+
+	memcpy(array->data + array->len, ptr, len);
+	array->len += len;
+	return len;
+}
+
+size_t
+ni_byte_array_putb(ni_byte_array_t *array, const unsigned char one)
+{
+	return ni_byte_array_put(array, &one, 1);
+}
+
+size_t
+ni_byte_array_puts(ni_byte_array_t *array, const char *str)
+{
+	const unsigned char *ptr = (const unsigned char *)str;
+	return ni_byte_array_put(array, ptr, ni_string_len(str));
+}
+
+/*
+ * Variable utils
+ */
+ni_bool_t
+ni_var_set(ni_var_t *var, const char *name, const char *value)
+{
+	if (var) {
+		ni_var_t old = {
+			.name  = var->name,
+			.value = var->value
+		};
+
+		var->name  = xstrdup(name);
+		var->value = xstrdup(value);
+
+		if ((name  && !var->name) ||
+		    (value && !var->value)) {
+			ni_var_destroy(var);
+			return FALSE;
+		}
+
+		ni_var_destroy(&old);
+		return TRUE;
+	}
+	return FALSE;
+}
+
+int
+ni_var_name_cmp(const ni_var_t *v1, const ni_var_t *v2)
+{
+	if (!v1 || !v2)
+		return v1 == v2;
+	return ni_string_cmp(v1->name, v2->name);
+}
+
+int
+ni_var_value_cmp(const ni_var_t *v1, const ni_var_t *v2)
+{
+	if (!v1 || !v2)
+		return v1 == v2;
+	return ni_string_cmp(v1->value, v2->value);
+}
+
+ni_bool_t
+ni_var_name_equal(const ni_var_t *v1, const ni_var_t *v2)
+{
+	return ni_var_name_cmp(v1, v2) == 0;
+}
+
+ni_bool_t
+ni_var_value_equal(const ni_var_t *v1, const ni_var_t *v2)
+{
+	return ni_var_value_cmp(v1, v2) == 0;
+}
+
+void
+ni_var_destroy(ni_var_t *var)
+{
+	if (var) {
+		ni_string_free(&var->name);
+		ni_string_free(&var->value);
+	}
+}
+
 /*
  * Array of variables
  */
@@ -497,33 +684,26 @@ ni_var_array_destroy(ni_var_array_t *nva)
 	memset(nva, 0, sizeof(*nva));
 }
 
-ni_var_t *
-ni_var_array_get(const ni_var_array_t *nva, const char *name)
-{
-	unsigned int i;
-	ni_var_t *var;
-
-	for (i = 0, var = nva->data; i < nva->count; ++i, ++var) {
-		if (ni_string_eq(var->name, name))
-			return var;
-	}
-	return NULL;
-}
-
-static void
-__ni_var_array_realloc(ni_var_array_t *nva, unsigned int newsize)
+static inline ni_bool_t
+ni_var_array_realloc(ni_var_array_t *nva, unsigned int newsize)
 {
 	unsigned int i;
 	ni_var_t *newdata;
 
+	if ((UINT_MAX - NI_VAR_ARRAY_CHUNK) <= newsize)
+		return FALSE;
+
 	newsize = (newsize + NI_VAR_ARRAY_CHUNK);
 	newdata = xrealloc(nva->data, newsize * sizeof(ni_var_t));
+	if (!newdata)
+		return FALSE;
 
 	nva->data = newdata;
 	for (i = nva->count; i < newsize; ++i) {
 		nva->data[i].name = NULL;
 		nva->data[i].value = NULL;
 	}
+	return TRUE;
 }
 
 ni_bool_t
@@ -562,52 +742,103 @@ ni_var_array_remove(ni_var_array_t *array, const char *name)
 	return FALSE;
 }
 
-static void
-__ni_var_array_append(ni_var_array_t *nva, const char *name, const char *value)
-{
-	ni_var_t *var;
-
-	if ((nva->count % NI_VAR_ARRAY_CHUNK) == 0)
-		__ni_var_array_realloc(nva, nva->count);
-	var = &nva->data[nva->count++];
-	var->name = xstrdup(name);
-	var->value = xstrdup(value);
-}
-
-void
-ni_var_array_set(ni_var_array_t *nva, const char *name, const char *value)
-{
-	ni_var_t *var;
-
-	if ((var = ni_var_array_get(nva, name)) == NULL) {
-		if ((nva->count % NI_VAR_ARRAY_CHUNK) == 0)
-			__ni_var_array_realloc(nva, nva->count);
-
-		var = &nva->data[nva->count++];
-		var->name = xstrdup(name);
-		var->value = NULL;
-	}
-
-	ni_string_dup(&var->value, value);
-}
-
-void
+ni_bool_t
 ni_var_array_copy(ni_var_array_t *dst, const ni_var_array_t *src)
 {
 	unsigned int i;
 
+	if (!dst || !src)
+		return FALSE;
+
 	for (i = 0; i < src->count; ++i) {
 		const ni_var_t *var = &src->data[i];
-		__ni_var_array_append(dst, var->name, var->value);
+		if (!ni_var_array_append(dst, var->name, var->value))
+			return FALSE;
 	}
+	return TRUE;
 }
 
-void
+ni_bool_t
 ni_var_array_move(ni_var_array_t *dst, ni_var_array_t *src)
 {
+	if (!dst || !src)
+		return FALSE;
+
 	ni_var_array_destroy(dst);
 	*dst = *src;
 	memset(src, 0, sizeof(*src));
+	return TRUE;
+}
+
+ni_bool_t
+ni_var_array_insert(ni_var_array_t *nva, unsigned int pos, const char *name, const char *value)
+{
+	ni_var_t *var, tmp = NI_VAR_INIT;
+
+	if (!nva)
+		return FALSE;
+
+	if (!ni_var_set(&tmp, name, value))
+		return FALSE;
+
+	if ((nva->count % NI_VAR_ARRAY_CHUNK) == 0 &&
+	    !ni_var_array_realloc(nva, nva->count)) {
+		ni_var_destroy(&tmp);
+		return FALSE;
+	}
+
+	if (pos >= nva->count) {
+		var = &nva->data[nva->count];
+	} else {
+		memmove(&nva->data[pos + 1], &nva->data[pos], (nva->count - pos) * sizeof(ni_var_t));
+		var = &nva->data[pos];
+	}
+	nva->count++;
+	var->name = tmp.name;
+	var->value = tmp.value;
+	return TRUE;
+}
+
+ni_bool_t
+ni_var_array_append(ni_var_array_t *nva, const char *name, const char *value)
+{
+	return ni_var_array_insert(nva, -1U, name, value);
+}
+
+unsigned int
+ni_var_array_find(const ni_var_array_t *nva, unsigned int pos, const ni_var_t *var,
+		ni_bool_t (*match)(const ni_var_t *, const ni_var_t *),
+		const ni_var_t **ret)
+{
+	const ni_var_t *ptr;
+
+	if (!nva || !var || !match)
+		return -1U;
+
+	for ( ; pos < nva->count; ++pos) {
+		ptr = &nva->data[pos];
+		if (match(ptr, var)) {
+			if (ret)
+				*ret = ptr;
+			return pos;
+		}
+	}
+	return -1U;
+}
+
+ni_var_t *
+ni_var_array_get(const ni_var_array_t *nva, const char *name)
+{
+	unsigned int i;
+	ni_var_t *var;
+
+	if (nva) {
+		for (i = 0, var = nva->data; i < nva->count; ++i, ++var) {
+			if (ni_string_eq(var->name, name))
+				return var;
+		}
+	}
+	return NULL;
 }
 
 int
@@ -615,6 +846,9 @@ ni_var_array_get_string(ni_var_array_t *nva, const char *name, char **p)
 {
 	ni_var_t *var;
 
+	if (!nva || !p)
+		return -1; /* Error */
+
 	if (*p) {
 		free(*p);
 		*p = NULL;
@@ -628,11 +862,33 @@ ni_var_array_get_string(ni_var_array_t *nva, const char *name, char **p)
 	return 0; /* Not found */
 }
 
+int
+ni_var_array_get_int(ni_var_array_t *nva, const char *name, int *p)
+{
+	ni_var_t *var;
+
+	if (!nva || !p)
+		return -1; /* Error */
+
+	*p = 0;
+	if ((var = ni_var_array_get(nva, name)) != NULL) {
+		if (ni_parse_int(var->value, p, 0) < 0)
+			return -1; /* Error */
+		else
+			return 1; /* Found */
+	}
+
+	return 0; /* Not found */
+}
+
 int
 ni_var_array_get_uint(ni_var_array_t *nva, const char *name, unsigned int *p)
 {
 	ni_var_t *var;
 
+	if (!nva || !p)
+		return -1; /* Error */
+
 	*p = 0;
 	if ((var = ni_var_array_get(nva, name)) != NULL) {
 		if (ni_parse_uint(var->value, p, 0) < 0)
@@ -644,11 +900,71 @@ ni_var_array_get_uint(ni_var_array_t *nva, const char *name, unsigned int *p)
 	return 0; /* Not found */
 }
 
+int
+ni_var_array_get_long(ni_var_array_t *nva, const char *name, long *p)
+{
+	ni_var_t *var;
+
+	if (!nva || !p)
+		return -1; /* Error */
+
+	*p = 0;
+	if ((var = ni_var_array_get(nva, name)) != NULL) {
+		if (ni_parse_long(var->value, p, 0) < 0)
+			return -1; /* Error */
+		else
+			return 1; /* Found */
+	}
+
+	return 0; /* Not found */
+}
+
+int
+ni_var_array_get_ulong(ni_var_array_t *nva, const char *name, unsigned long *p)
+{
+	ni_var_t *var;
+
+	if (!nva || !p)
+		return -1; /* Error */
+
+	*p = 0;
+	if ((var = ni_var_array_get(nva, name)) != NULL) {
+		if (ni_parse_ulong(var->value, p, 0) < 0)
+			return -1; /* Error */
+		else
+			return 1; /* Found */
+	}
+
+	return 0; /* Not found */
+}
+
+int
+ni_var_array_get_double(ni_var_array_t *nva, const char *name, double *p)
+{
+	ni_var_t *var;
+
+	if (!nva || !p)
+		return -1; /* Error */
+
+	*p = 0;
+	if ((var = ni_var_array_get(nva, name)) != NULL) {
+		if (ni_parse_double(var->value, p) < 0)
+			return -1; /* Error */
+		else
+			return 1; /* Found */
+	}
+
+	return 0; /* Not found */
+}
+
 int
 ni_var_array_get_boolean(ni_var_array_t *nva, const char *name, ni_bool_t *p)
 {
 	ni_var_t *var;
 
+	if (!nva || !p)
+		return -1; /* Error */
+
 	*p = 0;
 	if ((var = ni_var_array_get(nva, name)) != NULL) {
 		if (ni_parse_boolean(var->value, p))
@@ -660,61 +976,118 @@ ni_var_array_get_boolean(ni_var_array_t *nva, const char *name, ni_bool_t *p)
 	return 0; /* Not found */
 }
 
-void
+ni_bool_t
+ni_var_array_set(ni_var_array_t *nva, const char *name, const char *value)
+{
+	ni_var_t *var;
+
+	if (!nva)
+		return FALSE;
+
+	if ((var = ni_var_array_get(nva, name))) {
+		return ni_string_dup(&var->value, value);
+	} else {
+		return ni_var_array_append(nva, name, value);
+	}
+}
+
+ni_bool_t
+ni_var_array_set_int(ni_var_array_t *nva, const char *name, int value)
+{
+	char buffer[32];
+
+	snprintf(buffer, sizeof(buffer), "%d", value);
+	return ni_var_array_set(nva, name, buffer);
+}
+
+ni_bool_t
 ni_var_array_set_uint(ni_var_array_t *nva, const char *name, unsigned int value)
 {
 	char buffer[32];
 
 	snprintf(buffer, sizeof(buffer), "%u", value);
-	ni_var_array_set(nva, name, buffer);
+	return ni_var_array_set(nva, name, buffer);
 }
 
-void
-ni_var_array_set_long(ni_var_array_t *nva, const char *name, unsigned long value)
+ni_bool_t
+ni_var_array_set_long(ni_var_array_t *nva, const char *name, long value)
+{
+	char buffer[32];
+
+	snprintf(buffer, sizeof(buffer), "%ld", value);
+	return ni_var_array_set(nva, name, buffer);
+}
+
+ni_bool_t
+ni_var_array_set_ulong(ni_var_array_t *nva, const char *name, unsigned long value)
 {
 	char buffer[32];
 
 	snprintf(buffer, sizeof(buffer), "%lu", value);
-	ni_var_array_set(nva, name, buffer);
+	return ni_var_array_set(nva, name, buffer);
 }
 
-void
+ni_bool_t
 ni_var_array_set_double(ni_var_array_t *nva, const char *name, double value)
 {
 	char buffer[32];
 
 	snprintf(buffer, sizeof(buffer), "%g", value);
-	ni_var_array_set(nva, name, buffer);
+	return ni_var_array_set(nva, name, buffer);
 }
 
-void
+ni_bool_t
 ni_var_array_set_boolean(ni_var_array_t *nva, const char *name, int value)
 {
-	ni_var_array_set(nva, name, value? "yes" : "no");
+	return ni_var_array_set(nva, name, value? "yes" : "no");
+}
+
+ni_bool_t
+ni_var_array_set_var(ni_var_array_t *nva, const ni_var_t *var)
+{
+	if (!nva || !var)
+		return FALSE;
+	return ni_var_array_set(nva, var->name, var->value);
+}
+
+ni_bool_t
+ni_var_array_set_vars(ni_var_array_t *nva, const ni_var_array_t *vars, ni_bool_t overwrite)
+{
+	const ni_var_t *var;
+	unsigned int i;
+
+	if (!nva || !vars)
+		return FALSE;
+
+	for (i = 0; i < vars->count; ++i) {
+		var = &vars->data[i];
+
+		if (!overwrite && ni_var_array_get(nva, var->name))
+			continue;
+
+		if (!ni_var_array_set_var(nva, var))
+			return FALSE;
+	}
+	return TRUE;
+}
+void
+ni_var_array_sort(ni_var_array_t *nva, ni_var_compare_fn_t fn)
+{
+	qsort(nva->data, nva->count, sizeof(ni_var_t),
+			(int (*)(const void *, const void *)) fn);
 }
 
 void
-ni_var_array_list_append(ni_var_array_t **list, ni_var_array_t *nva)
+ni_var_array_sort_by_name(ni_var_array_t *nva)
 {
-	if (list && nva) {
-		while (*list)
-			list = &(*list)->next;
-		*list = nva;
-	}
+	ni_var_array_sort(nva, ni_var_name_cmp);
 }
 
-void
-ni_var_array_list_destroy(ni_var_array_t **list)
-{
-	ni_var_array_t *nva;
-
-	if (list) {
-		while ((nva = *list)) {
-			*list = nva->next;
-			ni_var_array_free(nva);
-		}
-	}
-}
+/*
+ * var_array list
+ */
+extern ni_define_slist_append(ni_var_array);
+extern ni_define_slist_destroy(ni_var_array);
 
 
 /*
@@ -805,7 +1178,7 @@ ni_bitfield_setbit(ni_bitfield_t *bf, unsigned int bit)
 {
 	if (!bf || !ni_bitfield_grow(bf, bit))
 		return FALSE;
-	bf->field[bit / 32] |= (1 << (bit % 32));
+	bf->field[bit / 32] |= NI_BIT(bit % 32u);
 	return TRUE;
 }
 
@@ -814,7 +1187,7 @@ ni_bitfield_clearbit(ni_bitfield_t *bf, unsigned int bit)
 {
 	if (!bf || !ni_bitfield_grow(bf, bit))
 		return FALSE;
-	bf->field[bit / 32] &= ~(1 << (bit % 32));
+	bf->field[bit / 32] &= ~NI_BIT(bit % 32u);
 	return TRUE;
 }
 
@@ -832,7 +1205,7 @@ ni_bitfield_testbit(const ni_bitfield_t *bf, unsigned int bit)
 {
 	if (!bf || bit / 32 >= bf->size)
 		return FALSE;
-	return !!(bf->field[bit / 32] & (1 << (bit % 32)));
+	return !!(bf->field[bit / 32] & NI_BIT(bit % 32u));
 }
 
 ni_bool_t
@@ -1103,8 +1476,8 @@ ni_bool_t
 ni_file_exists_fmt(const char *fmt, ...)
 {
 	char *path = NULL;
-	ni_bool_t ret;
 	va_list ap;
+	int ret;
 
 	if (!fmt)
 		return FALSE;
@@ -1112,7 +1485,7 @@ ni_file_exists_fmt(const char *fmt, ...)
 	va_start(ap, fmt);
 	ret = vasprintf(&path, fmt, ap) > 0;
 	va_end(ap);
-	if (!ret)
+	if (ret < 0 || !path)
 		return FALSE;
 
 	ret = ni_file_exists(path);
@@ -1179,6 +1552,20 @@ ni_string_set(char **pp, const char *value, size_t len)
 	return TRUE;
 }
 
+ni_bool_t
+ni_string_move(char **dst, char **src)
+{
+	if (!src || !dst)
+		return FALSE;
+
+	if (*dst)
+		free(*dst);
+	*dst = *src;
+	*src = NULL;
+
+	return TRUE;
+}
+
 const char *
 ni_string_printf(char **str, const char *fmt, ...)
 {
@@ -1257,18 +1644,11 @@ ni_string_split(ni_string_array_t *nsa, const char *str, const char *sep,
 const char *
 ni_string_join(char **str, const ni_string_array_t *nsa, const char *sep)
 {
-	ni_stringbuf_t buf;
-	unsigned int i;
+	ni_stringbuf_t buf = NI_STRINGBUF_INIT_DYNAMIC;
 
-	if (nsa == NULL || sep == NULL || str == NULL)
+	if (!sep || !ni_stringbuf_join(&buf, nsa, sep))
 		return NULL;
 
-	ni_stringbuf_init(&buf);
-	for (i=0; i < nsa->count; ++i) {
-		if (i)
-			ni_stringbuf_puts(&buf, sep);
-		ni_stringbuf_puts(&buf, nsa->data[i]);
-	}
 	ni_string_dup(str, buf.string);
 	ni_stringbuf_destroy(&buf);
 
@@ -1354,11 +1734,40 @@ ni_sprint_uint(unsigned int value)
 const char *
 ni_sprint_timeout(unsigned int timeout)
 {
-	if (timeout == NI_IFWORKER_INFINITE_TIMEOUT)
+	if (timeout == NI_SECONDS_INFINITE)
 		return "infinite";
 	return ni_sprint_uint(timeout);
 }
 
+const char *
+ni_format_seconds_timeout(char **str, unsigned int timeout)
+{
+	if (!str) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (timeout == NI_SECONDS_INFINITE)
+		return ni_string_printf(str, "infinite");
+	else
+		return ni_string_printf(str, "%u", timeout);
+}
+
+int
+ni_parse_seconds_timeout(const char *str, unsigned int *timeout)
+{
+	if (!timeout) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	if (ni_string_eq(str, "infinite")) {
+		*timeout = NI_SECONDS_INFINITE;
+		return 0;
+	} else {
+		return ni_parse_uint(str, timeout, 10);
+	}
+}
 
 int
 ni_parse_long(const char *input, long *result, int base)
@@ -1794,28 +2203,236 @@ ni_parse_hex(const char *string, unsigned char *data, unsigned int datasize)
 	return len;
 }
 
-const char *
-ni_format_bitmap(ni_stringbuf_t *buf, const ni_intmap_t *map,
-		unsigned int flags, const char *sep)
+unsigned int
+ni_parse_bitmap_array(unsigned int *mask, const ni_intmap_t *map, const ni_string_array_t *array,
+			ni_string_array_t *invalid)
 {
 	unsigned int i, flag;
+	const char *name;
+	unsigned int err = 0;
+
+	if (!mask || !map || !array)
+		return -1U;
+
+	for (i = 0; i < array->count; ++i) {
+		name = array->data[i];
+
+		if (!ni_parse_uint_mapped(name, map, &flag) &&
+				flag < 8 * sizeof(*mask)) {
+			*mask |= NI_BIT(flag);
+		} else {
+			if (invalid)
+				ni_string_array_append(invalid, name);
+			err++;
+		}
+	}
+
+	return err;
+}
+
+unsigned int
+ni_parse_bitmap_string(unsigned int *mask, const ni_intmap_t *map, const char *input,
+		const char *sep, ni_string_array_t *invalid)
+{
+	ni_string_array_t array = NI_STRING_ARRAY_INIT;
+	unsigned int err;
+
+	if (!mask || !map || !input)
+		return -1U;
+
+	if (!sep)
+		sep = NI_BITMAP_SEPARATOR_PARSE;
+
+	ni_string_split(&array, input, sep, 0);
+	err = ni_parse_bitmap_array(mask, map, &array, invalid);
+	ni_string_array_destroy(&array);
+
+	return err;
+}
+
+unsigned int
+ni_format_bitmap_array(ni_string_array_t *array, const ni_intmap_t *map,
+		unsigned int mask, unsigned int *done)
+{
+	unsigned int flag;
+
+	if (!array || !map)
+		return -1U;
+
+	for (; map->name; ++map) {
+		flag = NI_BIT(map->value);
+		if (mask & flag) {
+			if (ni_string_array_append(array, map->name) < 0)
+				continue;
+
+			mask &= ~flag;
+			if (done)
+				*done |= flag;
+		}
+	}
+
+	return mask;
+}
+
+const char *
+ni_format_bitmap_string(ni_stringbuf_t *buf, const ni_intmap_t *map,
+		unsigned int mask, unsigned int *done, const char *sep)
+{
+	ni_string_array_t array = NI_STRING_ARRAY_INIT;
+	const char *ptr;
 
 	if (!buf || !map)
 		return NULL;
 
 	if (ni_string_empty(sep))
-		sep = "|";
+		sep = NI_BITMAP_SEPARATOR_FORMAT;
 
-	for (i = 0; map->name; ++map) {
-		flag = (1 << map->value);
-		if (flags & flag) {
-			flags &= ~flag;
-			if (i++)
-				ni_stringbuf_puts(buf, sep);
-			ni_stringbuf_puts(buf, map->name);
+	if (ni_format_bitmap_array(&array, map, mask, done) == -1U)
+		return NULL;
+
+	ptr = ni_stringbuf_join(buf, &array, sep);
+	ni_string_array_destroy(&array);
+	return ptr;
+}
+
+const char *
+ni_format_bitmap(ni_stringbuf_t *buf, const ni_intmap_t *map,
+		unsigned int mask, const char *sep)
+{
+	return ni_format_bitmap_string(buf, map, mask, NULL, sep);
+}
+
+extern unsigned int
+ni_parse_bitmask_array(unsigned int *out, const ni_intmap_t *map, const ni_string_array_t *in, ni_string_array_t *invalid)
+{
+	unsigned int i, flag, err = 0;
+	const char *name;
+
+	if (!out || !map || !in)
+		return -1U;
+
+	for (i = 0; i < in->count; ++i){
+		name = in->data[i];
+
+		if (!ni_parse_uint_mapped(name, map, &flag))
+			*out |= flag;
+		else {
+			err++;
+			if (invalid)
+				ni_string_array_append(invalid, name);
 		}
 	}
-	return buf->string;
+
+	return err;
+}
+
+extern unsigned int
+ni_parse_bitmask_string(unsigned int *out, const ni_intmap_t *map, const char *in, const char *sep, ni_string_array_t *invalid)
+{
+	ni_string_array_t array = NI_STRING_ARRAY_INIT;
+	unsigned int err;
+
+	if (!out || !map || !in)
+		return -1U;
+
+	if (ni_string_empty(sep))
+		sep = NI_BITMASK_SEPARATOR_PARSE;
+
+	ni_string_split(&array, in, sep, 0);
+	err = ni_parse_bitmask_array(out, map, &array, invalid);
+	ni_string_array_destroy(&array);
+
+	return err;
+}
+
+extern unsigned int
+ni_parse_bitmask(unsigned int *out, const ni_intmap_t *map, const char *in, const char *sep, ni_string_array_t *invalid)
+{
+	ni_string_array_t temp_inval = NI_STRING_ARRAY_INIT;
+	const char *val;
+	unsigned int i, flag, err = 0;
+
+	if (ni_parse_bitmask_string(out, map, in, sep, &temp_inval)) {
+		for (i = 0; i < temp_inval.count; i++) {
+			val = temp_inval.data[i];
+
+			if (!ni_parse_uint(val, &flag, 16))
+				*out |= flag;
+			else {
+				err++;
+				if (invalid)
+					ni_string_array_append(invalid, val);
+			}
+		}
+	}
+	ni_string_array_destroy(&temp_inval);
+	return err;
+}
+
+extern unsigned int
+ni_format_bitmask_array(ni_string_array_t *out, const ni_intmap_t *map, unsigned int in, unsigned int *done)
+{
+	unsigned int value;
+
+	if (!out || !map)
+		return -1U;
+
+	for (; map->name; ++map) {
+		value = map->value;
+		if ((in & value) == value) {
+			if (ni_string_array_append(out, map->name) < 0)
+				continue;
+
+			in &= ~value;
+			if (done)
+				*done |= value;
+		}
+	}
+
+	return in;
+}
+
+extern const char *
+ni_format_bitmask_string(ni_stringbuf_t *out, const ni_intmap_t *map, unsigned int in, unsigned int *done, const char *sep)
+{
+	ni_string_array_t array = NI_STRING_ARRAY_INIT;
+	const char *ptr;
+
+	if (!out || !map)
+		return NULL;
+
+	if (ni_string_empty(sep))
+		sep = NI_BITMASK_SEPARATOR_FORMAT;
+
+	if (ni_format_bitmask_array(&array, map, in, done) == -1U)
+		return NULL;
+
+	ptr = ni_stringbuf_join(out, &array, sep);
+	ni_string_array_destroy(&array);
+	return ptr;
+}
+
+extern const char *
+ni_format_bitmask(ni_stringbuf_t *out, const ni_intmap_t *map, unsigned int in, const char *sep)
+{
+	unsigned int done = 0;
+	size_t offset = 0;
+
+	if (!out || !map)
+		return NULL;
+
+	if (ni_string_empty(sep))
+		sep = NI_BITMASK_SEPARATOR_FORMAT;
+
+	offset = out->len;
+	ni_format_bitmask_string(out, map, in, &done, sep);
+	if (done != in) {
+		in &= ~done;
+		if (!ni_stringbuf_empty(out))
+			ni_stringbuf_puts(out, sep);
+		ni_stringbuf_printf(out, "0x%x", in);
+	}
+	return out->string ? out->string + offset : NULL;
 }
 
 static ni_bool_t
@@ -1930,6 +2547,42 @@ ni_stringbuf_empty(const ni_stringbuf_t *sb)
 	return sb->len == 0; /* bool */
 }
 
+const char *
+ni_stringbuf_join(ni_stringbuf_t *buf, const ni_string_array_t *nsa, const char *sep)
+{
+	unsigned int i;
+	size_t len;
+
+	if (!buf || !nsa)
+		return NULL;
+
+	len = buf->len;
+	for (i = 0; i < nsa->count; ++i) {
+		if (sep && buf->len)
+			ni_stringbuf_puts(buf, sep);
+		ni_stringbuf_puts(buf, nsa->data[i]);
+	}
+	return buf->string ? buf->string + len : NULL;
+}
+
+const char *
+ni_stringbuf_join_uint(ni_stringbuf_t *buf, const ni_uint_array_t *nsa, const char *sep)
+{
+	unsigned int i;
+	size_t len;
+
+	if (!buf || !nsa)
+		return NULL;
+
+	len = buf->len;
+	for (i = 0; i < nsa->count; ++i) {
+		if (sep && buf->len)
+			ni_stringbuf_puts(buf, sep);
+		ni_stringbuf_printf(buf, "%u", nsa->data[i]);
+	}
+	return buf->string ? buf->string + len : NULL;
+}
+
 inline static size_t
 __ni_stringbuf_size(ni_stringbuf_t *sb, size_t len)
 {
@@ -2261,7 +2914,7 @@ ni_pidfile_write(const char *pidfile, unsigned int permissions, pid_t pid)
  * Check for presence of pidfile
  *  0:	no or stale pidfile
  *  >0:	pid of active process
- *  <0:	error occured
+ *  <0:	error occurred
  */
 pid_t
 ni_pidfile_check(const char *pidfile)
@@ -2645,6 +3298,12 @@ ni_sibling_path_printf(const char *path, const char *fmt, ...)
 /*
  * Utility functions for handling uuids
  */
+void
+ni_uuid_init(ni_uuid_t *uuid)
+{
+	memset(uuid, 0, sizeof(*uuid));
+}
+
 const char *
 ni_uuid_print(const ni_uuid_t *uuid)
 {
@@ -2653,6 +3312,7 @@ ni_uuid_print(const ni_uuid_t *uuid)
 
 	if (!uuid)
 		return NULL;
+
 	if (ni_uuid_is_null(uuid))
 		return "";
 
@@ -2802,7 +3462,7 @@ ni_srandom(void)
 }
 
 /*
- * Alloc helpers with NULLL check
+ * Alloc helpers with NULL check
  */
 void *
 xmalloc(size_t size)
