@@ -1,7 +1,8 @@
 /*
  *	XML objects - document and node
  *
- *	Copyright (C) 2009-2012  Olaf Kirch <okir@suse.de>
+ *	Copyright (C) 2009-2012 Olaf Kirch <okir@suse.de>
+ *	Copyright (C) 2009-2024 SUSE LLC
  *
  *	This program is free software; you can redistribute it and/or modify
  *	it under the terms of the GNU General Public License as published by
@@ -13,11 +14,8 @@
  *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *	GNU General Public License for more details.
  *
- *	You should have received a copy of the GNU General Public License along
- *	with this program; if not, see <http://www.gnu.org/licenses/> or write 
- *	to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
- *	Boston, MA 02110-1301 USA.
- *
+ *	You should have received a copy of the GNU General Public License
+ *	along with this program. If not, see <http://www.gnu.org/licenses/>.
  */
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -25,11 +23,13 @@
 
 #include <wicked/xml.h>
 #include <wicked/logging.h>
+#include "array_priv.h"
 #include "util_priv.h"
+#include "slist_priv.h"
 #include <inttypes.h>
 
-#define XML_DOCUMENTARRAY_CHUNK		1
-#define XML_NODEARRAY_CHUNK		8
+#define XML_DOCUMENT_ARRAY_CHUNK	8
+#define XML_NODE_ARRAY_CHUNK		8
 
 xml_document_t *
 xml_document_new()
@@ -44,19 +44,19 @@ xml_document_new()
 xml_node_t *
 xml_document_root(xml_document_t *doc)
 {
-	return doc->root;
+	return doc ? doc->root : NULL;
 }
 
 const char *
-xml_document_type(const xml_document_t *doc)
+xml_document_dtd(const xml_document_t *doc)
 {
-	return doc->dtd;
+	return doc ? doc->dtd : NULL;
 }
 
 void
 xml_document_set_root(xml_document_t *doc, xml_node_t *root)
 {
-	if (doc->root != root) {
+	if (doc && doc->root != root) {
 		xml_node_free(doc->root);
 		doc->root = root;
 	}
@@ -65,9 +65,12 @@ xml_document_set_root(xml_document_t *doc, xml_node_t *root)
 xml_node_t *
 xml_document_take_root(xml_document_t *doc)
 {
-	xml_node_t *root = doc->root;
+	xml_node_t *root = NULL;
 
-	doc->root = NULL;
+	if (doc) {
+		root = doc->root;
+		doc->root = NULL;
+	}
 	return root;
 }
 
@@ -81,6 +84,63 @@ xml_document_free(xml_document_t *doc)
 	}
 }
 
+xml_document_t *
+xml_document_create(const char *dtd, xml_node_t *root)
+{
+	xml_document_t *doc;
+
+	if ((doc = calloc(1, sizeof(*doc)))) {
+		if (ni_string_dup(&doc->dtd, dtd)) {
+			doc->root = root;
+			return doc;
+		}
+		xml_document_free(doc);
+	}
+	return NULL;
+}
+
+ni_bool_t
+xml_document_expand(xml_document_array_t *docs, xml_document_t *raw)
+{
+	ni_bool_t ret = TRUE;
+	xml_document_t *doc;
+	xml_node_t *node;
+	xml_node_t *next;
+
+	if (!docs || !raw || !raw->root)
+		return FALSE;
+
+	/*
+	 * Append if "raw" contains single named document root.
+	 */
+	if (!ni_string_empty(raw->root->name))
+		return xml_document_array_append(docs, raw);
+
+	/*
+	 * Otherwise, expand the just parsed data into multiple
+	 * "normalized" documents with one named root node each.
+	 */
+	for (node = raw->root->children; node; node = next) {
+		next = node->next;
+
+		if (ni_string_empty(node->name))
+			continue;
+
+		xml_node_detach(node);
+
+		if (!(doc = xml_document_create(raw->dtd, node)))
+			xml_node_free(node);
+		else if (!xml_document_array_append(docs, doc))
+			xml_document_free(doc);
+		else continue;
+
+		ret = FALSE;
+	}
+	if (ret)
+		xml_document_free(raw);
+	return ret;
+}
+
 /*
  * Helper functions for xml node list management
  */
@@ -305,79 +365,65 @@ xml_node_free(xml_node_t *node)
 	free(node);
 }
 
-void
+ni_bool_t
 xml_node_set_cdata(xml_node_t *node, const char *cdata)
 {
-	ni_string_dup(&node->cdata, cdata);
+	return ni_string_dup(&node->cdata, cdata);
 }
 
-void
+ni_bool_t
 xml_node_set_int(xml_node_t *node, int value)
 {
-	char buffer[32];
 
-	snprintf(buffer, sizeof(buffer), "%d", value);
-	ni_string_dup(&node->cdata, buffer);
+	return node && ni_string_printf(&node->cdata, "%d", value);
 }
 
-void
+ni_bool_t
 xml_node_set_int64(xml_node_t *node, int64_t value)
 {
-	char buffer[32];
-
-	snprintf(buffer, sizeof(buffer), "%"PRId64, value);
-	ni_string_dup(&node->cdata, buffer);
+	return node && ni_string_printf(&node->cdata, "%"PRId64, value);
 }
 
-void
+ni_bool_t
 xml_node_set_uint(xml_node_t *node, unsigned int value)
 {
-	char buffer[32];
-
-	snprintf(buffer, sizeof(buffer), "%u", value);
-	ni_string_dup(&node->cdata, buffer);
+	return node && ni_string_printf(&node->cdata, "%u", value);
 }
 
-void
+ni_bool_t
 xml_node_set_uint64(xml_node_t *node, uint64_t value)
 {
-	char buffer[32];
-
-	snprintf(buffer, sizeof(buffer), "%"PRIu64, value);
-	ni_string_dup(&node->cdata, buffer);
+	return node && ni_string_printf(&node->cdata, "%"PRIu64, value);
 }
 
-void
+ni_bool_t
 xml_node_set_uint_hex(xml_node_t *node, unsigned int value)
 {
-	char buffer[32];
-
-	snprintf(buffer, sizeof(buffer), "0x%x", value);
-	ni_string_dup(&node->cdata, buffer);
+	return node && ni_string_printf(&node->cdata, "0x%x", value);
 }
 
-void
+ni_bool_t
 xml_node_add_attr(xml_node_t *node, const char *name, const char *value)
 {
-	ni_var_array_set(&node->attrs, name, value);
+	return ni_var_array_set(&node->attrs, name, value);
 }
 
-void
+ni_bool_t
 xml_node_add_attr_uint(xml_node_t *node, const char *name, unsigned int value)
 {
-	ni_var_array_set_uint(&node->attrs, name, value);
+	return ni_var_array_set_uint(&node->attrs, name, value);
 }
 
-void
+ni_bool_t
 xml_node_add_attr_ulong(xml_node_t *node, const char *name, unsigned long value)
 {
-	ni_var_array_set_long(&node->attrs, name, value);
+	return ni_var_array_set_ulong(&node->attrs, name, value);
 }
 
-void
+ni_bool_t
 xml_node_add_attr_double(xml_node_t *node, const char *name, double value)
 {
-	ni_var_array_set_double(&node->attrs, name, value);
+	return ni_var_array_set_double(&node->attrs, name, value);
 }
 
 const ni_var_t *
@@ -468,12 +514,22 @@ xml_node_get_next_child(const xml_node_t *top, const char *name, const xml_node_
 	return NULL;
 }
 
-inline xml_node_t *
+xml_node_t *
 xml_node_get_child(const xml_node_t *node, const char *name)
 {
 	return xml_node_get_next_child(node, name, NULL);
 }
 
+const char *
+xml_node_get_child_cdata(const xml_node_t *node, const char *name)
+{
+	const xml_node_t *child;
+
+	if ((child = xml_node_get_child(node, name)))
+		return child->cdata;
+	return NULL;
+}
+
 /*
  * Find a child element given its name and a list of attributes
  */
@@ -717,85 +773,20 @@ xml_node_match_attrs(const xml_node_t *node, const ni_var_array_t *attrlist)
 /*
  * XML document arrays
  */
-void
-xml_document_array_init(xml_document_array_t *array)
-{
-	memset(array, 0, sizeof(*array));
-}
-
-void
-xml_document_array_destroy(xml_document_array_t *array)
-{
-	unsigned int i;
-
-	for (i = 0; i < array->count; ++i)
-		xml_document_free(array->data[i]);
-
-	if (array->data)
-		free(array->data);
-	memset(array, 0, sizeof(*array));
-}
-
-xml_document_array_t *
-xml_document_array_new(void)
-{
-	xml_document_array_t *array;
-
-	array = xcalloc(1, sizeof(*array));
-	return array;
-}
-
-void
-xml_document_array_free(xml_document_array_t *array)
-{
-	xml_document_array_destroy(array);
-	free(array);
-}
-
-static void
-__xml_document_array_realloc(xml_document_array_t *array, unsigned int newsize)
-{
-	xml_document_t **newdata;
-	unsigned int i;
-
-	newsize = (newsize + XML_DOCUMENTARRAY_CHUNK) + 1;
-	newdata = xrealloc(array->data, newsize * sizeof(xml_document_t *));
-
-	array->data = newdata;
-	for (i = array->count; i < newsize; ++i)
-		array->data[i] = NULL;
-}
-
-void
-xml_document_array_append(xml_document_array_t *array, xml_document_t *doc)
-{
-	if ((array->count % XML_DOCUMENTARRAY_CHUNK) == 0)
-		__xml_document_array_realloc(array, array->count);
-
-	array->data[array->count++] = doc;
-}
+extern ni_define_ptr_array_init(xml_document);
+extern ni_define_ptr_array_destroy(xml_document);
+static ni_define_ptr_array_realloc(xml_document, XML_DOCUMENT_ARRAY_CHUNK);
+extern ni_define_ptr_array_append(xml_document);
+extern ni_define_ptr_array_insert(xml_document);
+extern ni_define_ptr_array_index(xml_document);
 
 /*
  * XML node arrays
  */
-void
-xml_node_array_init(xml_node_array_t *array)
-{
-	memset(array, 0, sizeof(*array));
-}
-
-void
-xml_node_array_destroy(xml_node_array_t *array)
-{
-	unsigned int i;
-
-	for (i = 0; i < array->count; ++i)
-		xml_node_free(array->data[i]);
-
-	if (array->data)
-		free(array->data);
-	memset(array, 0, sizeof(*array));
-}
+extern ni_define_ptr_array_init(xml_node);
+extern ni_define_ptr_array_destroy(xml_node);
+static ni_define_ptr_array_realloc(xml_node, XML_NODE_ARRAY_CHUNK)
+extern ni_define_ptr_array_append(xml_node);
 
 xml_node_array_t *
 xml_node_array_new(void)
@@ -813,32 +804,6 @@ xml_node_array_free(xml_node_array_t *array)
 	free(array);
 }
 
-static void
-__xml_node_array_realloc(xml_node_array_t *array, unsigned int newsize)
-{
-	xml_node_t **newdata;
-	unsigned int i;
-
-	newsize = (newsize + XML_NODEARRAY_CHUNK) + 1;
-	newdata = xrealloc(array->data, newsize * sizeof(array->data[0]));
-
-	array->data = newdata;
-	for (i = array->count; i < newsize; ++i)
-		array->data[i] = NULL;
-}
-
-void
-xml_node_array_append(xml_node_array_t *array, xml_node_t *node)
-{
-	if (!array || !node)
-		return;
-
-	if ((array->count % XML_NODEARRAY_CHUNK) == 0)
-		__xml_node_array_realloc(array, array->count);
-
-	array->data[array->count++] = xml_node_clone_ref(node);
-}
-
 xml_node_t *
 xml_node_create(xml_node_t *parent, const char *name)
 {
@@ -860,3 +825,105 @@ xml_node_dict_set(xml_node_t *parent, const char *name, const char *value)
 	child = xml_node_create(parent, name);
 	xml_node_set_cdata(child, value);
 }
+
+typedef struct xml_node_name_path	xml_node_name_path_t;
+
+struct xml_node_name_path {
+	xml_node_name_path_t *	next;
+	ni_string_array_t	path;
+};
+
+static xml_node_name_path_t *
+xml_node_name_path_new(void)
+{
+	return calloc(1, sizeof(xml_node_name_path_t));
+}
+
+static void
+xml_node_name_path_free(xml_node_name_path_t *item)
+{
+	if (item) {
+		ni_string_array_destroy(&item->path);
+		free(item);
+	}
+}
+
+static inline ni_bool_t
+xml_node_name_path_match(xml_node_t *node, const ni_string_array_t *path)
+{
+	ni_bool_t ret = FALSE;
+	const char *name;
+	unsigned int i;
+
+	if (!node || !path)
+		return FALSE;
+
+	for (i = 0; i < path->count; ++i) {
+		name = path->data[i];
+
+		if (!node || !ni_string_eq(node->name, name))
+			return FALSE;
+
+		node = node->parent;
+		ret = TRUE;
+	}
+	return ret;
+}
+
+static ni_define_slist_destroy(xml_node_name_path);
+static ni_define_slist_append(xml_node_name_path);
+
+static ni_bool_t
+xml_node_name_path_list_create(xml_node_name_path_t **list, const char * const npaths[])
+{
+	xml_node_name_path_t *item;
+	const char * const *nptr;
+
+	if (!list || !npaths)
+		return FALSE;
+
+	for (nptr = npaths; *nptr; ++nptr) {
+		if (!(item = xml_node_name_path_new())) {
+			xml_node_name_path_list_destroy(list);
+			return FALSE;
+		}
+		if (!ni_string_split(&item->path, *nptr, "/", 0))
+			xml_node_name_path_free(item);
+		else
+			xml_node_name_path_list_append(list, item);
+	}
+	return TRUE;
+}
+
+static void
+xml_node_name_path_list_hide_cdata(xml_node_t *node,
+		const xml_node_name_path_t *list, const char *hidden)
+{
+	const xml_node_name_path_t *item;
+	xml_node_t *child;
+
+	ni_slist_foreach(list, item) {
+		if (!xml_node_name_path_match(node, &item->path))
+			continue;
+
+		xml_node_set_cdata(node, hidden);
+	}
+
+	for (child = node->children; child; child = child->next)
+		xml_node_name_path_list_hide_cdata(child, list, hidden);
+}
+
+extern void
+xml_node_hide_cdata(xml_node_t *node, const char * const npaths[], const char *hidden)
+{
+	xml_node_name_path_t *list = NULL;
+
+	if (!node || !npaths)
+		return;
+
+	if (!xml_node_name_path_list_create(&list, npaths) || !list)
+		return;
+
+	xml_node_name_path_list_hide_cdata(node, list, hidden);
+	xml_node_name_path_list_destroy(&list);
+}
